/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: daemon.c
 *
 */

#define _GNU_SOURCE	/* Needed to get prototype for lssek64() */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <semaphore.h>
#include <pthread.h>
#include <limits.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "fullengine.h"
#include "dmonmsg.h"
#include "daemon.h"
#include "cluster.h"
#include "remote.h"
#include "engine.h"
#include "dload.h"
#include "discover.h"
#include "volume.h"
#include "message.h"
#include "memman.h"
#include "internalAPI.h"
#include "dm.h"

		
#ifndef O_DIRECT
#define O_DIRECT 0
#endif


#define SEND_MSG(msg)                                                                             \
{                                                                                                 \
	int rc;                                                                                   \
	LOG_DEBUG("Send %s to node %s.\n", msg_name((msg)->cmd), nodeid_to_string(&(msg)->node)); \
	rc = ece_funcs->send_msg(msg);                                                            \
	if (rc != 0) {                                                                            \
		LOG_SERIOUS("send_msg() returned error code %d: %s\n", rc, evms_strerror(rc));    \
	}                                                                                         \
}


typedef struct msg_pool_ent_s {
	LIST_FIELDS(struct msg_pool_ent_s);
	ece_msg_t msg;
} msg_pool_ent_t;

static msg_pool_ent_t msg_pool = LIST_DECL(&msg_pool);
static pthread_mutex_t msg_pool_mutex = PTHREAD_MUTEX_INITIALIZER;
static sem_t msg_pool_sem;

/*
 * Get a message from the message pool or allocate a new one.
 * get_msg() is guaranteed to return a pointer to an ece_msg_t.
 * It may take a while or get blocked forever if there is a
 * deadlock.  But if it returns, the pointer will be valid.
 */
ece_msg_t * get_msg(const ece_msg_t * copy_msg) {

	msg_pool_ent_t * msg_pool_ent = NULL;
	ece_msg_t * msg = NULL;

	pthread_mutex_lock(&msg_pool_mutex);

	while (msg_pool_ent == NULL) {
		if (!LIST_EMPTY(&msg_pool)) {
			msg_pool_ent = msg_pool.next;
			LIST_REMOVE(msg_pool_ent);

		} else {
			/* Try to allocate a new one. */
			msg_pool_ent = engine_alloc(sizeof(msg_pool_ent_t));

			if (msg_pool_ent == NULL) {
				/*
				 * Couldn't allocate a new one.
				 * Wait until one becomes available.
				 */
				sem_init(&msg_pool_sem, 0, 0);
				pthread_mutex_unlock(&msg_pool_mutex);
				sem_wait(&msg_pool_sem);
				pthread_mutex_lock(&msg_pool_mutex);
			}
		}
	}

	msg = &(msg_pool_ent->msg);
	*msg = *copy_msg;

	pthread_mutex_unlock(&msg_pool_mutex);
	return msg;
}


/*
 * Return a message to the message pool.
 */
void free_msg(ece_msg_t * msg) {

	int sem_value = 0;
	msg_pool_ent_t * msg_pool_ent = (msg_pool_ent_t *) ((void *)msg - offsetof(msg_pool_ent_t, msg));

	pthread_mutex_lock(&msg_pool_mutex);

	LIST_INSERT_BEFORE(msg_pool_ent, &msg_pool);

	sem_getvalue(&msg_pool_sem, &sem_value);
	if (sem_value != 0) {
		sem_post(&msg_pool_sem);
	}

	pthread_mutex_unlock(&msg_pool_mutex);
}


static void msg_get_version(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_get_version_response_t response;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Daemon message version is %d.%d.%d.\n",
		  EVMS_DMONMSG_MAJOR_VERSION, EVMS_DMONMSG_MINOR_VERSION, EVMS_DMONMSG_PATCH_LEVEL);
	response.version.major      = EVMS_DMONMSG_MAJOR_VERSION;
	response.version.minor      = EVMS_DMONMSG_MINOR_VERSION;
	response.version.patchlevel = EVMS_DMONMSG_PATCH_LEVEL;

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_get_version_response_t);

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(0);
}


static void msg_lock_engine(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_lock_engine_t * lock_engine_msg = msg->msg;
	msg_lock_engine_response_t response;

	const char * lock_msg;

	LOG_PROC_ENTRY();

	response_msg = get_msg(msg);

	response.status = lock_engine(lock_engine_msg->mode, &lock_msg);
	LOG_DEBUG("lock_engine() returned error code %d: %s\n", response.status, strerror(response.status));

	if (response.status == 0) {
		/* Update the lock file contents. */
		const char * node_name = nodeid_to_string(&msg->node);
		size_t req_lock_file_size = sizeof(lock_file_t) + strlen(node_name) + 1;

		if (lock_file_size < req_lock_file_size) {
			lock_file = engine_realloc(lock_file, req_lock_file_size);
			if (lock_file != NULL) {
				lock_file_size = req_lock_file_size;

			} else {
				LOG_CRITICAL("Error reallocating memory for the lock file.\n");
				response.status = ENOMEM;
			}
		}

		if (response.status == 0) {
			strcpy(lock_file->node, node_name);
			lseek(lock_file_fd, offsetof(lock_file_t, node), SEEK_SET);
			write(lock_file_fd, node_name, strlen(node_name) + 1);
		}

	} else {
		ece_msg_t * lock_error_response_msg = get_msg(msg);

		LOG_DEBUG("Lock message is : %s\n", lock_msg);
		lock_error_response_msg->cmd = MSG_USER_MSG;
		lock_error_response_msg->corrolator = 0;
		lock_error_response_msg->msg = (void *) lock_msg;
		lock_error_response_msg->size = strlen(lock_msg) + 1;

		SEND_MSG(lock_error_response_msg);

		free_msg(lock_error_response_msg);
	}

	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_lock_engine_response_t);

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


static void msg_unlock_engine(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_unlock_engine_response_t response;

	LOG_PROC_ENTRY();
	
	response.status = unlock_engine(ENGINE_READWRITE);

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_unlock_engine_response_t);

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


static void msg_chknod(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_chknod_t * chknod_msg = msg->msg;
	msg_chknod_response_t response;

	LOG_PROC_ENTRY();
	LOG_DEBUG("chknod for %s, major: %#x, minor: %#x.\n", chknod_msg->name, chknod_msg->major, chknod_msg->minor);

	response.status = hasa_dev_node(chknod_msg->name, chknod_msg->major, chknod_msg->minor);

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_chknod_response_t);

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


static void msg_mknod(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_mknod_t * mknod_msg = msg->msg;
	msg_mknod_response_t response;

	LOG_PROC_ENTRY();
	LOG_DEBUG("mknod for %s, major: %#x, minor: %#x.\n", mknod_msg->name, mknod_msg->major, mknod_msg->minor);

	response.status = ensure_dev_node(mknod_msg->name, mknod_msg->major, mknod_msg->minor);

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_mknod_response_t);

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


static int free_disk(ADDRESS object,
		     TAG     object_tag,
		     ADDRESS object_handle,
		     ADDRESS PParameters) {

	storage_object_t * disk = (storage_object_t *) object;

	engine_free_logical_disk(disk);

	return DLIST_SUCCESS;
}


static void msg_discover_disks(const ece_msg_t * msg) {

	int rc = 0;
	ece_msg_t * response_msg;
	msg_discover_disks_response_t   err_response = {num_disks: 0};
	msg_discover_disks_response_t * response = NULL;
	dlist_t disk_list;
	uint disk_count = 0;
	TAG tag;
	void * thing;

	LOG_PROC_ENTRY();

	response_msg = get_msg(msg);
	response_msg->msg = &err_response;
	response_msg->size = sizeof(msg_discover_disks_response_t);

	disk_list = CreateList();
	if (disk_list != NULL) {
		rc = discover_logical_disks(disk_list);

		if (rc == 0) {
			uint alloc_size;

			GetListSize(DiskList, &disk_count);
			alloc_size = sizeof(msg_discover_disks_response_t) +
				    (sizeof(storage_object_t) * disk_count);
			response = engine_alloc(alloc_size);

			if (response != NULL) {
				int i = 0;
				storage_object_t * disk;

				response_msg->msg  = response;
				response_msg->size = alloc_size;

				DLIST_FOR_EACH(disk, DiskList, rc) {
					response->disks[i] = *disk;
					/*
					 * Substitute the plug-in
					 * pointer with the plug-in ID
					 * so that the remote Engine
					 * can lookup its own plug-in
					 * pointer.
					 */
					response->disks[i].plugin = (plugin_record_t *) ((unsigned long) disk->plugin->id);
					i++;
				}

				if ((rc == DLIST_END_OF_LIST) || (rc == DLIST_EMPTY)) {
					rc = DLIST_SUCCESS;
				}

				response->status = rc;
				if (rc == 0) {
					response->num_disks = disk_count;
				}

			} else {
				LOG_CRITICAL("Error allocating memory for a response.\n");
				err_response.status = ENOMEM;
				rc = ENOMEM;
			}

			ForEachItem(disk_list, free_disk, NULL, TRUE);
			DestroyList(&disk_list, FALSE);

			/*
			 * Can't use DeleteAllItems(VolumeDeleteList, TRUE)
			 * because it uses free().  Volumes are
			 * allocted with engine_alloc() which may add a
			 * debug header.  Extract each item and use
			 * engine_free().
			 */
			while (BlindExtractObject(DiskList, &tag,NULL, &thing) == DLIST_SUCCESS) {
				engine_free(thing);
			}

		} else {
			LOG_SERIOUS("Error code %d from discovery of disks.\n", rc);
			err_response.status = rc;
		}

	} else {
		LOG_CRITICAL("Error allocating memory for a disk dlist.\n");
		err_response.status = ENOMEM;
		rc = ENOMEM;
	}

	response_msg->cmd |= COMMAND_RESPONSE;

	SEND_MSG(response_msg);

	/* engine_free() handles NULL pointers. */
	engine_free(response);
	free_msg(response_msg);

	LOG_PROC_EXIT_INT(rc);
}


static void msg_get_dev_node_names(const ece_msg_t * msg) {
	ece_msg_t  * response_msg;
	msg_get_dev_node_names_t * get_dev_node_names_msg = msg->msg;
	msg_get_dev_node_names_response_t   err_response;
	msg_get_dev_node_names_response_t * response = NULL;
	int response_size;
	char * * name_list = NULL;
	int count = 0;
        int rc = 0;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Get dev node names for directory %s.\n", get_dev_node_names_msg->dir_name);

	response_msg = get_msg(msg);

	rc = get_dev_node_names(get_dev_node_names_msg->dir_name, &name_list, &count);

	if (rc == 0) {
		int i;

		response_size = sizeof(msg_get_dev_node_names_response_t);
		for (i = 0; i < count; i++) {
			response_size += strlen(name_list[i]) + 1;
		}

		response = engine_alloc(response_size);
		if (response != NULL) {
			char * pch;

			response_msg->size = response_size;
			response_msg->msg = response;
			response->status = 0;
			response->count = count;

			/* Flatten the array into the response buffer. */
			pch = response->names;
			for (i = 0; i < count; i++) {
				strcpy(pch, name_list[i]);
				pch += strlen(pch) + 1;
			}

		} else {
			rc = ENOMEM;
			err_response.status = rc;
			err_response.count = 0;
			response_msg->msg = &err_response;
			response_msg->size = sizeof(msg_get_dev_node_names_response_t);
		}
		
		for (i = 0; i < count; i++) {
			engine_free(name_list[i]);
		}
                engine_free(name_list);

	} else {
		err_response.status = rc;
		err_response.count = 0;
		response_msg->msg = &err_response;
		response_msg->size = sizeof(msg_get_dev_node_names_response_t);
	}

	response_msg->cmd |= COMMAND_RESPONSE;

	SEND_MSG(response_msg);

	/* engine_free() handles NULL pointers. */
	engine_free(response);
	free_msg(response_msg);
	LOG_PROC_EXIT_INT(rc);
}


static void msg_open(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_open_t * open_msg = msg->msg;
	msg_open_response_t response;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Open file %s, flags: %#x, mode: %#x.\n", open_msg->name, open_msg->flags, open_msg->mode);

	response.fd = open(open_msg->name, open_msg->flags, open_msg->mode);
	LOG_DEBUG("File descriptor is %d.\n", response.fd);

	if (response.fd != -1) {
		response.status = 0;

	} else {
		response.status = errno;
	}

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_open_response_t);

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_DEBUG("rc: %d fd: %d\n", response.status, response.fd);
	LOG_PROC_EXIT_INT(response.status);
}


#define BLKSSZGET _IO(0x12,104)

static int alloc_aligned(int fd, size_t size, int offset_to_align,
			 void * * buffer, void * * aligned_buffer) {

	int rc = 0;
	int status;
	uint hardsector_size;
	void * buff;
	void * aligned_buff;

	LOG_PROC_ENTRY();

	/* Initialize in case of errors. */
	*buffer = NULL;
	*aligned_buffer = NULL;

	status = ioctl(fd, BLKSSZGET, &hardsector_size);

	if (status == 0) {
		hardsector_size = max(hardsector_size, 1024);

		buff = engine_alloc(size + offset_to_align + hardsector_size - 1);

		if (buff != NULL) {
			aligned_buff = (void *) ((((unsigned long) buff + offset_to_align + hardsector_size - 1)
						  & ~(hardsector_size - 1))
						 - offset_to_align);

			/* Sanity check */
			if (((unsigned long)(aligned_buff + offset_to_align) % hardsector_size) == 0) {
				*buffer = buff;
				*aligned_buffer = aligned_buff;
			} else {
				rc = EFAULT;
				LOG_SERIOUS("NOT Aligned!!: aligned_buff: %p, offset_to_align: %#x, hardsector_size: %#x\n", aligned_buff, offset_to_align, hardsector_size);
				engine_free(buff);
			}

		} else {
			/* Unable to get memory. */
			rc = ENOMEM;
		}

	} else {
		/* Unable to get hard sector size. */
		rc = errno;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void msg_read(const ece_msg_t * msg) {
	ece_msg_t  * response_msg;
	msg_read_t * read_msg = msg->msg;
	msg_read_response_t   err_response;
	void *                buffer;
	msg_read_response_t * response = NULL;
	u_int64_t response_size = sizeof(msg_read_response_t) + read_msg->length;
	int rc = 0;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Read %d bytes from fd %d at offset %"PRIu64".\n", read_msg->length, read_msg->fd, read_msg->offset);

	response_msg = get_msg(msg);

	if (fcntl(read_msg->fd, F_GETFL, 0) & O_DIRECT) {
		rc = alloc_aligned(read_msg->fd, response_size, sizeof(msg_read_response_t),
				   &buffer, (void *) &response);

	} else {
		buffer = engine_alloc(response_size);
		if (buffer != NULL) {
			response = buffer;

		} else {
                        rc = ENOMEM;
			LOG_CRITICAL("Unable to allocate a response buffer.\n");
		}
	}

	if (response != NULL) {
		off64_t curr_offset;
		int     bytes_read;

		response_msg->msg = response;

		/*
		 * We don't have to align the file offset and length if opened
		 * with O_DIRECT.  The caller should have done that.  If it's
		 * not aligned, the caller will get the error.
		 */
		curr_offset = lseek64(read_msg->fd, (off64_t) read_msg->offset, SEEK_SET);

		if (curr_offset == read_msg->offset) {
			bytes_read = read((int) read_msg->fd, (void *) &response->data, (size_t) read_msg->length);

			if (bytes_read > 0) {
				response->status = 0;
				response->bytes_read = (u_int32_t) bytes_read;
				response_msg->size = sizeof(msg_read_response_t) + bytes_read;

			} else {
				rc = errno;
				response->status = errno;
				response_msg->size = sizeof(msg_read_response_t);
			}

		} else {
			rc = ESPIPE;
			response->status = ESPIPE; /* Illegal seek */
			response_msg->size = sizeof(msg_read_response_t);
		}

	} else {
		err_response.bytes_read = 0;
		err_response.status = rc;
		response_msg->msg = &err_response;
		response_msg->size = sizeof(msg_read_response_t);
	}

	response_msg->cmd |= COMMAND_RESPONSE;

	SEND_MSG(response_msg);

	/* engine_free() handles NULL pointers. */
	LOG_DEBUG("rc: %d bytes read: %d\n", response->status, response->bytes_read);
	engine_free(buffer);
	free_msg(response_msg);
	LOG_PROC_EXIT_INT(rc);
}


static void msg_write(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_write_t * write_msg = msg->msg;
	msg_write_response_t response = {bytes_written: 0};

	off64_t   curr_offset;
	void *    buffer = NULL;
	void *    write_buffer;
	int       bytes_written;
	int       rc = 0;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Write %d bytes to fd %d at offset %"PRIu64".\n", write_msg->length, write_msg->fd, write_msg->offset);

	/*
	 * The memory buffer must be aligned if the file was opened with
	 * O_DIRECT.
	 */
	if (fcntl(write_msg->fd, F_GETFL, 0) & O_DIRECT) {
		rc = alloc_aligned(write_msg->fd, write_msg->length, 0,
				   &buffer, &write_buffer);

		if (rc == 0) {
			memcpy(write_buffer, write_msg->data, write_msg->length);

		} else {
                        rc = ENOMEM;
			LOG_CRITICAL("Unable to allocate an aligned write buffer.\n");
		}

	} else {
		write_buffer = write_msg->data;
	}

	if (rc == 0) {
		/*
		 * We don't have to align the file offset and length if opened
		 * with O_DIRECT.  The caller should have done that.  If it's
		 * not aligned, the caller will get the error.
		 */
		curr_offset = lseek64((int) write_msg->fd, (off64_t) write_msg->offset, SEEK_SET);

		if (curr_offset == write_msg->offset) {
			bytes_written = write((int) write_msg->fd, write_buffer, (size_t) write_msg->length);

			if (bytes_written > 0) {
				response.bytes_written = bytes_written;

			} else {
				rc = errno;
			}

		} else {
			rc = ESPIPE; /* Illegal seek */
		}
	}

	response.status = rc;

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->size = sizeof(msg_write_response_t);
	response_msg->msg  = &response;

	SEND_MSG(response_msg);

	/* engine_free() handles NULL pointers. */
	engine_free(buffer);
	free_msg(response_msg);
	LOG_DEBUG("rc: %d bytes written: %d\n", response.status, response.bytes_written);
	LOG_PROC_EXIT_INT(response.status);
}


static void msg_ioctl(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_ioctl_t * ioctl_msg = msg->msg;

	/* The incoming message buffer is used for the outgoing response. */

	int status;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Send ioctl %"PRIu64" to fd %d.  Arg length is %"PRIu64".\n", ioctl_msg->cmd, ioctl_msg->fd, _IOC_SIZE(ioctl_msg->cmd));

	response_msg = get_msg(msg);
	status = ioctl(ioctl_msg->fd, ioctl_msg->cmd, (_IOC_SIZE(ioctl_msg->cmd) == 0) ? NULL : ioctl_msg->arg);

	if (status == 0) {
		ioctl_msg->status = status;
	} else {
		ioctl_msg->status = errno;
	}

	response_msg->cmd |= COMMAND_RESPONSE;

	SEND_MSG(response_msg);
	free_msg(response_msg);
	LOG_PROC_EXIT_INT(ioctl_msg->status);
}


static void msg_close(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_close_t * close_msg = msg->msg;
	msg_close_response_t response;

	int status;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Close file descriptor %d.\n", close_msg->fd);

	status = close(close_msg->fd);

	if (status != -1) {
		response.status = 0;

	} else {
		response.status = errno;
	}

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_close_response_t);

	SEND_MSG(response_msg);
	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


static void msg_unlink(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_unlink_t * unlink_msg = msg->msg;
	msg_unlink_response_t response;

	int status;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Unlink %s.\n", unlink_msg->file_name);


	status = unlink(unlink_msg->file_name);

	if (status == 0) {
		response.status = status;
	} else {
		response.status = errno;
	}

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


static void msg_is_mounted(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_is_mounted_t * is_mounted_msg = msg->msg;
	msg_is_mounted_response_t static_response = {0};
	msg_is_mounted_response_t * response = &static_response;
	char * mount_name;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Check if %s is mounted.\n", is_mounted_msg->volume_name);

	static_response.mounted = is_mounted(is_mounted_msg->volume_name, &mount_name);

	if (static_response.mounted) {
		response = engine_alloc(sizeof(msg_is_mounted_response_t) + strlen(mount_name) + 1);

		if (response != NULL) {
			memcpy(response, &static_response, sizeof(msg_is_mounted_response_t));
			strcpy(response->mount_name, mount_name);
		} else {
			static_response.status = ENOMEM;
			response = &static_response;
		}
	}

        response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = response;
	response_msg->size = sizeof(msg_is_mounted_response_t);
	if ((response->status == 0) && response->mounted) {
		 response_msg->size += strlen(mount_name) + 1;
	}
	
	SEND_MSG(response_msg);

	if ((response->status == 0) && response->mounted) {
		 engine_free(response);
	}

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(static_response.status);
}


static void msg_check_open_status(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_check_open_status_t * check_open_status_msg = msg->msg;
	msg_check_open_status_response_t response;
        int fd;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Check open status of %s.\n", check_open_status_msg->volume_name);

        response.status = 0;

	fd = open(check_open_status_msg->volume_name, O_EXCL | O_RDWR);

	if (fd >= 0) {
		response.opened = FALSE;
		close(fd);

	} else {
		response.opened = TRUE;
	}

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_check_open_status_response_t);
	
	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(response.status);
}


/*
 * local_copy()'s parameters are in bytes.
 */
static int local_copy(int             src_fd,
                      u_int64_t       src_offset,
                      int             trg_fd,
                      u_int64_t       trg_offset,
                      u_int64_t       len,
                      u_int64_t       chunk_size,
                      void          * context,
                      copy_callback_t callback) {

	int rc = 0;
	void * copy_buffer;
	u_int64_t bytes_written = 0;
	u_int64_t chunks_copied = 0;
	off64_t curr_offset;

	struct timezone tz;
	struct timeval copy_start_time;
	struct timeval copy_stop_time;
	struct timeval copy_time;
	int copy_hours;
	int copy_minutes;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Source offset: %"PRIu64" bytes (%"PRIu64" sectors).\n", src_offset, src_offset >> EVMS_VSECTOR_SIZE_SHIFT);
	LOG_DEBUG("Target offset: %"PRIu64" bytes (%"PRIu64" sectors).\n", trg_offset, trg_offset >> EVMS_VSECTOR_SIZE_SHIFT);
	LOG_DEBUG("Length:        %"PRIu64" bytes (%"PRIu64" sectors).\n", len, len >> EVMS_VSECTOR_SIZE_SHIFT);
	LOG_DEBUG("Chunk size:    %"PRIu64" bytes (%"PRIu64" sectors).\n", chunk_size, chunk_size >> EVMS_VSECTOR_SIZE_SHIFT);

	copy_buffer = engine_alloc(chunk_size);

	gettimeofday(&copy_start_time, &tz);

	curr_offset = lseek64(src_fd, src_offset, SEEK_SET);
	if (curr_offset != src_offset) {
		rc = errno;
		LOG_SERIOUS("lseek64() to set the source offset to %"PRIu64" returned error code %d: %s\n", src_offset, rc, strerror(rc));

		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	curr_offset = lseek64(trg_fd, trg_offset, SEEK_SET);
	if (curr_offset != trg_offset) {
		rc = errno;
		LOG_SERIOUS("lseek64() to set the target offset to %"PRIu64" returned error code %d: %s\n", trg_offset, rc, strerror(rc));

		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	if (copy_buffer == NULL) {
		LOG_CRITICAL("Error getting a buffer for the copy.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	while ((rc == 0) && (bytes_written < len)) {
		u_int64_t chunk_bytes_to_copy = min(len - bytes_written, chunk_size);
		u_int64_t chunk_bytes_written = 0;

		LOG_DEBUG("Chunk bytes to copy: %"PRIu64" (%"PRIu64" sectors)\n", chunk_bytes_to_copy, chunk_bytes_to_copy >> EVMS_VSECTOR_SIZE_SHIFT);
		while ((rc == 0) && (chunk_bytes_written < chunk_bytes_to_copy)) {
			int int_bytes_to_copy = (int) min(chunk_bytes_to_copy, (u_int64_t) INT_MAX);
			int int_bytes_written = 0;

			LOG_DEBUG("int bytes to copy: %d (%d sectors)\n", int_bytes_to_copy, int_bytes_to_copy >> EVMS_VSECTOR_SIZE_SHIFT);

			while ((rc == 0) && (int_bytes_written < int_bytes_to_copy)) {
				int bytes_read;

				/*
				 * Use regular read() and write() instead of
				 * engine_read_object() and engine_write_object().
				 * The engine_*() functions do checks for remote objects
				 * and seeks to the offset to read/write.  We know this
				 * copy is local.  We don't need to seek every time we
				 * read/write.
				 */
				bytes_read = read(src_fd, copy_buffer, int_bytes_to_copy - int_bytes_written);

				if (bytes_read > 0) {
					void * write_buffer = copy_buffer;
					int total_bytes_written = 0;

					while ((rc == 0) && (total_bytes_written < bytes_read)) {
						int bytes_written;

						bytes_written = write(trg_fd, write_buffer, bytes_read - total_bytes_written);

						if (bytes_written != -1) {
							total_bytes_written += bytes_written;
							write_buffer += bytes_written;

						} else {
							rc = errno;
							LOG_SERIOUS("write() returned error code %d: %s\n", rc, strerror(rc));
						}
					}

					if (rc == 0) {
						int_bytes_written += total_bytes_written;
					}

				} else {
					if (bytes_read == 0) {
						rc = ENODATA;
						LOG_SERIOUS("read() returned 0 bytes read.\n");
					} else {
						rc = errno;
						LOG_SERIOUS("read() returned error code %d: %s\n", rc, strerror(rc));
					}
				}
			}

			chunk_bytes_written += int_bytes_written;
		}

		bytes_written += chunk_bytes_written;

		if (rc == 0) {
			chunks_copied++;

			LOG_DEBUG("Chunks copied: %"PRIu64"\n", chunks_copied);

			if ((rc == 0) && (callback != NULL)) {
				callback(chunks_copied, context);
			}
		}
	}

	gettimeofday(&copy_stop_time, &tz);

	copy_time.tv_sec = copy_stop_time.tv_sec - copy_start_time.tv_sec;
	copy_time.tv_usec = copy_stop_time.tv_usec - copy_start_time.tv_usec;
	if (copy_time.tv_usec < 0) {
		copy_time.tv_sec --;
		copy_time.tv_usec += 1000000;
	}

	copy_hours = copy_time.tv_sec / 3600;
	copy_time.tv_sec %= 3600;
	copy_minutes = copy_time.tv_sec / 60;
	copy_time.tv_sec %= 60;

	LOG_DEFAULT("Copy time: %02d:%02d:%02ld.%06ld\n", copy_hours, copy_minutes, copy_time.tv_sec, copy_time.tv_usec);

	LOG_DEBUG("Total bytes written: %"PRIu64" (%"PRIu64" sectors)\n", bytes_written, bytes_written >> EVMS_VSECTOR_SIZE_SHIFT);

	engine_free(copy_buffer);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void copy_callback(u_int64_t chunks_copied,
			  void    * context) {

	ece_msg_t * response_msg = (ece_msg_t *) context;
        msg_copy_response_t * response = (msg_copy_response_t *) response_msg->msg;

	response->chunks_copied = chunks_copied;

	SEND_MSG(response_msg);
}


static void msg_copy(const ece_msg_t * msg) {
	ece_msg_t * response_msg;
	msg_copy_t * copy_msg = msg->msg;
	msg_copy_response_t response;
	int rc;

	LOG_PROC_ENTRY();

	response.context = copy_msg->context;
	response.chunks_copied = 0;
	response.status = 0;

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg = &response;
	response_msg->size = sizeof(msg_copy_response_t);

	/*
	 * Use our own callback so that we can send a response back.
	 * Use our response_msg for the context so that we can
	 * refer to it on our callback.
	 */
	rc = local_copy(copy_msg->src_fd,
			copy_msg->src_offset << EVMS_VSECTOR_SIZE_SHIFT,
			copy_msg->trg_fd,
			copy_msg->trg_offset << EVMS_VSECTOR_SIZE_SHIFT,
			copy_msg->len        << EVMS_VSECTOR_SIZE_SHIFT,
			copy_msg->chunk_size << EVMS_VSECTOR_SIZE_SHIFT,
			response_msg,
			copy_callback);

	if (rc != 0) {
		/*
		 * The copy failed.  Send a response with the bad status code.
		 */
		response.status = rc;

		response_msg = get_msg(msg);
		response_msg->cmd |= COMMAND_RESPONSE;
		response_msg->msg = &response;
		response_msg->size = sizeof(msg_copy_response_t);

		SEND_MSG(response_msg);
	}

	/*
	 * There is no final response from this function if the copy succeeds.
	 * The last callback message will have a zero status code and
	 * chunks_complete == chunks to copy.
	 */

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(rc);
}


static void msg_dm_cmd(const ece_msg_t * msg) {
	ece_msg_t             * response_msg;
	msg_dm_cmd_t          * dm_cmd_msg = msg->msg;
	msg_dm_cmd_response_t * response   = msg->msg;
	void                  * dmi = (void *) dm_cmd_msg->ioctl_data;
	int status;

	LOG_PROC_ENTRY();

	status = dm_run_command(dmi, dm_cmd_msg->cmd);

	response->status = status;
	response->size = dm_cmd_msg->size;

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg  = response;
	response_msg->size = sizeof(msg_dm_cmd_response_t) + dm_cmd_msg->size;

	SEND_MSG(response_msg);

	free_msg(response_msg);
	LOG_PROC_EXIT_INT(status);
}


static void msg_execvp(const ece_msg_t * msg) {
	ece_msg_t           * response_msg;
	msg_execvp_t        * execvp_msg = msg->msg;
	msg_execvp_response_t response = {0, 0};
	ece_msg_t             io_msg;
	int rc;
	int status;
	int i;
	char * * argv;
	char * str;
	int    stdin_pipe[2];
	int    stdout_pipe[2];
	int    stderr_pipe[2];
	pid_t  pid;
	int    bytes_read;
	void * buffer;

	LOG_PROC_ENTRY();

	response_msg = get_msg(msg);
	response_msg->cmd |= COMMAND_RESPONSE;
	response_msg->msg  = &response;
	response_msg->size = sizeof(msg_execvp_response_t);

	status = pipe(stdin_pipe);
	if (status) {
		LOG_SERIOUS("Error creating a pipe for stdin.  Error code %d: %s\n", errno, strerror(errno));
		response.rc = errno;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		free_msg(response_msg);
		LOG_PROC_EXIT_INT(status);
		return;
	}
	LOG_DEBUG("stdin pipe descriptors: read: %d  write: %d\n", stdin_pipe[0], stdin_pipe[1]);

	status = pipe(stdout_pipe);
	if (status) {
		LOG_SERIOUS("Error creating a pipe for stdout.  Error code %d: %s\n", errno, strerror(errno));
		response.rc = errno;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		free_msg(response_msg);
		LOG_PROC_EXIT_INT(errno);
		return;
	}
	LOG_DEBUG("stdout pipe descriptors: read: %d  write: %d\n", stdout_pipe[0], stdout_pipe[1]);

	status = pipe(stderr_pipe);
	if (status) {
		LOG_SERIOUS("Error creating a pipe for stderr.  Error code %d: %s\n", errno, strerror(errno));
		response.rc = errno;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		free_msg(response_msg);
		LOG_PROC_EXIT_INT(errno);
		return;
	}
	LOG_DEBUG("stderr pipe descriptors: read: %d  write: %d\n", stderr_pipe[0], stderr_pipe[1]);

	/*
	 * Count the number of strings so we know how many extries to allocate
	 * for argv[].
	 */
        i = 0;
	str = execvp_msg->flat_argv;
	while (*str != END_OF_TEXT) {
		str += strlen(str) + 1;
		i++;
	}
	LOG_DEBUG("Number of args: %d\n", i);

	/* Must have at least a program name in argv[0]*/
	if (i == 0) {
		LOG_WARNING("No arguments provided.  Not even a program name.\n");
		response.rc = EINVAL;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		free_msg(response_msg);
		LOG_PROC_EXIT_INT(EINVAL);
		return;
	}

	argv = engine_alloc(sizeof(char *) * (i + 1));
	if (argv == NULL) {
		LOG_CRITICAL("Could not get memory to build an argv.\n");
		response.rc = ENOMEM;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		free_msg(response_msg);
		LOG_PROC_EXIT_INT(ENOMEM);
		return;
	}

	/* Build the argv array. */
        i = 0;
	str = execvp_msg->flat_argv;
	while (*str != END_OF_TEXT) {
		argv[i] = str;
		LOG_DEBUG("Arg %d: %s\n", i, str);
		str += strlen(str) + 1;
		i++;
	}
	/* Add last NULL entry needed by execvp(). */
	argv[i] = NULL;

	buffer = engine_alloc(MAX_USER_MESSAGE_LEN);
	if (buffer == NULL) {
		LOG_CRITICAL("Error allocating memory for a read buffer.\n");
		response.rc = ENOMEM;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		engine_free(argv);
		free_msg(response_msg);
		LOG_PROC_EXIT_INT(ENOMEM);
		return;
	}

	/* Fork and execute the program. */
	switch (pid = fork()) {

	/* error */
	case -1:
		LOG_SERIOUS("Error %d (%s) forking the process to launch %s.\n", errno, evms_strerror(errno), argv[0]);
		response.rc = errno;
		LOG_DEBUG("Error response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);

		engine_free(argv);
		free_msg(response_msg);
		LOG_PROC_EXIT_INT(errno);
		return;

	/* child */
	case 0:
		dup2(stdin_pipe[0],0);	// Replace stdin
		LOG_DEBUG("Child: Close stdin pipe write descriptor %d.\n", stdin_pipe[1]);
		close(stdin_pipe[1]);	// Don't need this here

		dup2(stdout_pipe[1],1);	// Replace stdout
		LOG_DEBUG("Child: Close stdout pipe read descriptor %d.\n", stdout_pipe[0]);
		close(stdout_pipe[0]);	// Don't need this here

		dup2(stderr_pipe[1],2);	// Replace stderr
		LOG_DEBUG("Child: Close stderr pipe read descriptor %d.\n", stderr_pipe[0]);
		close(stderr_pipe[0]);	// Don't need this here

		LOG_DEBUG("Child: Run %s...\n", argv[0]);
		rc = execvp(argv[0], argv);
		LOG_DEBUG("Child: execvp() returned error code %d: %s\n", rc, evms_strerror(rc));
		LOG_DEBUG("Child: errno is %d: %s\n", errno, evms_strerror(errno));

		/* using exit() can hang GUI, use _exit */
		_exit(errno);

	/* parent */
	default:
		/*
		 * WARNING: Do Not close the read handle of stdin or you will
		 * cause a SIGPIPE if you write after the child process has gone
		 * away.
		 */
		LOG_DEBUG("Parent: Close stdout pipe write descriptor %d.\n", stdout_pipe[1]);
		close(stdout_pipe[1]);

                memcpy(&io_msg.node, &msg->node, sizeof(ece_nodeid_t));
		io_msg.corrolator = msg->corrolator;

		/* Set the output read fds to be non-blocking. */
		fcntl(stdout_pipe[0], F_SETFL, fcntl(stdout_pipe[0], F_GETFL,0) | O_NONBLOCK);
		fcntl(stderr_pipe[0], F_SETFL, fcntl(stderr_pipe[0], F_GETFL,0) | O_NONBLOCK);
		
		LOG_DEBUG("Initial response: rc: %d status: %#x\n", response.rc, response.status);
		SEND_MSG(response_msg);
		
		while (!(pid = waitpid(pid, &response.status, WNOHANG))) {
			LOG_DEBUG("Loop.  Current status: %#x\n", response.status);
			bytes_read = read(stderr_pipe[0], buffer, MAX_USER_MESSAGE_LEN);
			if (bytes_read > 0) {
				io_msg.cmd = MSG_EXECVP_STDERR_IO;
				io_msg.size = bytes_read;
				io_msg.msg = buffer;
				LOG_DEBUG("Parent: Read %d bytes from stderr pipe fd %d.\n", bytes_read, stderr_pipe[0]);
				
				SEND_MSG(&io_msg);
			}

			bytes_read = read(stdout_pipe[0], buffer, MAX_USER_MESSAGE_LEN);
			if (bytes_read > 0) {
				io_msg.cmd = MSG_EXECVP_STDOUT_IO;
				io_msg.size = bytes_read;
				io_msg.msg = buffer;
				LOG_DEBUG("Parent: Read %d bytes from stdout pipe fd %d.\n", bytes_read, stdout_pipe[0]);
				
				SEND_MSG(&io_msg);
			}

			/* Don't hog the CPU */
			usleep(10000);
		}

		LOG_DEBUG("Final status: %#x\n", response.status);
		if (WIFEXITED(status)) {
			int exit_status = WEXITSTATUS(response.status);

			LOG_DEBUG("Parent: %s exited with error code %d: %s\n", argv[0], exit_status, evms_strerror(exit_status));

			/* Drain the output pipes. */
			do {
				bytes_read = read(stderr_pipe[0],buffer,MAX_USER_MESSAGE_LEN);
				if (bytes_read > 0) {
					io_msg.cmd = MSG_EXECVP_STDERR_IO;
					io_msg.size = bytes_read;
					io_msg.msg = buffer;
					LOG_DEBUG("Parent: Read %d bytes from stderr pipe fd %d.\n", stderr_pipe[0], bytes_read);

					SEND_MSG(&io_msg);
				}
			} while (bytes_read > 0);

			do {
				bytes_read = read(stdout_pipe[0],buffer,MAX_USER_MESSAGE_LEN);
				if (bytes_read > 0) {
					io_msg.cmd = MSG_EXECVP_STDOUT_IO;
					io_msg.size = bytes_read;
					io_msg.msg = buffer;
					LOG_DEBUG("Parent: Read %d bytes from stdout pipe fd %d.\n", stdout_pipe[0], bytes_read);

					SEND_MSG(&io_msg);
				}
			} while (bytes_read > 0);

		} else {
			int signal = WTERMSIG(status);

			LOG_WARNING("Parent: %s was interrupted by signal %d: %s\n", argv[0], signal, sys_siglist[signal]);
		}
	}

	/* Close out this process's access to the pipes. */
	close(stdin_pipe[0]);
	close(stdin_pipe[1]);
	close(stdout_pipe[0]);
	close(stdout_pipe[1]);
	close(stderr_pipe[0]);
	close(stderr_pipe[1]);

	LOG_DEBUG("Final response: rc: %d status: %#x\n", response.rc, response.status);
	SEND_MSG(response_msg);

	engine_free(buffer);
	engine_free(argv);
	free_msg(response_msg);
	LOG_PROC_EXIT_INT(status);
}



void daemon_router(ece_msg_t * msg) {

	u_int32_t cmd = msg->cmd;

	LOG_PROC_ENTRY();

	LOG_DEBUG("%s from node %s\n", msg_name(cmd), nodeid_to_string(&msg->node));

	if (cmd & COMMAND_RESPONSE) {
		/* Ignore responses. */

	} else {
		switch (msg->cmd) {
		case MSG_GET_VERSION:
			msg_get_version(msg);
			break;

		case MSG_LOCK_ENGINE:
			msg_lock_engine(msg);
			break;

		case MSG_UNLOCK_ENGINE:
			msg_unlock_engine(msg);
			break;

		case MSG_CHKNOD:
			msg_chknod(msg);
			break;

		case MSG_MKNOD:
			msg_mknod(msg);
			break;

		case MSG_DISCOVER_DISKS:
			msg_discover_disks(msg);
			break;

		case MSG_GET_DEV_NODE_NAMES:
			msg_get_dev_node_names(msg);
			break;

		case MSG_OPEN:
			msg_open(msg);
			break;

		case MSG_READ:
			msg_read(msg);
			break;

		case MSG_WRITE:
			msg_write(msg);
			break;

		case MSG_IOCTL:
			msg_ioctl(msg);
			break;

		case MSG_CLOSE:
			msg_close(msg);
			break;

		case MSG_UNLINK:
			msg_unlink(msg);
			break;

		case MSG_IS_MOUNTED:
			msg_is_mounted(msg);
			break;

		case MSG_CHECK_OPEN_STATUS:
			msg_check_open_status(msg);
			break;

		case MSG_COPY:
			msg_copy(msg);
			break;

		case MSG_DM_CMD:
			msg_dm_cmd(msg);
			break;

		case MSG_EXECVP:
			msg_execvp(msg);
			break;

		case MSG_EXECVP_STDIN_IO:
			msg_execvp_stdin_io(msg);
			break;			    	

		case MSG_EXECVP_STDOUT_IO:
			msg_execvp_stdout_io(msg);
			break;

		case MSG_EXECVP_STDERR_IO:
			msg_execvp_stderr_io(msg);
			break;

		default:
			LOG_DEBUG("%d is not a valid message command.\n", msg->cmd);
			((ece_msg_t *)msg)->cmd = MSG_INVALID_CMD | COMMAND_RESPONSE;
			SEND_MSG(msg);
			break;
		}
	}

	LOG_PROC_EXIT_VOID();
}

