/*
 *
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: engine.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include <wait.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "fullengine.h"
#include "engine.h"
#include "dload.h"
#include "faulthdlr.h"
#include "handlemgr.h"
#include "common.h"
#include "internalAPI.h"
#include "memman.h"
#include "commit.h"
#include "discover.h"
#include "message.h"
#include "volume.h"
#include "config.h"
#include "dm.h"
#include "cluster.h"
#include "remote.h"
#include "daemon.h"
#include "dmonmsg.h"


/*----------------------------------------------------------------------------+
+                                                                             +
+                                  GLOBAL DATA                                +
+                                                                             +
+-----------------------------------------------------------------------------*/

evms_version_t engine_version =
{
	MAJOR_VERSION,
	MINOR_VERSION,
	PATCH_LEVEL
};

evms_version_t engine_api_version =
{
	ENGINE_API_MAJOR_VERSION,
	ENGINE_API_MINOR_VERSION,
	ENGINE_API_PATCH_LEVEL
};

evms_version_t engine_services_api_version =
{
	ENGINE_SERVICES_API_MAJOR_VERION,
	ENGINE_SERVICES_API_MINOR_VERION,
	ENGINE_SERVICES_API_PATCH_LEVEL
};

evms_version_t engine_plugin_api_version =
{
	ENGINE_PLUGIN_API_MAJOR_VERION,
	ENGINE_PLUGIN_API_MINOR_VERION,
	ENGINE_PLUGIN_API_PATCH_LEVEL
};

evms_version_t engine_fsim_api_version =
{
	ENGINE_FSIM_API_MAJOR_VERION,
	ENGINE_FSIM_API_MINOR_VERION,
	ENGINE_FSIM_API_PATCH_LEVEL
};

evms_version_t engine_container_api_version =
{
	ENGINE_CONTAINER_API_MAJOR_VERION,
	ENGINE_CONTAINER_API_MINOR_VERION,
	ENGINE_CONTAINER_API_PATCH_LEVEL
};

evms_version_t engine_cluster_api_version =
{
	ENGINE_CLUSTER_API_MAJOR_VERION,
	ENGINE_CLUSTER_API_MINOR_VERION,
	ENGINE_CLUSTER_API_PATCH_LEVEL
};

evms_version_t engine_daemon_msg_version =
{
	EVMS_DMONMSG_MAJOR_VERSION,
	EVMS_DMONMSG_MAJOR_VERSION,
	EVMS_DMONMSG_PATCH_LEVEL
};


ece_nodeid_t no_nodeid = ECE_NO_NODE;
ece_nodeid_t current_nodeid;
char       * current_node_name;
BOOLEAN      local_focus = TRUE;

dlist_t PluginList           = NULL;
dlist_t DiskList             = NULL;
dlist_t SegmentList          = NULL;
dlist_t ContainerList        = NULL;
dlist_t RegionList           = NULL;
dlist_t EVMSObjectList       = NULL;
dlist_t VolumeList           = NULL;
dlist_t KillSectorList       = NULL;
dlist_t VolumeDeleteList     = NULL;
dlist_t DeactivateList       = NULL;

rename_volume_t rename_volume_list = LIST_DECL(&rename_volume_list);

plugin_record_t * local_disk_manager = NULL;
plugin_record_t * replace_plugin = NULL;
plugin_record_t * cluster_manager = NULL;
cluster_functions_t * ece_funcs = NULL;

int dm_control_fd = 0;

engine_mode_t engine_mode = ENGINE_CLOSED;

debug_level_t debug_level = DEFAULT;

char  * log_file_name = DEFAULT_LOG_FILE;
int     log_file_fd = 0;
BOOLEAN log_usec = FALSE;
BOOLEAN log_pid = FALSE;
u_char log_buf[LOG_BUF_SIZE];

#define ENGINE_LOCK_FILE_DIR	"/var/lock"
#define ENGINE_LOCK_FILE	ENGINE_LOCK_FILE_DIR "/evms-engine"
int           lock_file_fd = 0;
size_t        lock_file_size = 0;
lock_file_t * lock_file;

sem_t       shutdown_sem;

/*-------------------------------------------------------------------------------------+
+                                                                                      +
+                                PRIVATE SUBROUTINES                                   +
+                                                                                      +
+-------------------------------------------------------------------------------------*/

static char * build_archive_log_name(char * log_name, int index) {
	char * archive_log_file_name;
	char * pch1;
	char * pch2;

	/*
	 * Get memory to build the archive file names.  Get enough for the name,
	 * three bytes for the ".nn" that will be inserted/appended, plus one for
	 * the null terminator.
	 */
	archive_log_file_name = engine_alloc(strlen(log_name) + 4);

	if (archive_log_file_name != NULL) {

		strcpy(archive_log_file_name, log_name);

		/*
		 * The ".nn" will be inserted ahead of the last '.' in the file name.
		 * If the file name doesn't have a '.' in it, the .nn will be
		 * appended at the end.
		 */
		pch1 = strrchr(archive_log_file_name, '.');
		if (pch1 == NULL) {
			pch1 = archive_log_file_name + strlen(archive_log_file_name);
		}

		/* Add/insert the ".nn". */
		*(pch1++) = '.';
		sprintf(pch1, "%d", index);

		/* Copy the remainder of the file name, if it exists. */
		pch2 = strrchr(log_name, '.');
		if (pch2 != NULL) {
			strcat(pch1, pch2);
		}
	}

	return archive_log_file_name;
}


#define MAX_LOG_ARCHIVES 9

static int start_logging(char * filename) {
	int rc = 0;
	int i;
	char * old_archive_log_name;
	char * new_archive_log_name;

	if (log_file_fd == 0) {

		/* Roll back the archive log files. */
		old_archive_log_name = build_archive_log_name(filename, MAX_LOG_ARCHIVES);
		unlink(old_archive_log_name);

		for (i = MAX_LOG_ARCHIVES; i > 1; i--) {
			new_archive_log_name = build_archive_log_name(filename, i - 1);

			rename(new_archive_log_name, old_archive_log_name);

			engine_free(old_archive_log_name);
			old_archive_log_name = new_archive_log_name;
		}

		rename(filename, old_archive_log_name);
		engine_free(old_archive_log_name);

		log_file_fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY, 0664);

		if (log_file_fd < 0) {
			rc = errno;
			log_file_fd = 0;
		}

	} else {
		rc = EEXIST;
	}

	return rc;
}


static int stop_logging() {
	int rc = 0;

	if (log_file_fd > 0) {
		close(log_file_fd);
		log_file_fd = 0;
	} else {
		rc = ENOENT;
	}

	return rc;
}


/*
 * Put a time stamp in the buffer.
 * This function assumes the buffer has enough room for the timestamp.
 */
void timestamp(char * buf, size_t len) {
	time_t t;
	size_t stamp_len;

	time(&t);

	strftime(buf, len, "%b %d %H:%M:%S", localtime(&t));

	if (log_usec) {
		struct timezone tz;
		struct timeval  utime;

		gettimeofday(&utime, &tz);
		sprintf(buf + strlen(buf), ".%06ld ", utime.tv_usec);
	} else {
		strcat(buf, " ");
	}

	stamp_len = strlen(buf);

	gethostname(buf + stamp_len, len - stamp_len);

	if (log_pid) {
		sprintf(log_buf + strlen(log_buf), " %d ", getpid());

	} else {
		strcat(buf, " ");
	}
}


int engine_write_log_entry(debug_level_t level,
			   char        * fmt,
			   ...) {
	int rc = 0;
	int len;
	va_list args;

	if (dm_device_suspended) {
		return rc;
	}

	if (level <= debug_level) {
		if (log_file_fd > 0) {
			timestamp(log_buf, LOG_BUF_SIZE);

			if (engine_mode == ENGINE_DAEMON) {
				strcat(log_buf, "Daemon: ");
			} else {
				strcat(log_buf, "Engine: ");
			}
			len = strlen(log_buf);

			va_start(args, fmt);
			len += vsprintf(log_buf + strlen(log_buf), fmt, args);
			va_end(args);

			if (write(log_file_fd, log_buf, len) < 0) {
				rc = errno;
			}

		} else {
			rc = ENOENT;
		}
	}

	return rc;
}


/*
 * Scan a line for device info.
 * This is a simple, job specific function.  Rather than try to use sscanf() or
 * other complex parsers, we do a quick and dirty job here.
 * The input buffer should have the form:
 * | +[0-9]+ +name *[\n]|
 * In English, start with one or more blanks, then a number, one or more blanks,
 * then a name, followed by blanks and/or a newline.
 * Return the number and the name via parameters.
 * The function returns TRUE if it successfully found a number and a name,
 * else FALSE.
 */
static BOOLEAN scan_dev_info(char * buf, int * dev_num, char * dev_name) {

	int result = FALSE;
	char * p = buf;

	/* Skip over white space. */
	while ((*p != '\0') &&
	       ((*p == ' ') || (*p == '\t') || (*p == '\n'))) {
		p++;
	}

	if ((*p != '\0') &&
	    (*p >= '0') && (*p <= '9')) {
		*dev_num = atoi(p);

		/* Skip over the number. */
		while ((*p != '\0') &&
		       (*p != ' ') && (*p != '\t') && (*p != '\n')) {
			p++;
		}

		/* Skip over white space. */
		while ((*p != '\0') &&
		       ((*p == ' ') || (*p == '\t') || (*p == '\n'))) {
			p++;
		}

		if (*p != '\0') {
			/* Copy string up to white space. */
			while ((*p != '\0') &&
			       (*p != ' ') && (*p != '\t') && (*p != '\n')) {
				*dev_name++ = *p++;
			}

			result = TRUE;
		}
	}

	/*
	 * Terminate the name. If a name was not found, this will set the
	 * returned device name to "".
	 */
	*dev_name = '\0';

	return result;
}


static int load_dm_module() {

	int rc = 0;
	pid_t pid;
	int   status;
	char * argv[3] = {"modprobe", "dm-mod", NULL};

	LOG_PROC_ENTRY();

        pid = fork();

	switch (pid) {
	case -1:
		rc = errno;
		LOG_WARNING("fork() to run \"%s %s\" returned error %d: %s\n", argv[0], argv[1], rc, strerror(rc));
		break;

	case 0:
		/* Child process */

		execvp(argv[0], argv);

		/* Should not get here unless execvp() fails. */
		rc = errno;
		LOG_WARNING("execvp() to run \"%s %s\" returned error %d: %s\n", argv[0], argv[1], rc, strerror(rc));

		/* exit() kills the GUI.  Use _exit() instead. */
		_exit(rc);
		break;

	default:
		/* Parent process */
		waitpid(pid, &status, 0);
	}

	if (rc == 0) {
		if (WIFSIGNALED(status)) {
			LOG_WARNING("\"%s %s\" was terminated by signal %s\n", argv[0], argv[1], sys_siglist[WTERMSIG(status)]);
			rc = EINTR;

		} else {
			rc = WEXITSTATUS(status);
			LOG_DEBUG("\"%s %s\" exited with error code %d: %s\n", argv[0], argv[1], rc, strerror(rc));
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Setup a dev node for the Device Mapper control node.
 * The Device Mapper control node is a character device.
 * Its major number is the major number of "misc", found in /proc/devices.
 * Its Minor number is listed in /proc/misc as "device"mapper".
 */
static int open_dm_control_node(void) {

	int rc = 0;
	FILE * file;
	char work_buf[256];
	char dev_name[64];
	int dev_num;
	int dm_control_major = 0;
	int dm_control_minor = 0;
	BOOLEAN found_major = FALSE;
	BOOLEAN found_minor = FALSE;

	LOG_PROC_ENTRY();

	file = fopen("/proc/devices", "r");

	if (file != NULL) {
		while (fgets(work_buf,sizeof(work_buf),file) != NULL) {
			if (scan_dev_info(work_buf, &dev_num, dev_name)) {
				if (strcmp(dev_name, "misc") == 0) {
					dm_control_major = dev_num;
					found_major = TRUE;
					break;
				}
			}
		}

		fclose(file);

		if (found_major) {
			BOOLEAN retry = TRUE;

retry:
			file = fopen("/proc/misc", "r");

			if (file != NULL) {
				while (fgets(work_buf,sizeof(work_buf),file) != NULL) {
					if (scan_dev_info(work_buf, &dev_num, dev_name)) {
						if (strcmp(dev_name, "device-mapper") == 0) {
							dm_control_minor = dev_num;
							found_minor = TRUE;
							break;
						}
					}
				}

				fclose(file);

				if (found_minor) {
					dev_t devt = makedev(dm_control_major, dm_control_minor);

					make_directory(EVMS_DM_CONTROL_DIR, NO_MESSAGES);
					unlink(EVMS_DM_CONTROL);

					rc = mknod(EVMS_DM_CONTROL, (S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP), devt);

					if (rc == 0) {
						rc = open(EVMS_DM_CONTROL, O_RDWR);

						if (rc > 0) {
							dm_control_fd = rc;

							rc = dm_check_version();
							if (rc != 0) {
								LOG_WARNING("Checking Device-Mapper interface version failed with error code %d: %s\n", rc, strerror(rc));
								close(dm_control_fd);
								dm_control_fd = 0;
							}

						} else {
							rc = errno;
							LOG_WARNING("Open of " EVMS_DM_CONTROL " failed with error code %d: %s\n", rc, strerror(rc));
						}

					} else {
						rc = errno;
						LOG_WARNING("mknod of " EVMS_DM_CONTROL " (%d:%d) failed with error code %d: %s\n", dm_control_major, dm_control_minor, rc, strerror(rc));
					}

				} else {
					if (retry) {
						rc = load_dm_module();

						/* Whether success or error, don't try again. */
						retry = FALSE;

						if (rc == 0) {
							goto retry;
						}

					} else {
						LOG_WARNING("Could not find an entry for \"device-mapper\" in /proc/misc.\n");
						rc = ENOENT;
					}
				}

			} else {
				rc = errno;
				LOG_WARNING("Open of /proc/misc failed with error code %d: %s\n", rc, strerror(rc));
			}

		} else {
			LOG_WARNING("Could not find an entry for \"misc\" in /proc/devices.\n");
			rc = ENOENT;
		}

	} else {
		rc = errno;
		LOG_WARNING("Open of /proc/devices failed with error code %d: %s\n", rc, strerror(rc));
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int send_shutdown(pid_t pid) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * If the locking node is the local machine, send the shutdown
	 * signal locally, else send it to the remote node.
	 */
	if (lock_file->node[0] == '\0') {
		kill(pid, SIGUSR1);

		/*
		 * Wait for the process to die by sending it a
		 * harmless signal.  When the kill() fails we
		 * can know that the process has gone away.
		 */
		do {
			sleep(1);
		} while (kill(pid, SIGCONT) == 0);

	} else {
		/*
		 * Loading the plug-ins will result in an attempt to
		 * lock the Engine in critical mode on the remote
		 * nodes.  The lock critical will shutdown any
		 * Engine that is not running in critical mode.
		 */
		rc = load_plugins();
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static char lock_msg[MAX_LOCK_MESSAGE_LEN];

int lock_engine(engine_mode_t mode, const char * * msg) {

	int     rc = 0;
	int     status;
	struct  flock lockinfo = {0};
	BOOLEAN retry = FALSE;
	char  * buf1 = NULL;
	char  * buf2;
	struct stat statbuf = {0};

	LOG_PROC_ENTRY();

	lock_msg[0] = '\0';

	status = stat(ENGINE_LOCK_FILE, &statbuf);
	
	/*
	 * Get memory for the lock file contents.
	 * If the lock file doesn't exist, the stat() above failed and
	 * statbuf.st_size will be left at its initial value of 0.
	 */
	if (lock_file == NULL) {
		lock_file_size = max(statbuf.st_size, sizeof(lock_file_t));
		lock_file = engine_alloc(lock_file_size);
		if (lock_file == NULL) {
			LOG_CRITICAL("Error getting memory for the lock file.\n");
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

	} else {
		/* Make sure the memory for the lock file is big enough. */
		if (lock_file_size < statbuf.st_size) {
			lock_file = engine_realloc(lock_file, statbuf.st_size);
			if (lock_file != NULL) {
				lock_file_size = statbuf.st_size;

			} else {
				LOG_CRITICAL("Error reallocating memory for the lock file.\n");
				LOG_PROC_EXIT_INT(ENOMEM);
				return ENOMEM;
			}
		}
	}

	if (status != 0) {

		/*
		 * The lock file does not exist.  Make sure the lock file
		 * directory exists so that we can create the lock file.
		 */
		char dir_name[] = ENGINE_LOCK_FILE_DIR;

		rc = make_directory(dir_name, NO_MESSAGES);

		if (rc != 0) {
			LOG_SERIOUS("Unable to create the directory for the lock file.\n");
		}
	}

start:
	if (rc == 0) {
		lock_file_fd = open(ENGINE_LOCK_FILE, O_CREAT | O_RDWR, 0660);
		if (lock_file_fd < 0) {
			rc = errno;
		}
	}

	if (rc == 0) {

		lseek(lock_file_fd, offsetof(lock_file_t, pid), SEEK_SET);
		read(lock_file_fd, &lock_file->pid, lock_file_size - offsetof(lock_file_t, pid));

		/*
		 * Lock our access to the Engine.  Figure out whether to lock for
		 * reading or for writing based on the mode.
		 */
		if (mode == ENGINE_READONLY) {
			lockinfo.l_type = F_RDLCK;
		} else {
			lockinfo.l_type = F_WRLCK;
		}

		lockinfo.l_whence = SEEK_SET;
		if (mode == ENGINE_DAEMON) {
			lockinfo.l_start = offsetof(lock_file_t, daemon_lock);
			lockinfo.l_len = sizeof(lock_file->daemon_lock);

		} else {
			lockinfo.l_start = offsetof(lock_file_t, engine_lock);
			lockinfo.l_len = sizeof(lock_file->engine_lock);
		}

		rc = fcntl(lock_file_fd, F_SETLK, &lockinfo);

		if (rc == 0) {
			/* Put our information in the lock file. */
			if (mode == ENGINE_DAEMON) {
				lock_file->daemon_lock = 1;
				lseek(lock_file_fd, offsetof(lock_file_t, daemon_lock), SEEK_SET);
				write(lock_file_fd, &lock_file->daemon_lock, sizeof(lock_file->daemon_lock));

			} else {
				lock_file->engine_lock = 1;
				lock_file->pid = getpid();
				lock_file->mode = mode;
				/*
				 * Leave lock_file->node as an empty string.  If
				 * the Engine is being opened on behalf of
				 * another node, the daemon will fill in the
				 * node field.
				 */
				lseek(lock_file_fd, offsetof(lock_file_t, engine_lock), SEEK_SET);
				write(lock_file_fd, &lock_file->engine_lock, lock_file_size - offsetof(lock_file_t, engine_lock));
			}

		} else {
			/*
			 * We could not lock the Engine.  It must be in use by some
			 * other process.  Try to find out what process it is and tell
			 * the user.
			 */
			rc = fcntl(lock_file_fd, F_GETLK, &lockinfo);

			if (rc == 0) {
				int bytes;

				/*
				 * Check if the process is alive by sending it a harmless
				 * signal.
				 */
				rc = kill(lockinfo.l_pid, SIGCONT);
				if (rc != 0) {
					if ((errno == ESRCH) && !retry) {
						/*
						 * The locking process is gone.
						 * Try deleting the lock file and starting again.
						 */
						close(lock_file_fd);
						lock_file_fd = 0;
						rc = unlink(ENGINE_LOCK_FILE);
						if (rc == 0) {
							retry = TRUE;
							goto start;
						}
					}
				}

				if ((mode == ENGINE_READWRITE_CRITICAL) &&
				    (lock_file->mode != ENGINE_READWRITE_CRITICAL)) {
					send_shutdown(lockinfo.l_pid);
					close(lock_file_fd);
					goto start;
				}

				buf1 = engine_alloc(PATH_MAX + 1);
				if (buf1 == NULL) {
					LOG_CRITICAL("Error getting memory for buffer 1.\n");
					LOG_PROC_EXIT_INT(ENOMEM);
					return ENOMEM;
				}

				buf2 = engine_alloc(PATH_MAX + 1);
				if (buf2 == NULL) {
					LOG_CRITICAL("Error getting memory for buffer 1.\n");
					engine_free(buf1);
					LOG_PROC_EXIT_INT(ENOMEM);
					return ENOMEM;
				}
				
				/*
				 * The file name for the pid is symlinked at
				 * /proc/{pid}/exe
				 */
				sprintf(buf1, "/proc/%d/exe", lockinfo.l_pid);

				bytes = readlink(buf1, buf2, PATH_MAX + 1);

				if (bytes > 0) {
					/*
					 * readlink() does not null terminate the returned
					 * file name.
					 */
					buf2[bytes] = '\0';

					if (mode == ENGINE_DAEMON) {
						sprintf(lock_msg, "The EVMS Daemon is already running in process %d (%s).\n",
							lockinfo.l_pid, buf2);
					} else {
						sprintf(lock_msg, "The EVMS Engine is currently in use by process %d (%s).\n",
							lockinfo.l_pid, buf2);
					}

				} else {
					/*
					 * Couldn't get the name for the process ID.
					 * Log a message with just the PID.
					 */
					if (mode == ENGINE_DAEMON) {
						sprintf(lock_msg, "The EVMS Daemon is already running in process %d.\n",
							lockinfo.l_pid);
					} else {
						sprintf(lock_msg, "The EVMS Engine is currently in use by process %d.\n",
							lockinfo.l_pid);
					}
				}

				if ((mode != ENGINE_DAEMON) &&
				    (lock_file->node[0] != '\0')) {
					sprintf(lock_msg + strlen(lock_msg), "The process has locked the Engine on behalf of node %s.\n", lock_file->node);
				}

				engine_free(buf2);

			} else {
				/*
				 * Couldn't get the lockinfo of the locking process.
				 * Log a generic message.
				 */
				if (mode == ENGINE_DAEMON) {
					sprintf(lock_msg, "The EVMS Daemon is already running.\n");
				} else {
					sprintf(lock_msg, "The EVMS Engine is currently in use by another process.\n");
				}
			}

			engine_free(buf1);

			if (!((engine_mode == ENGINE_DAEMON) &&
			      (mode != ENGINE_DAEMON)) ) {
				engine_user_message(NULL, NULL, lock_msg);
			}

			close(lock_file_fd);
			lock_file_fd = 0;
			rc = EACCES;
		}

	} else {
		sprintf(lock_msg, "Unable to open the Engine lock file %s: %s.", ENGINE_LOCK_FILE, strerror(rc));
		if (rc == EROFS) {
			strcat(lock_msg, "  The Engine is not protected against other instances of the Engine being opened at the same time.\n");
			rc = 0;
		} else {
			strcat(lock_msg, "\n");
		}
		engine_user_message(NULL, NULL, lock_msg);
		lock_file_fd = 0;
	}

	if (msg != NULL) {
		*msg = lock_msg;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int unlock_engine(engine_mode_t mode) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (lock_file_fd != 0) {
		struct flock lockinfo = {0};

		/* Clear the contents of the lock file. */
		if (mode == ENGINE_DAEMON) {
			lock_file->daemon_lock = 0;
			lseek(lock_file_fd, offsetof(lock_file_t, daemon_lock), SEEK_SET);
			write(lock_file_fd, &lock_file->daemon_lock, sizeof(lock_file->daemon_lock));

		} else {
			lock_file->engine_lock = 0;
			lock_file->pid = 0;
			lock_file->mode = ENGINE_CLOSED;
			memset(lock_file->node, '\0', lock_file_size - offsetof(lock_file_t, node));
			lseek(lock_file_fd, offsetof(lock_file_t, engine_lock), SEEK_SET);
			write(lock_file_fd, &lock_file->engine_lock, lock_file_size - offsetof(lock_file_t, engine_lock));
		}

		lockinfo.l_type = F_UNLCK;
		lockinfo.l_whence = SEEK_SET;
		if (mode == ENGINE_DAEMON) {
			lockinfo.l_start = offsetof(lock_file_t, daemon_lock);
			lockinfo.l_len = sizeof(lock_file->daemon_lock);

		} else {
			lockinfo.l_start = offsetof(lock_file_t, engine_lock);
			lockinfo.l_len = sizeof(lock_file->engine_lock);
		}

		rc = fcntl(lock_file_fd, F_SETLK, &lockinfo);
		if (rc == 0) {
			close(lock_file_fd);
			lock_file_fd = 0;
		} else {
			rc = errno;
		}

		engine_free(lock_file);
		lock_file = NULL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int create_evms_dlists() {

	LOG_PROC_ENTRY();

	PluginList = CreateList();
	if (PluginList == NULL) {
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	DiskList = CreateList();
	if (DiskList == NULL) {
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	if (engine_mode != ENGINE_DAEMON) {

		SegmentList = CreateList();
		if (SegmentList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		ContainerList = CreateList();
		if (ContainerList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		RegionList = CreateList();
		if (RegionList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		EVMSObjectList = CreateList();
		if (EVMSObjectList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		VolumeList = CreateList();
		if (VolumeList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		KillSectorList = CreateList();
		if (KillSectorList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		VolumeDeleteList = CreateList();
		if (VolumeDeleteList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		DeactivateList = CreateList();
		if (DeactivateList == NULL) {
			LOG_PROC_EXIT_INT(ENOMEM);
			return ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(0);
	return 0;
}


static void free_list_items(dlist_t list) {

	int rc = 0;
	TAG tag;
	ADDRESS thing = NULL;

	rc = BlindGetObject(list,
			    &tag,
			    NULL,
			    FALSE,
			    &thing);

	while ((rc == DLIST_SUCCESS) && (thing != NULL)) {
		switch (tag) {
		case DISK_TAG:
			{
				storage_object_t * obj = (storage_object_t *) thing;

				/*
				 * Mark the object inactive so that
				 * engine_free_logical_disk() doesn't put the
				 * object on the DeactivateList.
				 */
				obj->flags &= ~SOFLAG_ACTIVE;
				engine_free_logical_disk((storage_object_t *) obj);
			}
			break;

		case SEGMENT_TAG:
			{
				storage_object_t * obj = (storage_object_t *) thing;

				/*
				 * Mark the object inactive so that
				 * engine_free_segment() doesn't put the
				 * object on the DeactivateList.
				 */
				obj->flags &= ~SOFLAG_ACTIVE;
				engine_free_segment((storage_object_t *) obj);
			}
			break;

		case REGION_TAG:
			{
				storage_object_t * obj = (storage_object_t *) thing;

				/*
				 * Mark the object inactive so that
				 * engine_free_region() doesn't put the
				 * object on the DeactivateList.
				 */
				obj->flags &= ~SOFLAG_ACTIVE;
				engine_free_region((storage_object_t *) obj);
			}
			break;

		case EVMS_OBJECT_TAG:
			{
				storage_object_t * obj = (storage_object_t *) thing;

				/*
				 * Mark the object inactive so that
				 * engine_free_evms_object() doesn't put the
				 * object on the DeactivateList.
				 */
				obj->flags &= ~SOFLAG_ACTIVE;
				engine_free_evms_object((storage_object_t *) obj);
			}
			break;

		case CONTAINER_TAG:
			{
				storage_container_t * con = (storage_container_t *) thing;

				engine_free_container(con);
			}
			break;

		case VOLUME_TAG:
			{
				logical_volume_t * vol = (logical_volume_t *) thing;

				destroy_handle(vol->app_handle);
				engine_unregister_name(vol->name);

				/*
				 * The volume may also be on the the
				 * VolumeDeleteList.  If we are processing the
				 * VolumeList, remove the volume from the
				 * VolumeDeleteList, if it is on the list.
				 */
				if (list == VolumeList) {
					DeleteObject(VolumeDeleteList, vol);
				}
			}

			/*
			 * Fall through to remove the logical_volume from
			 * the list and free its memory.
			 */

		default:
			/*
			 * Don't know what the object is.  Remove the item from the
			 * list and free its memory
			 */
			DeleteObject(list, thing);
			engine_free(thing);
			break;
		}

		/* Get another object from the list. */
		rc = BlindGetObject(list,
				    &tag,
				    NULL,
				    FALSE,
				    &thing);
	}
}


static void destroy_evms_dlists(void) {

	LOG_PROC_ENTRY();

	if (PluginList != NULL) {
		DestroyList(&PluginList, FALSE);
	}

	if (DiskList != NULL) {
		free_list_items(DiskList);
		DestroyList(&DiskList, FALSE);
	}

	if (SegmentList != NULL) {
		free_list_items(SegmentList);
		DestroyList(&SegmentList, FALSE);
	}

	if (ContainerList != NULL) {
		free_list_items(ContainerList);
		DestroyList(&ContainerList, FALSE);
	}

	if (RegionList != NULL) {
		free_list_items(RegionList);
		DestroyList(&RegionList, FALSE);
	}

	if (EVMSObjectList != NULL) {
		free_list_items(EVMSObjectList);
		DestroyList(&EVMSObjectList, FALSE);
	}

	if (VolumeList != NULL) {
		free_list_items(VolumeList);
		DestroyList(&VolumeList, FALSE);
	}

	if (KillSectorList != NULL) {
		DestroyList(&KillSectorList, FALSE);
	}

	if (VolumeDeleteList != NULL) {
		free_list_items(VolumeDeleteList);
		DestroyList(&VolumeDeleteList, FALSE);
	}

	if (DeactivateList != NULL) {
		DestroyList(&DeactivateList, FALSE);
	}

	LOG_PROC_EXIT_VOID();
}


int evms_close_engine(void) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (engine_mode != ENGINE_CLOSED) {
		close(dm_control_fd);
		dm_control_fd = 0;
		evms_free_config();
		unload_plugins();
		clear_name_registry();
		destroy_all_handles();
		destroy_evms_dlists();
		remove_signal_handlers();
		unlock_engine(engine_mode);
		engine_mode = ENGINE_CLOSED;

	} else {
		LOG_DEBUG("The Engine is already closed.\n");
		rc = EPERM;
	}

	LOG_PROC_EXIT_INT(rc);

	stop_logging();

	return rc;
}


static void * thread_engine_user_message(void * arg) {

	char * msg = (char *) arg;

	engine_user_message(NULL, NULL, msg);

	return NULL;
}


static char msg_buf[256];

static void shutdown_engine() {

	int seconds_until_death = 30;
	pthread_t msg_tid;

	LOG_PROC_ENTRY();

	while (seconds_until_death > 0) {
		if (engine_mode == ENGINE_READONLY) {
			sprintf(msg_buf,
				"Another process urgently needs the Engine.  "
				"Please quit now.  "
				"This process will self destruct in %d seconds.\n",
				seconds_until_death);

		} else {
			sprintf(msg_buf,
				"Another process urgently needs the Engine.  "
				"Please save your changes or quit now.  "
				"This process will self destruct in %d seconds.\n",
				seconds_until_death);
		}

		pthread_create(&msg_tid,
			       NULL,
			       thread_engine_user_message,
			       msg_buf);

		/* Warn every ten seconds and on the last five seconds. */

		if (seconds_until_death >= 10) {
			sleep(10);
			seconds_until_death -= 10;
		} else {
			sleep(5);
			seconds_until_death -= 5;
		}
	}

	if (seconds_until_death <= 0) {
		pthread_create(&msg_tid,
			       NULL,
			       thread_engine_user_message,
			       "Self destruct sequence initiated.\n");


	}

	/* Don't self destruct in the middle of a commit. */
	while (commit_in_progress) {
		sleep(1);
	}

	/* Close the Engine nicely if the user/UI didn't do it. */
	evms_close_engine();

	/* Try to exit nicely. */
	raise(SIGQUIT);

	sleep(3);

	/* Die, die, die. */
	raise(SIGKILL);
}


/*
 * Check if the old EVMS kernel is running on this system.
 */

static int check_for_evms_kernel(void) {

	int rc = 0;
	int status;
	struct stat statbuf;
	dev_t evms_devt = makedev(EVMS_MAJOR, 0);
	int fd;
	char temp_node_name[65];

	LOG_PROC_ENTRY();

	/*
	 * Simple check first. Is there a "/proc/evms" directory?
	 * If the EVMS kernel is running it will exist.
	 */
	status = stat("/proc/evms", &statbuf);
	if (status == 0) {
		if (S_ISDIR(statbuf.st_mode)) {

			/*
			 * Make a dev node for EVMS_MAJOR:0 and see if it can
			 * be opened.
			 */

			/* Find a random file name. */
			do {
				sprintf(temp_node_name, "%x", rand());

				status = stat(temp_node_name, &statbuf);
			} while (status == 0);

			status = mknod(temp_node_name, (S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP), evms_devt);

			if (status == 0) {
				fd = open(temp_node_name, O_RDWR|O_NONBLOCK);
				if (fd > 0) {
					/* Open succeeded.  EVMS kernel is running. */
					close(fd);
					rc = EEXIST;
				}

				unlink(temp_node_name);
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Check if the Engine is currently allowing read access.  The Engine must be
 * open and a commit must not be in progress.
 */
int check_engine_read_access() {

	int rc = 0;

	LOG_PROC_ENTRY();

	if ((engine_mode == ENGINE_CLOSED) ||
	    commit_in_progress) {
		if (engine_mode == ENGINE_CLOSED) {
			LOG_ERROR("The Engine is not open.\n");
		}
		if (commit_in_progress) {
			LOG_ERROR("The Engine is currently committing changes.\n");
		}
		rc = EACCES;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Check if the Engine is allowing write access, i.e., data structures, etc.
 * can be updated.  The Engine must be opened in a mode that allows writing
 * and a commit must not be in progress.
 */
int check_engine_write_access() {

	int rc = 0;

	LOG_PROC_ENTRY();

	if ((engine_mode == ENGINE_CLOSED) ||
	    (engine_mode == ENGINE_READONLY) ||
	    commit_in_progress) {
		if (engine_mode == ENGINE_CLOSED) {
			LOG_ERROR("The Engine is not open.\n");
		} else {
			if (engine_mode == ENGINE_READONLY) {
				LOG_ERROR("The Engine is open for reading only.\n");
			}
		}
		if (commit_in_progress) {
			LOG_ERROR("The Engine is currently committing changes.\n");
		}
		rc = EACCES;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/* Incremental size to allocate for the name array */
#define ALLOC_SIZE  4096

/* Mask to check if the number of array entries is on an allocation boundary */
#define ALLOC_MASK  ((ALLOC_SIZE / sizeof(char *)) - 1)

/*
 * Insert a name into the array, keeping the array sorted.
 */
static int insert_name(char * name, char * * * p_array, int * p_count) {

	int i;
	char * * array;
	int count = *p_count;
	int low;
	int curr;
	int high;
	int comp;

	LOG_PROC_ENTRY();

	if (!(count & ALLOC_MASK)) {
		/*
		 * The array has a number of entries that fills up an entire
		 * allocated space.  (Includes the initial array size of 0.)
		 * We need more space for the array.
		 */
		*p_array = engine_realloc(*p_array, (count * (ALLOC_SIZE / sizeof(char *))) + ALLOC_SIZE);
	}

	if (*p_array == NULL) {
		LOG_CRITICAL("Error allocating memory for the name list.\n");
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	array = *p_array;

	/* Binary search for the insertion slot */
	low = 0;
	high = count - 1;
	curr = (high + low) / 2;
	comp = 1;

	while ((low <= high) && (comp != 0)) {

		comp = strcmp(name, array[curr]);
		if (comp != 0) {
			if (comp < 0) {
				high = curr - 1;

			} else {
				low = curr + 1;
			}
			curr = (high + low) / 2;
		}
	}

	/*
	 * "low" is the index of the entry before which the new entry belongs.
	 * Move all the entries from "low" to "count -1" up one slot in the
	 * array to make room for the new entry.
	 */
	for (i = count; i > low; i--) {
		array[i] = array[i - 1];
	}

	array[low] = name;
	(*p_count)++;

	LOG_PROC_EXIT_INT(0);
	return 0;
}


/*
 * Remove a name from the array.
 */
static int remove_name(char * name, char * * array, int * p_count) {

	int rc = 0;
	int i;
	int count = *p_count;
	int low;
	int curr;
	int high;
	int comp;

	LOG_PROC_ENTRY();

	/* Binary search for the node name */
	low = 0;
	high = count - 1;
	curr = (high + low) / 2;
	comp = 1;

	while ((low <= high) && (comp != 0)) {

		comp = strcmp(name, array[curr]);
		if (comp != 0) {
			if (comp < 0) {
				high = curr - 1;
			} else {
				low = curr + 1;
			}
			curr = (high + low) / 2;
		}
	}

	if (comp == 0) {
		/* Found a match. */
		engine_free(array[curr]);

		/* Remove the match by collapsing the array. */
		for (i = curr + 1; i < count; i++) {
			array[i-1] = array[i];
		}

		array[count - 1] = NULL;
		(*p_count)--;

	} else {
		/* Did not find a match. */
		rc = ENOENT;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Get the names of the block device dev nodes in a given directory.
 * Return a pointer to an array of dev node name strings and a count of names
 * in the array.
 */
int get_dev_node_names(char * dir_name, char * * * name_list, int * count) {

	int rc = 0;
	struct stat stat_buf;
	DIR * dir;
	struct dirent * dir_ent;
	char * new_name;
	char name_buf[PATH_MAX];
	int dir_path_len;

	LOG_PROC_ENTRY();

	dir = opendir(dir_name);

	if (dir == NULL) {
		LOG_PROC_EXIT_INT(errno);
		return errno;
	}

	strcpy(name_buf, dir_name);
	dir_path_len = strlen(name_buf);
	name_buf[dir_path_len] = '/';
	dir_path_len++;

	for (dir_ent = readdir(dir);
	     (rc == 0) && (dir_ent != NULL);
	     dir_ent = readdir(dir)) {
		
		/* Skip the "." and ".." entries. */
		if ((strcmp(dir_ent->d_name,".") == 0) ||
		    (strcmp(dir_ent->d_name,"..") == 0)) {
			continue;
		}

		/*
		 * Build the full dev node name by appending the
		 * dir_ent->d_name to the directory name.
		 */
		strcpy(&name_buf[dir_path_len], dir_ent->d_name);

		if (stat(name_buf, &stat_buf) == 0) {
			/*
			 * If this entry is a directory, call this function
			 * recursively to process the subdirectory.
			 */
			if (S_ISDIR(stat_buf.st_mode)) {
				int prev_count = *count;
				
				get_dev_node_names(name_buf, name_list, count);

				if (prev_count == *count) {
					/*
					 * Nothing found in the directory.
					 * Remove it.
					 */
					rmdir(name_buf);
				}

			} else {
				if (S_ISBLK(stat_buf.st_mode)) {
					new_name = engine_strdup(name_buf);
					if (new_name != NULL) {
						rc = insert_name(new_name, name_list, count);

					} else {
						LOG_CRITICAL("Error getting memory for a name string.\n");
						rc = ENOMEM;
					}
				}
			}
		}
	}

	closedir(dir);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static void remove_dm_mapping(char * name) {

	if (strstr(name, EVMS_OBJECT_NODE_PATH) != NULL) {
		/*
		 * It's a dev node for an object.
		 * Build a dummy object to check if device-mapper has
		 * a map for the object name.  If so, deactivate the
		 * mapping.
		 */
		storage_object_t obj;

		memset(&obj, '\0', sizeof(storage_object_t));
		strcpy(obj.name, name + EVMS_OBJECT_NODE_PATH_LEN);

		dm_update_status(&obj);

		/*
		 * Object will be marked active if it has a
		 * device_mapper mapping.
		 */
		if (obj.flags & SOFLAG_ACTIVE) {
			dm_deactivate(&obj);
		}

	} else {
		/*
		 * It's a dev node for a volume.
		 * Build a dummy volume to check if device-mapper has
		 * a map for the volume name.  If so, deactivate the
		 * mapping.
		 */
		logical_volume_t vol;

		memset(&vol, '\0', sizeof(logical_volume_t));
		strcpy(vol.name, name);

		dm_update_volume_status(&vol);

		/*
		 * Volume will be marked active if it has a
		 * device_mapper mapping.
		 */
		if (vol.flags & VOLFLAG_ACTIVE) {
			dm_deactivate_volume(&vol);
		}
	}
}


void cleanup_dev_evms_tree(void) {

	int num_nodes = 0;
	int rc;
	int i;
	char * * dev_nodes = NULL;
	logical_volume_t * vol;
	storage_object_t * obj;
	dlist_t lists[] = {DiskList,
			   SegmentList,
			   RegionList,
			   EVMSObjectList,
			   NULL};
	char obj_name_buf[EVMS_OBJECT_NODE_PATH_LEN + EVMS_NAME_SIZE + 1] = EVMS_OBJECT_NODE_PATH;
	int status_count = 0;

	LOG_PROC_ENTRY();
	
	sync_volumes();

	status_message("Cleaning up the /dev/evms tree%s\n", status_dots(&status_count));

	/* Get an array of all the block dev nodes in the entire /dev/evms tree. */
	if (local_focus) {
		get_dev_node_names(EVMS_DEV_NODE_DIR, &dev_nodes, &num_nodes);
	} else {
		remote_get_dev_node_names(&current_nodeid, EVMS_DEV_NODE_DIR, &dev_nodes, &num_nodes);
	}

	status_message("Cleaning up the /dev/evms tree%s\n", status_dots(&status_count));

	/*
	 * dev_nodes contains a sorted array of nodes in the /dev/evms tree.
	 * Remove the entries for the volumes and object dev nodes that we know
	 * about.
	 */
	DLIST_FOR_EACH(vol, VolumeList, rc) {
		remove_name(vol->name, dev_nodes, &num_nodes);
	}

	for (i = 0; lists[i] != NULL; i++) {
		status_message("Cleaning up the /dev/evms tree%s\n", status_dots(&status_count));
		DLIST_FOR_EACH(obj, lists[i], rc) {
			if (obj->data_type == DATA_TYPE) {
				strcpy(&obj_name_buf[EVMS_OBJECT_NODE_PATH_LEN], obj->name);
				if (local_focus) {
					ensure_dev_node(obj_name_buf, obj->dev_major, obj->dev_minor);
				} else {
					remote_ensure_dev_node(&current_nodeid, obj_name_buf, obj->dev_major, obj->dev_minor);
				}
				remove_name(obj_name_buf, dev_nodes, &num_nodes);
			}
		}
	}
	status_message("Cleaning up the /dev/evms tree%s\n", status_dots(&status_count));

	/*
	 * All of the known nodes have been removed from the list.  What is left in
	 * the list are nodes for which we do not have a corresponding volume or
	 * object node.  unlink all the remaining nodes.
	 */
	for (i = 0; i < num_nodes; i++) {
		LOG_DETAILS("Removing dev node %s.\n", dev_nodes[i]);
		remove_dm_mapping(dev_nodes[i]);
		if (local_focus) {
			unlink(dev_nodes[i]);
		} else {
			remote_unlink(&current_nodeid, dev_nodes[i]);
		}
		engine_free(dev_nodes[i]);
	}

	engine_free(dev_nodes);

	status_message("Finished cleaning up the /dev/evms tree.\n");

	LOG_PROC_EXIT_VOID();
	return;
}


static void get_config_mode(char          * key,
			    engine_mode_t * open_mode) {

	int rc;
	const char * config_string;

	rc = evms_get_config_string(key, &config_string);
	if (rc == 0) {
		if (strcasecmp(config_string, "readonly") == 0) {
			*open_mode = ENGINE_READONLY;
		} else if (strcasecmp(config_string, "readwrite") == 0) {
			*open_mode = ENGINE_READWRITE;
		} else if (strcasecmp(config_string, "daemon") == 0) {
			*open_mode = ENGINE_DAEMON;
		}
	}
}


static void get_config_debug_level(char          * key,
				   debug_level_t * debug_level) {

	int rc;
	const char * config_string;

	rc = evms_get_config_string(key, &config_string);
	if (rc == 0) {
		if (strcasecmp(config_string, "CRITICAL") == 0) {
			*debug_level = CRITICAL;
		} else if (strcasecmp(config_string, "SERIOUS") == 0) {
			*debug_level = SERIOUS;
		} else if (strcasecmp(config_string, "ERROR") == 0) {
			*debug_level = ERROR;
		} else if (strcasecmp(config_string, "WARNING") == 0) {
			*debug_level = WARNING;
		} else if (strcasecmp(config_string, "DEFAULT") == 0) {
			*debug_level = DEFAULT;
		} else if (strcasecmp(config_string, "DETAILS") == 0) {
			*debug_level = DETAILS;
		} else if (strcasecmp(config_string, "DEBUG") == 0) {
			*debug_level = DEBUG;
		} else if (strcasecmp(config_string, "EXTRA") == 0) {
			*debug_level = EXTRA;
		} else if (strcasecmp(config_string, "ENTRY_EXIT") == 0) {
			*debug_level = ENTRY_EXIT;
		} else if (strcasecmp(config_string, "EVERYTHING") == 0) {
			*debug_level = EVERYTHING;
		}
	}
}


static void sigusr1_handler(int sig_no) {

	LOG_CRITICAL("***\n");
	LOG_CRITICAL("*** Received shutdown signal: %s.\n", sys_siglist[sig_no]);
	LOG_CRITICAL("***\n");

	sem_post(&shutdown_sem);
}


static void * shutdown_thread(void * arg) {

	sem_t * shutdown_sem = (sem_t *) arg;

	signal(SIGUSR1, sigusr1_handler);

	sem_init(shutdown_sem, 0, 0);

	sem_wait(shutdown_sem);
					
	shutdown_engine();

	return NULL;
}


#define VERSION_FIELDS(ver) ver.major, ver.minor, ver.patchlevel

int evms_get_api_version(evms_version_t * version) {
	*version = engine_api_version;
	return 0;
}


int evms_open_engine(char           * node_name,
		     engine_mode_t    mode,
		     ui_callbacks_t * callbacks,
		     debug_level_t    level,
		     char           * log_name) {
	int rc = 0;
	engine_mode_t open_mode;

	if (geteuid() != 0) {
		if (callbacks != NULL) {
			if (callbacks->user_message != NULL) {
                                callbacks->user_message("You must have root privilege to open the EVMS Engine.", NULL, NULL);
			}
		}

		return EACCES;
	}

	/* Setup defaults. */
	current_node_name = NULL;
	open_mode = ENGINE_READWRITE;
	debug_level = DEFAULT;
	log_usec = FALSE;
	log_pid = FALSE;

	if (mode == ENGINE_DAEMON) {
		log_file_name = DEFAULT_DAEMON_LOG_FILE;
	} else {
		log_file_name = DEFAULT_LOG_FILE;
	}

	ui_callbacks = callbacks;

	/*
	 * Get settings from the evms.conf.  They may override the default
	 * settings above.
	 */
	if (evms_get_config(NULL) == 0) {
		if (mode != ENGINE_DAEMON) {
			get_config_mode("engine.mode", &open_mode);
			get_config_debug_level("engine.debug_level", &debug_level);
			evms_get_config_string("engine.node", (const char * *) &current_node_name);
			evms_get_config_string("engine.log_file", (const char * *) &log_file_name);
			evms_get_config_bool("engine.log_usec", &log_usec);
			evms_get_config_bool("engine.log_pid", &log_pid);

		} else {
			get_config_debug_level("daemon.debug_level", &debug_level);
			evms_get_config_string("daemon.log_file", (const char * *) &log_file_name);
			evms_get_config_bool("daemon.log_usec", &log_usec);
			evms_get_config_bool("daemon.log_pid", &log_pid);
		}
	}

	/*
	 * Get user supplied settings, which override both the defaults and the
	 * settings in evms.conf.
	 */
	if (node_name != NULL) {
		current_node_name = node_name;
	}

	if (mode != -1) {
		open_mode = mode;
	}

	if (level != -1) {
		debug_level = level;
	}

	if (log_name != NULL) {
		log_file_name = log_name;
	}

	start_logging(log_file_name);

	LOG_PROC_ENTRY();

	LOG_DEFAULT("Engine version:          %2d.%d.%d\n", VERSION_FIELDS(engine_version));
	LOG_DEFAULT("External API version:    %2d.%d.%d\n", VERSION_FIELDS(engine_api_version));
	LOG_DEFAULT("Engine services version: %2d.%d.%d\n", VERSION_FIELDS(engine_services_api_version));
	LOG_DEFAULT("Plug-in API version:     %2d.%d.%d\n", VERSION_FIELDS(engine_plugin_api_version));
	LOG_DEFAULT("Container API version:   %2d.%d.%d\n", VERSION_FIELDS(engine_container_api_version));
	LOG_DEFAULT("FSIM API version:        %2d.%d.%d\n", VERSION_FIELDS(engine_fsim_api_version));
	LOG_DEFAULT("Cluster API version:     %2d.%d.%d\n", VERSION_FIELDS(engine_cluster_api_version));
	LOG_DEFAULT("Daemon protocol version: %2d.%d.%d\n", VERSION_FIELDS(engine_daemon_msg_version));

	if ((open_mode != ENGINE_READONLY) &&
	    (open_mode != ENGINE_READWRITE) &&
	    (open_mode != ENGINE_READWRITE_CRITICAL) &&
	    (open_mode != ENGINE_DAEMON)) {
		LOG_ERROR("Open mode of %d is not valid.\n", open_mode);

		evms_free_config();
		LOG_PROC_EXIT_INT(EINVAL);
		return EINVAL;
	}

	if (engine_mode != ENGINE_CLOSED) {
		/* This Engine is already open. */
		LOG_ERROR("The Engine is already opened.\n");
		evms_free_config();
		LOG_PROC_EXIT_INT(EINVAL);
		return EINVAL;
	}

	/* Check if we can access the EVMS kernel code. */
	rc = check_for_evms_kernel();
	if (rc != 0) {
		engine_user_message(NULL, NULL, "This system is running the EVMS kernel.  "
				    "EVMS Engine versions 1.9.0 and above do not run with the EVMS kernel.  "
				    "Use EVMS Engine version 1.2 to configure with the EVMS kernel.\n");
		evms_free_config();
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	engine_mode = open_mode;

	/* Initialize the random number generator. */
	srand(time(NULL) + getpid());

	/* Install signal handlers. */
	install_signal_handlers();

	if (open_dm_control_node() != 0) {
		engine_user_message(NULL, NULL,
				    "Unable to open the control node for Device-Mapper.  "
				    "The Engine will run without Device-Mapper support.\n");
	}

	rc = create_evms_dlists();
	if (rc == 0) {

		rc = lock_engine(open_mode, NULL);
		if (rc == 0) {

			rc = load_plugins();
			if (rc == 0) {

				if (open_mode == ENGINE_DAEMON) {

					/* Daemon can't run without a cluster manager? */
					if (cluster_manager == NULL) {
						engine_user_message(NULL, NULL,
								    "There is no cluster manager plug-in loaded on this system.\n");
						rc = ENODEV;
					}

				} else {
					pthread_t shutdown_tid;

					if ((cluster_manager == NULL) && (current_node_name != NULL)) {
						engine_user_message(NULL, NULL,
								    "There is no cluster manager plug-in loaded on this system.  "
								    "The node_name parameter \"%s\" is ignored.\n",
								    current_node_name);
					}

					pthread_create(&shutdown_tid,
						       NULL,
						       shutdown_thread,
						       &shutdown_sem);

					/* Initialize the Handle Manager. */
					if (initialize_handle_manager()) {

						rc = do_discovery();
						if (rc == 0) {
							cleanup_dev_evms_tree();

						} else {
							destroy_all_handles();
						}

					} else {
						LOG_CRITICAL("Handle Manager failed to initialize.\n");
						rc = ENOMEM;
					}
				}

				if (rc != 0) {
					/* Something went wrong.  Mark the Engine closed. */
					unlock_engine(engine_mode);
					engine_mode = ENGINE_CLOSED;
				}
			}
		}

		if (rc != 0) {
			/*
			 * Something went wrong.
			 * Destroy the lists we created.
			 */
			destroy_evms_dlists();
		}

	} else {
		LOG_CRITICAL("create_evms_dlists failed with return code %d.\n", rc);

		/* Destroy any dlists that might have been created. */
		destroy_evms_dlists();
	}

	if (rc != 0) {
		close(dm_control_fd);
		evms_free_config();
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


void free_changes_pending_record_array_contents(void * thing) {

	change_record_array_t * changes = (change_record_array_t *) thing;
	int i;

	for (i = 0; i < changes->count; i++) {
		engine_free(changes->changes_pending[i].name);
	}
}


int evms_changes_pending(BOOLEAN                 * result,
			 change_record_array_t * * p_changes) {

	int rc = 0;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		if (p_changes == NULL) {
			*result = are_changes_pending(NULL);

		} else {
			change_record_array_t * changes = NULL;
			change_record_array_t * user_changes = NULL;

			*result = are_changes_pending(&changes);

			if (changes != NULL) {
				user_changes = alloc_app_struct(sizeof(change_record_array_t) + (sizeof(change_record_t) * changes->count),
								free_changes_pending_record_array_contents);

				if (user_changes != NULL) {
					int i;

					user_changes->count = changes->count;

					for (i = 0; i < changes->count; i++) {
						user_changes->changes_pending[i].name = engine_strdup(changes->changes_pending[i].name);
						user_changes->changes_pending[i].type = changes->changes_pending[i].type;
						user_changes->changes_pending[i].changes = changes->changes_pending[i].changes;
					}

				} else {
					LOG_CRITICAL("Error allocating memory for a change_record_array_t.\n");
				}

				engine_free(changes);
			}

			*p_changes = user_changes;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Get the Engine's current setting of the debug level.
 */
int evms_get_debug_level(debug_level_t * level) {

	int rc = 0;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();

	if (rc == 0) {
		*level = debug_level;
	}

	LOG_PROC_EXIT_INT(0)
	return 0;
}


/*
 * Set the Engine's debug level.
 */
int evms_set_debug_level(debug_level_t level) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * People that opened the Engine for read-only are allowed to set their
	 * debug level.
	 */
	rc = check_engine_read_access();

	if (rc == 0) {

		if ((level >= CRITICAL) &&
		    (level <= EVERYTHING)) {
			debug_level = level;

		} else {
			LOG_ERROR("Debug level is out of range.  It must be between %d and %d, inclusive.\n", CRITICAL, EVERYTHING);
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc)
	return(rc);
}


/*-----------------------------------------------------------------------------+
+                                                                              +
+                              Logging Functions                               +
+                                                                              +
+-----------------------------------------------------------------------------*/

int evms_write_log_entry(debug_level_t level,
			 char        * module_name,
			 char        * fmt,
			 ...) {
	int rc = 0;
	int len;
	va_list args;

	if (dm_device_suspended) {
		return rc;
	}

	rc = check_engine_read_access();
	if (rc == 0) {
		if (level <= debug_level) {
			if (log_file_fd > 0) {
				timestamp(log_buf, LOG_BUF_SIZE);
				strcat(log_buf, module_name);
				strcat(log_buf, ":  ");
				len = strlen(log_buf);

				va_start(args, fmt);
				len += vsprintf(log_buf + strlen(log_buf), fmt, args);
				va_end(args);

				if (write(log_file_fd, log_buf, len) < 0) {
					rc = errno;
				}

			} else {
				rc = ENOENT;
			}
		}
	}

	return rc;
}


/*-----------------------------------------------------------------------------+
+                                                                              +
+                             Some List Handlers                               +
+                                                                              +
+-----------------------------------------------------------------------------*/

/*
 * Make sure this object has an app_handle.
 */
int ensure_app_handle(void * object, object_type_t object_type, object_handle_t * app_handle) {
	int rc = 0;

	LOG_PROC_ENTRY();

	if (*app_handle == 0) {
		LOG_DEBUG("Create a handle for a thing of type %d.\n", object_type);
		rc = create_handle(object,
				   object_type,
				   app_handle);
		if (rc == 0) {
			LOG_DEBUG("Handle is %d.\n", *app_handle);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Add the app_handle for a given object to a handle_array_t.
 * The parameters of this function are structured so that it can be called by
 * ForEachItem().
 */
int make_handle_entry(ADDRESS Object,
		      TAG     Tag,
		      ADDRESS ObjectHandle,
		      ADDRESS Parameters) {
	int rc = 0;

	handle_array_t * ha = (handle_array_t *) Parameters;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Current number of entries in handle array:  %d.\n", ha->count);
	switch (Tag) {
	case PLUGIN_TAG:
		{
			plugin_record_t * pPlugRec = (plugin_record_t *) Object;

			LOG_DEBUG("Add entry for plug-in %s.\n", pPlugRec->short_name);
			rc = ensure_app_handle(Object,
					       PLUGIN,
					       &pPlugRec->app_handle);
			if (rc == 0) {
				ha->handle[ha->count] = pPlugRec->app_handle;
				ha->count++;
			}
		}
		break;

	case DISK_TAG:
	case SEGMENT_TAG:
	case REGION_TAG:
	case EVMS_OBJECT_TAG:
		{
			storage_object_t * pObj = (storage_object_t *) Object;

			LOG_DEBUG("Add entry for storage object %s.\n", pObj->name);
			rc = ensure_app_handle(Object,
					       pObj->object_type,
					       &pObj->app_handle);
			if (rc == 0) {
				ha->handle[ha->count] = pObj->app_handle;
				ha->count++;
			}
		}
		break;

	case CONTAINER_TAG:
		{
			storage_container_t * pCon = (storage_container_t *) Object;

			LOG_DEBUG("Add entry for container %s.\n", pCon->name);
			rc = ensure_app_handle(Object,
					       CONTAINER,
					       &pCon->app_handle);
			if (rc == 0) {
				ha->handle[ha->count] = pCon->app_handle;
				ha->count++;
			}
		}
		break;

	case VOLUME_TAG:
		{
			logical_volume_t * pVol = (logical_volume_t *) Object;

			LOG_DEBUG("Add entry for volume %s.\n", pVol->name);
			rc = ensure_app_handle(Object,
					       VOLUME,
					       &pVol->app_handle);
			if (rc == 0) {
				ha->handle[ha->count] = pVol->app_handle;
				ha->count++;
			}
		}
		break;

	default:
		/* It's nothing for which we hand out app handles.  Skip it. */
		LOG_WARNING("Attempt to make an app handle for an object of tag %ld.\n", Tag);
		break;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Make a user (it has a Engine memory block header) array of handles
 * (handle_array_t) for the objects in a dlist_t.
 */
int make_user_handle_array(dlist_t list, handle_array_t * * ha) {
	int rc = 0;
	uint count;
	uint size;

	LOG_PROC_ENTRY();

	rc = GetListSize(list, &count);

	if (rc == 0) {
		LOG_DEBUG("Number of objects in the list:  %d\n", count);
		if (count > 1) {
			size = sizeof(handle_array_t) + ((count -1) * sizeof(object_handle_t));
		} else {
			size = sizeof(handle_array_t);
		}

		*ha = alloc_app_struct(size, NULL);
		if (*ha != NULL) {
			rc = ForEachItem(list,
					 make_handle_entry,
					 *ha,
					 TRUE);
		} else {
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Get a list of handles for plug-ins, optionally filtering by plug-in type.
 */
int evms_get_plugin_list(evms_plugin_code_t    type,
			 plugin_search_flags_t flags,
			 handle_array_t    * * plugin_handle_list) {
	int rc = 0;
	dlist_t plugin_list;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();
	if (rc == 0) {

		if (plugin_handle_list != NULL) {
			rc = engine_get_plugin_list(type, flags, &plugin_list);

			if (rc == 0) {
				rc = make_user_handle_array(plugin_list, plugin_handle_list);

				DestroyList(&plugin_list, FALSE);
			}

		} else {
			LOG_DEBUG("User specified NULL pointer for plugin_handle_list. I'm not doing anything since there is nowhere to put the results.\n");
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Get the handle for a plug-in with a given ID.
 */
int evms_get_plugin_by_ID(plugin_id_t       plugin_ID,
			  plugin_handle_t * plugin_handle) {
	int rc = 0;
	plugin_record_t * pPlugRec;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();
	if (rc == 0) {
		rc = engine_get_plugin_by_ID(plugin_ID, &pPlugRec);

		if (rc == 0) {
			rc = ensure_app_handle(pPlugRec,
					       PLUGIN,
					       &pPlugRec->app_handle);

			if (rc == 0) {
				*plugin_handle = pPlugRec->app_handle;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Get the handle for a plug-in with a given short name.
 */
int evms_get_plugin_by_name(char            * plugin_name,
			    plugin_handle_t * plugin_handle) {
	int rc = 0;

	plugin_record_t * pPlugRec;

	LOG_PROC_ENTRY();

	rc = check_engine_read_access();
	if (rc == 0) {
		rc = engine_get_plugin_by_name(plugin_name, &pPlugRec);

		if (rc == 0) {
			rc = ensure_app_handle(pPlugRec,
					       PLUGIN,
					       &pPlugRec->app_handle);

			if (rc == 0) {
				*plugin_handle = pPlugRec->app_handle;
			}
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/********************************************/
/* Functions for updating the dev node tree */
/********************************************/

/*
 * Make a directory in the EVMS dev node tree.
 * This function is a utility function for make_evms_dev_node() below.
 */
int make_directory(char * dir_name, evms_verbose_level_t verbose) {
	char name_buf[EVMS_VOLUME_NAME_SIZE + 1];
	char * tmp_ptr = name_buf;
	struct stat statbuf;
	int rc = 0;

	LOG_PROC_ENTRY();

	/* See if this directory exists */
	rc = stat(dir_name, &statbuf);
	if (rc != 0) {
		if (errno == ENOENT) {
			/*
			 * This directory doesn't exist.  Peel off the top-most directory
			 * name, and try again.
			 */
			strcpy(name_buf, dir_name);
			tmp_ptr = strrchr(name_buf, '/');

			if (tmp_ptr != NULL) {
				*tmp_ptr = 0;

				/* Call myself recursively to make the parent directory. */
				rc = make_directory(name_buf, verbose);
				if (rc == 0) {
					rc = mkdir(dir_name, (S_IFDIR | S_IRWXU |
							      S_IRGRP | S_IXGRP |
							      S_IROTH | S_IXOTH));

					if (rc != 0) {
						rc = errno;
						LOG_WARNING("mkdir(%s) failed with error code %d.\n", dir_name, rc);
						if (verbose >= ERRORS_ONLY) {
							fprintf(stderr, "mkdir(%s) failed with error code %d.\n", dir_name, rc);
						}
					}
				}

			} else {
				/*
				 * Either we were passed a relative path name, or the root
				 * file system doesn't exist.
				 */
				rc = ENODEV;
			}

		} else {
			rc = errno;
			LOG_WARNING("stat(%s) failed with error code %d.\n", dir_name, rc);
			if (verbose >= ERRORS_ONLY) {
				fprintf(stderr, "stat(%s) failed with error code %d.\n", dir_name, rc);
			}
		}

	} else {
		/* Make sure this is a directory */
		if (!(statbuf.st_mode & S_IFDIR)) {
			LOG_ERROR("%s is a non-directory file\n", dir_name);
			if (verbose >= ERRORS_ONLY) {
				fprintf(stderr, "%s is a non-directory file\n", dir_name);
			}
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


int evms_update_evms_dev_tree(evms_verbose_level_t verbose) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/* Opening the Engine updates the /dev/evms tree. */
	rc = evms_open_engine(NULL, ENGINE_READONLY, NULL, DEFAULT, NULL);

	if (rc == 0) {
		evms_close_engine();
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


BOOLEAN is_local_object(storage_object_t * obj) {

	BOOLEAN result = TRUE;
	LOG_PROC_ENTRY();

	LOG_DEBUG("Check object %s.\n", obj->name);

	/* Must have a cluster manager installed for the object to be remote. */
	if (cluster_manager != NULL) {

		/*
		 * Any sharable storage, private or sharable, can be accessed
		 * locally.
		 */
		if (!(obj->flags & (SOFLAG_CLUSTER_PRIVATE | SOFLAG_CLUSTER_SHARED))) {
			/*
			 * If the Engine was opened for a remote node, then the
			 * object is remote.
			 */
			result = local_focus;
		}
	}

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}


BOOLEAN is_local_volume(logical_volume_t * vol) {

	BOOLEAN result = TRUE;
	LOG_PROC_ENTRY();

	LOG_DEBUG("Check volume %s.\n", vol->name);

	/* Must have a cluster manager installed for the volume to be remote. */
	if (cluster_manager != NULL) {

		/*
		 * If the Engine was opened for a remote node, then the volume
		 * is remote, regardless if it is on sharable storage.  Volumes
		 * must be accessed on the node on which they reside.
		 */
		result = local_focus;
	}

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}


const char * evms_strerror(int err_num) {

	unsigned int err = abs(err_num);

	if (err <= EMEDIUMTYPE) {
		return strerror(err_num);
	}

	if (IS_DLIST_ERROR(err)) {
		return dlist_strerror(err);
	}

	if (IS_HANDLE_MANAGER_ERROR(err)) {
		return handlemgr_strerror(err);
	}

	if (err == E_CANCELED) {
		return "Operation canceled";
	}

	if (err == E_NOLOAD) {
		return "Plug-in did not want to load";
	}

	return "Unknown error code";
}
