/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libaixregmgr.so
 *
 *   File: aix_dm.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "aixregmgr.h"	


/*
 *  Function: aix_activate
 *
 *  Called in order to create a mapping of an AIX Logical
 *  Volume (evms region) in the kernel.
 */
int aix_activate_LV(storage_object_t * region)
{
	dm_target_t *target_list;
	int rc = EINVAL;
        aix_region_data_t *r_pdata;

	LOG_ENTRY();

	r_pdata = (region != NULL && region->plugin==my_plugin_record) ? region->private_data : NULL;

	if (r_pdata != NULL) {
			
		LOG_DEBUG("activating region %s.\n", region->name);

		rc = aix_build_LV_targets(region, &target_list);
		if (!rc) {
		
			rc = EngFncs->dm_activate(region, target_list);

			EngFncs->dm_deallocate_targets(target_list);
		}

		if (!rc) {				
			region->flags |=  SOFLAG_ACTIVE;
			region->flags &= ~SOFLAG_NEEDS_ACTIVATE;
		}
		else if (rc == ENOSYS) {
			LOG_DEBUG("clearing needs activate flag ... not supported\n");
			region->flags &= ~SOFLAG_ACTIVE;
			region->flags &= ~SOFLAG_NEEDS_ACTIVATE;
			rc = 0;
		}

	}
	else {
		LOG_ERROR("error, region not mine or NULL ... returning EINVAL\n");
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Function: aix_deactivate
 *
 *  Called in order to delete an existing mapping of an AIX
 *  Logical Volume (evms region) in the kernel.
 */
int aix_deactivate_LV( storage_object_t * region )
{
        int rc=EINVAL;

        LOG_ENTRY();

	if (region != NULL && region->plugin==my_plugin_record) {
	
		rc = EngFncs->dm_deactivate(region);
		if (!rc) {
			region->flags &= ~SOFLAG_ACTIVE;
		}

	}

        LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Function: is_next_lps_phys_contiguous
 *
 *  Called to see if the next logical to physical mapping (lps) for 
 *  the specified region (LV) would describe a physical extent (pp)
 *  on an evms storage object (PV) that is physically contiguous to
 *  the current physical extent (pp) being mapped by the current lps.
 */
static BOOLEAN is_next_lps_phys_contiguous( aix_region_data_t *r_pdata, u_int32_t  current )
{
	BOOLEAN result = FALSE;			     // BOOLEAN result
	u_int32_t  num_lps = r_pdata->lv->num_lps;   // total lps mappings in this volume
	u_int32_t  pp_current;			     // current phys extent number
	u_int32_t  pp_next;			     // next    phys extent number
	storage_object_t *pv_current;		     // current phys volume
	storage_object_t *pv_next;		     // next    phys volume

	LOG_ENTRY();

	// dont walk off end of lps mapping list
	if ( current+1 < num_lps ) {  

		// get current pv and next pv
		pv_current = r_pdata->lp_map[current].object;
		pv_next    = r_pdata->lp_map[current+1].object;

		// get current pp and next pp
		pp_current = r_pdata->lp_map[current].pp_number;
		pp_next    = r_pdata->lp_map[current+1].pp_number;

		// test if contiguous
		if ( pv_current   == pv_next &&    // if same PV (evms object) and
		     pp_current+1 == pp_next ) {   // next pp follows the current pp
			result=TRUE;		   // then the next lps is contiguous
		}				   // to the current lps.

	}

	LOG_EXIT_BOOL(result);
	return result;
}
 
/*
 *  Function: build mirrored target list
 *
 *  Called when the aix region contains mirrored physical extents.
 *
 *  Returns:  device mapper target list if successful
 *            NULL on failure
 *
 *  NOTE: there isnt an aix device mapper plugin for the kernel yet that
 *        would support aix mirrored volumes. this is because aix mirrors
 *        physical extents and could have up to a 3-way mirror of pps.
 */
static 	dm_target_t *  build_mirrored_target_list( aix_region_data_t    *r_pdata,
				 	           aix_container_data_t *c_pdata )
{
	LOG_ENTRY();
	LOG_DEBUG("error, mirroring not supported.\n");
	LOG_EXIT_PTR(NULL);
	return NULL;
}


/*
 *  Function: build linear target list
 *
 *  Called when the aix region contains a simple linear mapping to 
 *  physical volumes.
 *
 *  Returns:  device mapper target list if successful
 *            NULL on failure
 */
static 	dm_target_t *  build_linear_target_list( aix_region_data_t    *r_pdata,
				 	         aix_container_data_t *c_pdata )
{
	int i;
	dm_target_t    *target = NULL;
	dm_target_t    *target_list = NULL;
	dm_device_t    *linear = NULL;	
	u_int32_t	pp_size;
	u_int32_t 	logical_extent_count;   
	u_int64_t 	start_le, num_le; 
	    
	pp_size              = PP_SIZE_TO_VSECTORS(c_pdata->vgda->vg_head->pp_size);
	logical_extent_count = r_pdata->lv->num_lps;
	start_le             = 0;
	num_le               = 1;	

	for (i=0; i < logical_extent_count; i++, num_le++) {
  
		if ( is_next_lps_phys_contiguous( r_pdata, i ) == FALSE ) {
			         
			target = EngFncs->dm_allocate_target( DM_TARGET_LINEAR,          // type
								 start_le * pp_size,        // start
								 num_le * pp_size,          // length
								 0 );                       // ignored for linear
			if (!target) {
				LOG_ERROR("error: dm_allocate_target failed\n");
				if (target_list) EngFncs->dm_deallocate_targets(target_list); 
				LOG_EXIT_PTR(NULL);
				return NULL;
			}
 			
			linear        = (dm_device_t *) target->data.linear;
			linear->major = r_pdata->lp_map[start_le].object->dev_major;        // device major
			linear->minor = r_pdata->lp_map[start_le].object->dev_minor;        // device minor
			linear->start = (r_pdata->lp_map[start_le].pp_number * pp_size) +   // sector offset on 				 
					           c_pdata->vgda->pv_headers[0]->psn_part1; // the device

			EngFncs->dm_add_target(target, &target_list);			
  
			start_le += num_le;
			num_le = 0;		
		}
	
	}

	LOG_EXIT_PTR(target_list);
	return target_list;
}

/*
 *  Function: build stripped target list
 *
 *  Called when the aix region contains a stripped mapping to 
 *  physical volumes.
 *
 *  Returns:  device mapper target list if successful
 *            NULL on failure
 */
static 	dm_target_t *  build_striped_target_list( aix_region_data_t    *r_pdata,
				                  aix_container_data_t *c_pdata )
{
	int i, j;
	u_int64_t pp_size;
	u_int64_t logical_extent_count;	
	u_int64_t stripes;
	u_int64_t stripe_chunk_size;
	u_int64_t start_le; 

	dm_target_t          *target = NULL;
	dm_target_t          *target_list = NULL;
	dm_target_stripe_t   *stripe = NULL;	
 	dm_device_t    	     *linear = NULL; 

	LOG_ENTRY();

	pp_size                     = PP_SIZE_TO_VSECTORS(c_pdata->vgda->vg_head->pp_size);
 	stripes                     = r_pdata->lv->striping_width;
	stripe_chunk_size           = 1 << r_pdata->lv->stripe_exp;	
	logical_extent_count 	    = r_pdata->lv->num_lps;
 	start_le             	    = 0;			
   
	for (i=0; i < logical_extent_count; i+=stripes, start_le+=stripes ) {
   			         
       		target = EngFncs->dm_allocate_target( DM_TARGET_STRIPE,                  // type
       							 start_le * pp_size,                // start
       							 stripes  * pp_size,                // length
       							 stripes );                         // stripe count
       		if (!target) {
			LOG_ERROR("error: dm_allocate_target failed\n");
			if (target_list) EngFncs->dm_deallocate_targets(target_list); 
			LOG_EXIT_PTR(NULL);
			return NULL;
       		}

       		stripe              = (dm_target_stripe_t *) target->data.stripe;
       		stripe->num_stripes = stripes;
       		stripe->chunk_size  = stripe_chunk_size;			

       		for (j = 0; j < stripes; j++) {
       			linear        = &stripe->devices[j];
       			linear->major = r_pdata->lp_map[start_le + j].object->dev_major;    // device major
       			linear->minor = r_pdata->lp_map[start_le + j].object->dev_minor;    // device minor
       			linear->start = (r_pdata->lp_map[start_le + j].pp_number * pp_size) + // sector offset on
       						c_pdata->vgda->pv_headers[0]->psn_part1;    // the device
       		}

       		EngFncs->dm_add_target(target, &target_list); 

	}

	LOG_EXIT_PTR(target_list);
	return target_list;	
}



/*
 *  Function:  aix_build_LV_targets
 *
 *  Build a DM target list to represent the AIX Logical 
 *  Volume (evms region).
 *
 *  Returns:  0 on success
 *           -1 on failure and NULLs the callers ptr
 */
int aix_build_LV_targets( storage_object_t *region, dm_target_t **target_list )
{
	storage_container_t  *container;
	aix_container_data_t *c_pdata;
	aix_region_data_t    *r_pdata;
	int rc;

	LOG_ENTRY();

	container  = (region != NULL)    ?  region->producing_container  :  NULL;
	c_pdata    = (container != NULL) ?  (aix_container_data_t *) container->private_data  :  NULL;
	r_pdata    = (region != NULL)    ?  (aix_region_data_t *) region->private_data  :  NULL;

	if ( c_pdata == NULL || r_pdata == NULL ) {
		LOG_ERROR("error, bad container or region object.\n");
		*target_list = NULL;
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}

	// check for complex LV
	if ( ( r_pdata->lv->striping_width > 1 ) &&    // if stripping
	     ( r_pdata->lv->mirror > 1) ){	       // and mirroring
		LOG_ERROR("cant handle an LV that is both mirrored and stripped.\n");
		*target_list = NULL;
		LOG_EXIT_INT(ENOSYS);
		return ENOSYS;
	}
      
	// check stripes ... 1 means LV is built on a single PV and is not stripped
	if ( r_pdata->lv->striping_width > 1 ) {
		*target_list = build_striped_target_list(r_pdata, c_pdata);
	}
	// check mirror policy ... 1 means no pp's are mirrored 
	else if ( r_pdata->lv->mirror > 1) {		
		LOG_ERROR("mirroring not supported\n");
		LOG_EXIT_INT(ENOSYS);
		return ENOSYS;
		*target_list = build_mirrored_target_list(r_pdata, c_pdata);		
	}
	// else must be a simple linear volume
	else {
		*target_list = build_linear_target_list(r_pdata,c_pdata);
	}

	rc = (*target_list != NULL) ? 0 : -1;      
 
	LOG_EXIT_INT(rc);
	return rc;
}



/*
 *  Function:  aix_compare_volume_targets
 *
 *  Compare the engine's current view of the mapping for this AIX
 *  Logical Volume (evms region) against the kernel's current mapping 
 *  of the same Logical Volume. 
 *  
 *  Returns:  -1 if they are different                                         
 *             0 if they match
 */
int aix_compare_LV_targets( storage_object_t *region )
{
	dm_target_t 	   *dm_target_list=NULL;
 	dm_target_t 	   *dm_target=NULL;
	dm_device_t 	   *dm_linear=NULL;
	dm_target_stripe_t *dm_stripe=NULL;
	int                 dm_target_count=0;
	BOOLEAN             dm_consistant=TRUE;
	int                 dm_target_type=DM_TARGET_LINEAR;

	dm_target_t 	   *aix_target_list=NULL;
	dm_target_t 	   *aix_target=NULL;
	dm_device_t 	   *aix_linear=NULL;
	dm_target_stripe_t *aix_stripe=NULL;
	int		    aix_target_count=0;
	BOOLEAN             aix_consistant=TRUE;
	int		    aix_target_type=DM_TARGET_LINEAR;

	u_int32_t i, num_stripes=1;
	int rc;

	LOG_ENTRY();

	LOG_DETAILS("examining region %s\n", region->name );

	/* Get the kernel's mapping for this region. */
	rc = EngFncs->dm_get_targets(region, &dm_target_list);
	if (!rc) {
		 /* Count targets in list */
		for (dm_target = dm_target_list; dm_target; dm_target = dm_target->next) {
			++dm_target_count;
		}

		/* see if they are all a consistant target type */
		dm_target_type = dm_target_list->type;
		for (dm_target = dm_target_list; dm_target; dm_target = dm_target->next) {
			if (dm_target->type != dm_target_type){
				dm_consistant=FALSE;
				break;
			}
		}
	}
	else {
		rc = -1;
		LOG_ERROR("error, getting kernel mapping for region failed\n");		
	}

	/* Get the engine's mapping for this region. */
	rc = aix_build_LV_targets(region, &aix_target_list);
	if (!rc) {
		/* Count targets in list */
		for (aix_target = aix_target_list; aix_target; aix_target = aix_target->next) {
			++aix_target_count;
		}

		/* see if they are all a consistant target type */
		aix_target_type = aix_target_list->type;
		for (aix_target = aix_target_list; aix_target; aix_target = aix_target->next) {
			if (aix_target->type != aix_target_type){
				aix_consistant=FALSE;
				break;
			}
		}

	}
	else {
		rc = -1;
		LOG_ERROR("error, building engine mapping for region failed\n");
	}
 
	/* Sanity check the two lists prior to comparison of individual targets */
	if (!rc) {
		if ( dm_target_count != aix_target_count ||
		     dm_target_type  != aix_target_type  ||
		     dm_consistant   == FALSE            ||
		     aix_consistant  == FALSE  ) {
			rc = -1;
			LOG_ERROR("target lists differ with respect to: count, target type or target consistancy.\n");
		}
	}

	/* Compare the kernel's mapping with the current mapping. */
	for ( dm_target = dm_target_list, aix_target = aix_target_list;
	      dm_target && aix_target && rc==0;
	      dm_target = dm_target->next, aix_target = aix_target->next ) {

		if ( dm_target->type != aix_target->type ) {
			LOG_ERROR("error, target type mismatch\n");
			rc = -1;
			break;
		}

		if ( dm_target->start  != aix_target->start ||
		     dm_target->length != aix_target->length ||
		     dm_target->type   != aix_target->type ) {

			LOG_ERROR("error, start/length/type dont match\n");
			rc = -1;
			break;

		}

		if ( dm_target->type == DM_TARGET_STRIPE ) {

			dm_stripe  = dm_target->data.stripe;
			aix_stripe = aix_target->data.stripe;

			if ( dm_stripe->num_stripes != aix_stripe->num_stripes ||
			     dm_stripe->chunk_size  != aix_stripe->chunk_size ) {

				LOG_ERROR("error, stripe info doesnt match\n");
				rc = -1;
				break;

			}

			num_stripes = dm_stripe->num_stripes;
		}

		for (i = 0; i < num_stripes; i++) {

			dm_linear = (dm_target->type == DM_TARGET_STRIPE) ?
				    &dm_stripe->devices[i] : dm_target->data.linear;

			aix_linear = (aix_target->type == DM_TARGET_STRIPE) ?
				     &aix_stripe->devices[i] : aix_target->data.linear;

			if ( dm_linear->major != aix_linear->major ||
			     dm_linear->minor != aix_linear->minor ||
			     dm_linear->start != aix_linear->start ) {

				LOG_ERROR("error, linear info doesnt match\n");
				rc = -1;
				break;

			}

		}
	}

	if (dm_target_list)  EngFncs->dm_deallocate_targets(dm_target_list);
	if (aix_target_list) EngFncs->dm_deallocate_targets(aix_target_list);

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Function: aix_update_LV_devmap_info
 *
 *  Called to update device mapper info for the AIX Logical
 *  Volume (evms region).
 */
int aix_update_LV_devmap_info( storage_object_t *region )
{
        int rc=EINVAL;
	aix_region_data_t *r_pdata;
	dm_target_t *targets=NULL;

	LOG_ENTRY();

       	r_pdata = (region != NULL) ? (aix_region_data_t *) region->private_data : NULL;

	if (r_pdata != NULL) {
	
		LOG_DEBUG("Updating device mapper info for region %s.\n", region->name);

		rc = EngFncs->dm_update_status(region);
		if (!rc) {

			if (region->flags & SOFLAG_ACTIVE) {

				if ( aix_compare_LV_targets(region) ){
					LOG_DEBUG("marking region ...needs activate\n");
					region->flags |= SOFLAG_NEEDS_ACTIVATE;
				}

			}
			else {
			
				if (r_pdata->flags & AIX_REGION_EXPORTED) {
					
					rc =  aix_build_LV_targets(region, &targets);
					if (!rc) {
						LOG_DEBUG("region not active ... marking it ... needs activate\n");
						region->flags |= SOFLAG_NEEDS_ACTIVATE;
					}
					else if (rc == ENOSYS) {
						LOG_DEBUG("region is not supported ... cant activate\n");
						region->flags &= ~SOFLAG_NEEDS_ACTIVATE;
					}
					else {
						LOG_DEBUG("errors building the target list ... marking it ... needs activate\n");
						region->flags |= SOFLAG_NEEDS_ACTIVATE;
					}
					
					if (targets) EngFncs->dm_deallocate_targets(targets);

					rc = 0;
 				}
				else {
					LOG_DEBUG("region not active ... but not exported ... removing needs activate\n");
					region->flags &= ~SOFLAG_NEEDS_ACTIVATE;
				}

			}

		}

	}
	else {
		LOG_ERROR("error, called with invalid region object ... returning EINVAL\n");
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Function: aix_update_VG_devmap_info
 *
 *  Called to update the devmap info for the AIX Volume Group
 *  (evms container) which means updating the info for each 
 *  Logical Volume (evms region) produced by the Volume Group.
 */
int aix_update_VG_devmap_info( storage_container_t *container )
{
        storage_object_t *region;
        int rc=0;

        LOG_ENTRY();

	LOG_DETAILS("Updating device mapper info for container %s.\n", container->name);

	DLIST_FOR_EACH( region, container->objects_produced, rc ) {

		if (region->object_type == REGION) {		

			aix_update_LV_devmap_info(region);

		}

	}

        LOG_EXIT_INT(0);
	return 0;
}

 
