/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_info.c
 *
 * Description:
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"


/* Function: aix_get_volume_info
 *
 *	Fill in all the basic information about this volume.
 */
int aix_get_volume_info(storage_object_t         * region,
			extended_info_array_t   ** info_array )
{
	aix_region_data_t       * r_data = region->private_data;
	extended_info_array_t   * info = NULL;
	int                     i = 0;

	LOG_ENTRY();

	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*14)) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	// Region Name
	SET_STRING(info->info[i].name, "LV_Name");
	SET_STRING(info->info[i].title, "Region Name");
	SET_STRING(info->info[i].desc, "Name of AIX Region (LV)");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, region->name);
	i++;

	// Container Name
	SET_STRING(info->info[i].name, "VG_Name");
	SET_STRING(info->info[i].title, "Container Name");
	SET_STRING(info->info[i].desc, "Name of AIX Container (VG)");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, region->producing_container->name);
	i++;

	// LV Number
	SET_STRING(info->info[i].name, "LV_Number");
	SET_STRING(info->info[i].title, "Region Number");
	SET_STRING(info->info[i].desc, "ID number for this region in this container");
	info->info[i].type = EVMS_Type_Int;
	info->info[i].value.i = r_data->lv->lvname;
	i++;

	// LV Size
	SET_STRING(info->info[i].name, "LV_Size");
	SET_STRING(info->info[i].title, "Region Size");
	SET_STRING(info->info[i].desc, "Usable space for this region in sectors");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = region->size;
	info->info[i].flags |= EVMS_EINFO_FLAGS_NO_UNIT_CONVERSION;
	i++;

	// Allocated LEs
	SET_STRING(info->info[i].name, "Extents");
	SET_STRING(info->info[i].title, "Logical Extents");
	SET_STRING(info->info[i].desc, "Number of logical extents used by this region");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = r_data->lv->num_lps;
	info->info[i].flags = EVMS_EINFO_FLAGS_MORE_INFO_AVAILABLE;
	i++;

	// LV Status
	SET_STRING(info->info[i].name, "LV_Status");
	SET_STRING(info->info[i].title, "Region Status");
	info->info[i].type = EVMS_Type_String;
	if ( r_data->lv->lv_state & LVM_PVACTIVE ) {
		SET_STRING(info->info[i].value.s, "Available");
	}
	else {
		SET_STRING(info->info[i].value.s, "Unavailable");
	}
	i++;

	// Allocation Scheme
	SET_STRING(info->info[i].name, "Allocation");
	SET_STRING(info->info[i].title, "Allocation Policy");
	SET_STRING(info->info[i].desc, "How extents are allocated to this region");
	info->info[i].type = EVMS_Type_String;
//	if ( region->flags & LV_CONTIGUOUS ) {
//		SET_STRING(info->info[i].value.s, "Contiguous");
//	}
//	else {
	SET_STRING(info->info[i].value.s, "Next-free");
//	}
	i++;

	// Write Permissions
	SET_STRING(info->info[i].name, "Permissions");
	SET_STRING(info->info[i].title, "Access Permissions");
	info->info[i].type = EVMS_Type_String;
	if ( r_data->lv->lv_state & LVM_RDWR && r_data->lv->lv_state & LVM_RDONLY ) {
		SET_STRING(info->info[i].value.s, "Read-Write");
	}
	else if ( r_data->lv->lv_state & LVM_RDWR ) {
		SET_STRING(info->info[i].value.s, "Write-Only");
	}
	else if ( r_data->lv->lv_state & LVM_RDONLY ) {
		SET_STRING(info->info[i].value.s, "Read-Only");
	}
	else {
		SET_STRING(info->info[i].value.s, "No-Access");
	}
	i++;

	// Striping Info
	if ( r_data->lv->striping_width > 1 ) {
		// Number of Stripes
		SET_STRING(info->info[i].name, "Stripes");
		SET_STRING(info->info[i].title, "Stripes");
		SET_STRING(info->info[i].desc, "Number of objects this region is striped across");
		info->info[i].type = EVMS_Type_Unsigned_Int32;
		info->info[i].value.ui32 = r_data->lv->striping_width;
		i++;

		// Stripe Size
		SET_STRING(info->info[i].name, "Stripe_Size");
		SET_STRING(info->info[i].title, "Stripe Size");
		info->info[i].type = EVMS_Type_Unsigned_Int32;
		info->info[i].unit = EVMS_Unit_Sectors;
		info->info[i].value.ui32 = PP_SIZE_TO_VSECTORS(r_data->lv->stripe_exp);
		i++;
	}

	// Minor Number
	if ( region->volume ) {
		SET_STRING(info->info[i].name, "Minor_Number");
		SET_STRING(info->info[i].title, "Device Minor Number");
		info->info[i].type = EVMS_Type_Unsigned_Int32;
		info->info[i].value.ui32 = region->volume->dev_minor;
		i++;
	}

	if ( r_data->flags & AIX_REGION_INCOMPLETE ) {
		SET_STRING(info->info[i].name, "Incomplete_LV");
		SET_STRING(info->info[i].title, "INCOMPLETE REGION!!!");
		SET_STRING(info->info[i].desc, "This region has extents which are not mapped! One or more objects are probably missing from this container!");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, "");
		i++;
	}

	info->count = i;
	*info_array = info;

	LOG_EXIT_INT(0);
	return 0;
}


static inline int is_next_le_consecutive(storage_object_t * region,
					 int                     this_le )
{

	aix_region_data_t       * r_data = region->private_data;

	LOG_DEBUG("--------------------------------\n");
	LOG_DEBUG("r_data->lp_map[this_le].object:%p \n", r_data->lp_map[this_le].object);
	LOG_DEBUG("r_data->lp_map[this_le+1].object:%p \n", r_data->lp_map[this_le+1].object);
	LOG_DEBUG("r_data->lp_map[this_le].pp_number +1 :%ld \n", r_data->lp_map[this_le].pp_number+1);
	LOG_DEBUG("r_data->lp_map[this_le+1].pp_number:%ld \n", r_data->lp_map[this_le+1].pp_number);
	LOG_DEBUG("--------------------------------\n");

	if (r_data->lp_map[this_le+1].object != NULL) {
		if (( r_data->lp_map[this_le].object == r_data->lp_map[this_le+1].object) &&
		    ((r_data->lp_map[this_le].pp_number + 1) == r_data->lp_map[this_le+1].pp_number )) {
			LOG_DEBUG("return TRUE\n");
			return TRUE;
		} else {
			LOG_DEBUG("return FALSE\n");
			return FALSE;
		}
	} else
		LOG_DEBUG("return FALSE\n");
	return FALSE;
}


/* Function: aix_get_volume_extent_info
 *
 *	Fill in the "extra" extended info about the LE-to-PE mapping for
 *	this volume.
 */
int aix_get_volume_extent_info( storage_object_t   * region,
				extended_info_array_t   ** info_array )
{
	aix_region_data_t       * r_data = region->private_data;
	extended_info_array_t   * info = NULL;
	char                    buffer[150] = {0};
	int                     printed_dots = FALSE;
	int                     consecutive_run = FALSE;
	int                     info_line = 0;
	int                     j, i = 0;

	LOG_ENTRY();

	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t))) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	// Extent mappings
	SET_STRING(info->info[i].name, "Extent_Map");
	SET_STRING(info->info[i].title, "Logical Extents");
	SET_STRING(info->info[i].desc, "LE Number : PV Name : PE Number");
	info->info[i].type = EVMS_Type_String;
	info->info[i].collection_type = EVMS_Collection_List;
	info->info[i].collection.list = EngFncs->engine_alloc(sizeof(value_list_t) + sizeof(value_t)*(r_data ->lv->num_lps));
	snprintf(buffer, 150, "%-5s : %-15s : %-5s", "LE #", "PV Name", "PE #");
	SET_STRING(info->info[i].collection.list->value[info_line].s, buffer);
	info_line++;

	// Print info for each LE. Condense runs of consecutive LEs into
	// the first and last, with a ... in between.
	for ( j = 0; j < r_data->lv->num_lps; j++ ) {
		if ( ! r_data->lp_map[j].object ) {
			snprintf(buffer, 150, "%-5d : %-15s : %-5s", j, "Missing PV", "");
			consecutive_run = FALSE;
		}
		// Are we at the end of the LV or at the end of a run
		// of consecutive LEs?
		else if ( j == r_data->lv->num_lps - 1 || ! is_next_le_consecutive(region, j) ) {
			snprintf(buffer, 150, "%-5d :%-15s : %-5ld", j, r_data->lp_map[j].object->name, r_data->lp_map[j].pp_number);
			consecutive_run = FALSE;
		}
		// Are we at the start of the LV or at the start of a
		// run of consecutive LEs?
		else if ( ! consecutive_run ) {
			snprintf(buffer, 150, "%-5d :%-15s : %-5ld", j, r_data->lp_map[j].object->name, r_data->lp_map[j].pp_number);
			consecutive_run = TRUE;
			printed_dots = FALSE;
		}
		// Have we printed the ... yet?
		else if ( ! printed_dots ) {
			snprintf(buffer, 150, ".....");
			printed_dots = TRUE;
		}
		else {
			continue;
		}

		LOG_DEBUG("r_data->lp_map[j].object->name:%s\n", r_data->lp_map[j].object->name);

		SET_STRING(info->info[i].collection.list->value[info_line].s, buffer);
		info_line++;
	}
	info->info[i].collection.list->count = info_line;
	i++;

	info->count = i;
	*info_array = info;

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_get_group_info
 */
int aix_get_group_info( storage_container_t     * container,
			extended_info_array_t   ** info_array )
{
	extended_info_array_t   * info = NULL;
	aix_container_data_t    * c_data = container->private_data;
	int                     i = 0;

	LOG_ENTRY();

	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*12)) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	// Container Name
	SET_STRING(info->info[i].name, "VG_Name");
	SET_STRING(info->info[i].title, "Container Name");
	SET_STRING(info->info[i].desc, "Name of AIX Container (VG)");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, container->name);
	i++;

	// VG Number
	SET_STRING(info->info[i].name, "VG_Number");
	SET_STRING(info->info[i].title, "Container Number");
	SET_STRING(info->info[i].desc, "ID number for this container in the aix plugin");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.i32 = c_data->vgda->vg_head->vg_id.word2;
	i++;

	// VG Size
	SET_STRING(info->info[i].name, "VG_Size");
	SET_STRING(info->info[i].title, "Container Size");
	SET_STRING(info->info[i].desc, "Total accumulated space in this container");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = container->size;
	i++;

	// Available VG Size
	SET_STRING(info->info[i].name, "VG_Free_Size");
	SET_STRING(info->info[i].title, "Available Size");
	SET_STRING(info->info[i].desc, "Amount of space currently available for allocating to regions");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = c_data->freespace->size;
	i++;

	// Percent Allocated
	SET_STRING(info->info[i].name, "VG_Percent_Allocated");
	SET_STRING(info->info[i].title, "Percent Allocated");
	SET_STRING(info->info[i].desc, "Percentage of space currently allocated to regions");
	info->info[i].type = EVMS_Type_Real32;
	info->info[i].unit = EVMS_Unit_Percent;
	info->info[i].value.r32 = (100 - (((float)(c_data->freespace->size) / (float)(container->size)) * 100.0));
	i++;

	// PE Size
	SET_STRING(info->info[i].name, "PE_Size");
	SET_STRING(info->info[i].title, "Extent Size");
	SET_STRING(info->info[i].desc, "Size of each extent available for allocating to regions");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	i++;

	// Total PEs
	SET_STRING(info->info[i].name, "Total_PEs");
	SET_STRING(info->info[i].title, "Total PEs");
	SET_STRING(info->info[i].desc, "Total number of extents in the container");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->pp_count;
	i++;

	// Available PEs
	SET_STRING(info->info[i].name, "Available_PEs");
	SET_STRING(info->info[i].title, "Available PEs");
	SET_STRING(info->info[i].desc, "Number of extents available for allocating to regions");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->pp_count - (c_data->pp_count - (c_data->freespace->size / PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size)));
	i++;

	// Number of PVs
	SET_STRING(info->info[i].name, "Current_PVs");
	SET_STRING(info->info[i].title, "Number of Objects (PVs)");
	SET_STRING(info->info[i].desc, "Number of objects in this container");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->vgda->vg_head->numpvs;
	info->info[i].flags = EVMS_EINFO_FLAGS_MORE_INFO_AVAILABLE;
	i++;

	// Number of LVs
	SET_STRING(info->info[i].name, "Current_LVs");
	SET_STRING(info->info[i].title, "Number of Regions (LVs)");
	SET_STRING(info->info[i].desc, "Number of regions in this container");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->vgda->vg_head->numlvs;
	info->info[i].flags = EVMS_EINFO_FLAGS_MORE_INFO_AVAILABLE;
	i++;

	// Max LV Size
	SET_STRING(info->info[i].name, "Max_LV_Size");
	SET_STRING(info->info[i].title, "Maximum Region Size");
	SET_STRING(info->info[i].desc, "Maximum possible size of any region in this container for the given PE size");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = c_data->vgda->vg_head->pp_size * 64 * 1024;
	i++;

	if ( c_data->flags & LVM_PVMISSING ) {
		SET_STRING(info->info[i].name, "Incomplete_VG");
		SET_STRING(info->info[i].title, "INCOMPLETE CONTAINER!!!");
		SET_STRING(info->info[i].desc, "This container seems to be missing one or more objects!");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, "");
		i++;
	}

	info->count = i;
	*info_array = info;

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_get_group_pv_list_info
 */
int aix_get_group_pv_list_info( storage_container_t     * container,
				extended_info_array_t   ** info_array )
{
	extended_info_array_t   * info = NULL;
	aix_container_data_t    * c_data = container->private_data;
	char                    buf[50] = {0};
	int                     j, i = 0;

	LOG_ENTRY();

	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*c_data->vgda->vg_head->numpvs)) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	for ( j = 0; j < AIX_MAX_PVS; j++ ) {
		if ( c_data->vgda->pv_headers[j]) {
			sprintf(buf, "PV%d", j);
			SET_STRING(info->info[i].name, buf);
			sprintf(buf, "Object (PV) %d", j);
			SET_STRING(info->info[i].title, buf);
			info->info[i].type = EVMS_Type_Int16;
			info->info[i].value.i16 = c_data->vgda->pv_headers[j]->pv_num;
			info->info[i].flags = EVMS_EINFO_FLAGS_MORE_INFO_AVAILABLE;
			i++;
		}
	}

	info->count = i;
	*info_array = info;

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_get_group_lv_list_info
 */
int aix_get_group_lv_list_info( storage_container_t     * container,
				extended_info_array_t   ** info_array )
{
	extended_info_array_t   * info = NULL;
	aix_container_data_t    * c_data = container->private_data;
	char                    buf[50] = {0};
	int                     j, i = 0;

	LOG_ENTRY();

	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*c_data->vgda->vg_head->numlvs)) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	for ( j = 0; j < AIX_MAX_LVS; j++ ) {
		if (c_data->regions[j]) {
			if ( c_data->regions[j]->volume ) {
				sprintf(buf, "LV%d", j);
				SET_STRING(info->info[i].name, buf);
				sprintf(buf, "Region (LV) %d", j);
				SET_STRING(info->info[i].title, buf);
				info->info[i].type = EVMS_Type_String;
				SET_STRING(info->info[i].value.s, c_data->regions[j]->name);
				i++;
			}
		}
	}

	info->count = i;
	*info_array = info;

	LOG_EXIT_INT(0);
	return 0;
}


/* Function: aix_get_pv_info
 */
int aix_get_pv_info(    storage_container_t     * container,
			extended_info_array_t   ** info_array )
{
	extended_info_array_t   * info = NULL;
	aix_container_data_t    * c_data = container->private_data;
//	char			* buf1 = NULL;
	char                    buf2[50] = {0};
	int                     i = 0;

	LOG_ENTRY();

	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*9)) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}
/*
	// PV Name
	SET_STRING(info->info[i].name, "PV_Name");
	SET_STRING(info->info[i].title, "Object Name");
	SET_STRING(info->info[i].desc, "Name of AIX Object (PV)");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, container->name);
	i++;
*/
	// VG Name
	SET_STRING(info->info[i].name, "VG_Name");
	SET_STRING(info->info[i].title, "Container Name");
	SET_STRING(info->info[i].desc, "Name of AIX Container (VG)");
	info->info[i].type = EVMS_Type_String;
	SET_STRING(info->info[i].value.s, container->name);
	i++;

	// PV Number
	SET_STRING(info->info[i].name, "PV_Number");
	SET_STRING(info->info[i].title, "Object Number");
	SET_STRING(info->info[i].desc, "ID number for this object (PV) in this container");
	info->info[i].type = EVMS_Type_Int;
	info->info[i].value.i = c_data->vgda->pv_headers[0]->pv_id.word2;
	i++;

	// PV Size
	SET_STRING(info->info[i].name, "PV_Size");
	SET_STRING(info->info[i].title, "Object Size");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = container->size;
	i++;

	// Usable Size
	SET_STRING(info->info[i].name, "Usable_PV_Size");
	SET_STRING(info->info[i].title, "Usable Space");
	SET_STRING(info->info[i].desc, "PE Size * Number of PEs. Some object space is used for metadata, some is unusable due to PE size");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size) * c_data->pp_count;

	i++;

	// Number of Regions
	SET_STRING(info->info[i].name, "Current_Regions");
	SET_STRING(info->info[i].title, "Current Regions");
	SET_STRING(info->info[i].desc, "Number of regions currently using space on this object");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->vgda->vg_head->numlvs;
	i++;

	// PE Size
	SET_STRING(info->info[i].name, "PE_Size");
	SET_STRING(info->info[i].title, "Extent Size");
	SET_STRING(info->info[i].desc, "Size of each extent available for allocating to regions");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].unit = EVMS_Unit_Sectors;
	info->info[i].value.ui32 = PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size);
	i++;

	// Total PEs
	sprintf(buf2, "PEMapPV%d", c_data->vgda->pv_headers[0]->pv_id.word2);
	SET_STRING(info->info[i].name, buf2);
	SET_STRING(info->info[i].title, "Total PEs");
	SET_STRING(info->info[i].desc, "Total number of extents in this object");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->pp_count;
	info->info[i].flags = 0;
	i++;

	// Available PEs
	SET_STRING(info->info[i].name, "Available_PEs");
	SET_STRING(info->info[i].title, "Available PEs");
	SET_STRING(info->info[i].desc, "Number of extents available for allocating to regions");
	info->info[i].type = EVMS_Type_Unsigned_Int32;
	info->info[i].value.ui32 = c_data->pp_count - (c_data->pp_count - (c_data->freespace->size / PP_SIZE_TO_VSECTORS(c_data->vgda->vg_head->pp_size)));
	i++;


	info->count = i;
	*info_array = info;
	LOG_EXIT_INT(0);
	return 0;
}


static inline int is_next_pe_consecutive(aix_pv_data_t  * pv_entry,
					 int                     this_pe )
{
	if ( pv_entry->pp_map[this_pe].lp_num == pv_entry->pp_map[this_pe+1].lp_num ) {
		if ( pv_entry->pp_map[this_pe].lp_num ) {
			if ( pv_entry->pp_map[this_pe].lp_num + 1 == pv_entry->pp_map[this_pe+1].lp_num ) {
				return TRUE;
			}
			else {
				return FALSE;
			}
		}
		else {
			return TRUE;
		}
	}
	else {
		return FALSE;
	}
}

/* Function: aix_get_pv_extent_info
 */
int aix_get_pv_extent_info(     storage_container_t     * container,
				extended_info_array_t   ** info_array )
{
	extended_info_array_t   * info = NULL;
//	char			buffer[150] = {0};
//	int			consecutive_run = FALSE;
//	int			printed_dots = FALSE;
//	int			info_line = 1;
//	int			j,i = 0;

	LOG_ENTRY();


	// Get memory for the info array
	if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t))) ) {
		LOG_CRITICAL("Memory error creating info array\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}
/*
	// Extent mappings
	SET_STRING(info->info[i].name, "Extent_Map");
	SET_STRING(info->info[i].title, "Physical Extents");
	SET_STRING(info->info[i].desc, "PE Number : LV Name : LE Number: Sector");
	info->info[i].type = EVMS_Type_String;
	info->info[i].collection_type = EVMS_Collection_List;
	info->info[i].collection.list = EngFncs->engine_alloc(sizeof(value_list_t) + sizeof(value_t)*(pv_entry->pv->pe_total));
	snprintf(buffer, 150, "%-5s : %-20s : %-5s : %-5s", "PE #", "LV Name", "LE #", "Sector");
	SET_STRING(info->info[i].collection.list->value[0].s, buffer);

	// Print info for each PE. Condense runs of consecutive PEs into
	// the first and last, with a ... in between.
	for ( j = 0; j < pv_entry->pv->pe_total; j++ ) {
		if ( pv_entry->group->volume_list[pv_entry->pe_map[j].lv_num] ) {
			// PE is mapped to a volume.
			if ( j == pv_entry->pv->pe_total ||
			     ! is_next_pe_consecutive(pv_entry, j) ) {
				snprintf(buffer, 150, "%-5d : %-20s : %-5d : %-5d",
					j, pv_entry->group->volume_list[pv_entry->pe_map[j].lv_num]->region->name,
					pv_entry->pe_map[j].le_num, pv_entry->pv->pe_start+pv_entry->pv->pe_size*j);
				consecutive_run = FALSE;
			}
			else if ( ! consecutive_run ) {
				snprintf(buffer, 150, "%-5d : %-20s : %-5d : %-5d",
					j, pv_entry->group->volume_list[pv_entry->pe_map[j].lv_num]->region->name,
					pv_entry->pe_map[j].le_num, pv_entry->pv->pe_start+pv_entry->pv->pe_size*j);
				consecutive_run = TRUE;
				printed_dots = FALSE;
			}
			else if ( ! printed_dots ) {
				snprintf(buffer, 150, ".....");
				printed_dots = TRUE;
			}
			else {
				continue;
			}
		}
		else {
			// PE is unallocated.
			if ( j == pv_entry->pv->pe_total - 1 ||
			     ! is_next_pe_consecutive(pv_entry, j) ) {
				snprintf(buffer, 150, "%-5d : %-20s : %-5s : %-5d",
					j, "none", "n/a", pv_entry->pv->pe_start+pv_entry->pv->pe_size*j);
				consecutive_run = FALSE;
			}
			else if ( ! consecutive_run ) {
				snprintf(buffer, 150, "%-5d : %-20s : %-5s : %-5d",
					j, "none", "n/a", pv_entry->pv->pe_start+pv_entry->pv->pe_size*j);
				consecutive_run = TRUE;
				printed_dots = FALSE;
			}
			else if ( ! printed_dots ) {
				snprintf(buffer, 150, ".....");
				printed_dots = TRUE;
			}
			else {
				continue;
			}
		}
		SET_STRING(info->info[i].collection.list->value[info_line].s, buffer);
		info_line++;
	}
	info->info[i].collection.list->count = info_line;
	i++;

	info->count = i;
	*info_array = info;

*/
	LOG_EXIT_INT(0);
	return 0;
}

int AIX_dump_expand_region(storage_object_t * region, int lineNum)
{
	aix_container_data_t  * c_data = region->producing_container->private_data;
	aix_region_data_t     * f_data = c_data->freespace->private_data;
	aix_region_data_t     * r_data = region->private_data;
	storage_object_t      * object;
	aix_vgda_t            * vgda   = c_data->vgda;
	pp_entries          * pp_map;
	pv_header           * pv_head;
	int i = 0, j, Found, rc;

	LOG_ENTRY();
	LOG_DEBUG("******************* Called from Line %d ****************************\n", lineNum);


	if (f_data) {
		LOG_DEBUG("Freespace num_lps:%d \n", f_data->lv->num_lps);

		for (i = 0, Found = 0; i < AIX_MAX_LVS && Found < f_data->lv->num_lps; i++) {
			if (f_data->lp_map) {
				if (f_data->lp_map[i].object != NULL) {
					LOG_DEBUG("f_data object:%s pp_number:%ld \n", f_data->lp_map[i].object->name, f_data->lp_map[i].pp_number);
					Found ++;
				}
			}
		}
	}

	if (c_data) {
		LOG_DEBUG("c_data:%s  pp_count:%ld numlvs:%d\n", region->producing_container->name, c_data->pp_count, c_data->vgda->vg_head->numlvs);

		for (i = 0, Found = 0; i < AIX_MAX_PVS && Found < c_data->vgda->vg_head->numpvs; i++) {
			if (c_data->vgda->pv_headers[i]) {
				LOG_DEBUG("c_data pv_num:%d pp_count:%d \n", c_data->vgda->pv_headers[i]->pv_num, c_data->vgda->pv_headers[i]->pp_count);

				Found++;
			}
		}
	}


	if (r_data) {
		LOG_DEBUG("r_data expand region name:%s lv_name:%d num_lps:%d\n", region->name, r_data->lv->lvname, r_data->lv->num_lps);

		for (i = 0, Found = 0; i < AIX_MAX_LVS && Found < r_data->lv->num_lps / r_data->lv->mirror; i++) {
			if (r_data->lp_map) {
				LOG_DEBUG("r_data object:%s lp_num:%d pp_number:%ld \n", r_data->lp_map[i].object->name, i, r_data->lp_map[i].pp_number);
				Found++;
			}
		}
	}

	LOG_DEBUG("Dumping the entire container region info: \n");

	DLIST_FOR_EACH(object, region->producing_container->objects_produced, rc) {
		r_data = object->private_data;


		LOG_DEBUG("** Region name:%s \n", object->name);

		if (r_data && object->data_type == DATA_TYPE) {
			if (r_data->lv) {
				LOG_DEBUG("lv_name:%d num_lps:%d\n", r_data->lv->lvname, r_data->lv->num_lps);

				for (i = 0, Found = 0; i < AIX_MAX_LVS && Found < r_data->lv->num_lps; i++) {
					if (r_data->lp_map) {
						LOG_DEBUG("r_data object:%s lp_num:%d pp_number:%ld \n", r_data->lp_map[i].object->name, i, r_data->lp_map[i].pp_number);
						Found++;
					}
				}

			}
		}

	}

	LOG_DEBUG("Dumping the entire container pp_maps info: \n");

	for (i = 0; i < AIX_MAX_PVS; i++) {
		if (vgda->pv_headers[i] && vgda->pp_array[i]) {

			pv_head = vgda->pv_headers[i];
			pp_map = vgda->pp_array[i];

			LOG_DEBUG("****** Dumping PV:%d ******\n",pv_head->pv_num);

			for (j = 0; j < pv_head->pp_count; j++) {

				LOG_DEBUG("lp_num:%d lv_index:%d pp_state:%d fap:%d fav:%d sap:%d sav:%d \n",
					  pp_map[j].lp_num-1,
					  pp_map[j].lv_index,
					  pp_map[j].pp_state,
					  pp_map[j].fst_alt_part,
					  pp_map[j].fst_alt_vol,
					  pp_map[j].snd_alt_part,
					  pp_map[j].snd_alt_vol);

			}
		}
	}




	LOG_DEBUG("***********************************************\n");
	LOG_EXIT_INT(0);
	return 0;

}
int aix_dump_region(storage_object_t * region, int lineNum)
{

	aix_region_data_t     * r_data = region->private_data;
	storage_object_t      * object;
	int i, Found, rc;

	LOG_ENTRY();

	LOG_DEBUG("******************* Called from Line %d ****************************\n", lineNum);

	if (r_data) {
		LOG_DEBUG("r_data expand region name:%s lv_name:%d num_lps:%d\n", region->name, r_data->lv->lvname, r_data->lv->num_lps);

		for (i = 0, Found = 0; i < AIX_MAX_LVS && Found < r_data->lv->num_lps; i++) {
			if (r_data->lp_map) {
				LOG_DEBUG("r_data object:%s lp_num:%d pp_number:%ld \n", r_data->lp_map[i].object->name, i, r_data->lp_map[i].pp_number);
				Found++;
			}
		}
	}

	LOG_DEBUG("Dumping the entire container region info: \n");

	DLIST_FOR_EACH(object, region->producing_container->objects_produced, rc) {
		r_data = object->private_data;


		LOG_DEBUG("** Region name:%s \n", object->name);

		if (r_data && object->data_type == DATA_TYPE) {
			if (r_data->lv) {
				LOG_DEBUG("lv_name:%d num_lps:%d\n", r_data->lv->lvname, r_data->lv->num_lps);

				for (i = 0, Found = 0; i < AIX_MAX_LVS && Found < r_data->lv->num_lps; i++) {
					if (r_data->lp_map) {
						LOG_DEBUG("r_data object:%s lp_num:%d pp_number:%ld \n", r_data->lp_map[i].object->name, i, r_data->lp_map[i].pp_number);
						Found++;
					}
				}

			}
		}

	}

	LOG_EXIT_INT(0);
	return 0;
}
