/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_pv.c
 *
 * Description: This file contains functions related to managing AIX PVs.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "aixregmgr.h"


/* aix_find_pv_by_number
 *
 *	Search the objects_consumed list in this container for the PV with
 *	the specified number.
 */
storage_object_t * aix_find_pv_by_number(storage_container_t    * container,
					 unsigned int            pv_num )
{
	storage_object_t        * object;
	aix_pv_data_t           * pv;
	int                     rc;

	LOG_ENTRY();

	DLIST_FOR_EACH(object, container->objects_consumed, rc) {
		pv = object->consuming_private_data;
		if (pv->pv_head) {
			if (pv->pv_head->pv_num == pv_num) {
				LOG_EXIT_PTR(object);
				return object;
			}
		}
	}

	LOG_EXIT_PTR(NULL);
	return NULL;
}

/* aix_deallocate_pv
 *
 *	Free up all memory used by the specified PV.
 */
void aix_deallocate_pv( storage_object_t * object )
{
	aix_pv_data_t * pv = object->consuming_private_data;

	LOG_ENTRY();

	if ( pv ) {
		object->consuming_private_data = NULL;

		if ( pv->ipl ) {
			EngFncs->engine_free(pv->ipl);
		}

		if ( pv->lvm ) {
			EngFncs->engine_free(pv->lvm);
		}

		EngFncs->engine_free(pv);
	}

	LOG_EXIT_VOID();
}


/* aix_allocate_pv
 *
 *	Allocate and initialize a PV structure for the specified object.
 */
int aix_allocate_pv(    storage_object_t        * object,
			aix_ipl_rec_t           * ipl,
			aix_lvm_rec_t           * lvm )
{
	aix_pv_data_t   * new_pv;

	LOG_ENTRY();

	// Allocate the pv structure
	new_pv = EngFncs->engine_alloc(sizeof(aix_pv_data_t));
	if ( ! new_pv ) {
		LOG_CRITICAL("Memory error creating PV structure for %s.\n", object->name);
		EngFncs->engine_free(ipl);
		EngFncs->engine_free(lvm);
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	new_pv->ipl             = ipl;
	new_pv->lvm             = lvm;

	object->consuming_private_data = new_pv;

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_compare_timestamps
 *
 *	Compare the various time-stamps in the VG headers and trailers for the
 *	specified PV to determine which VGDA is active.
 */
int aix_compare_timestamps( storage_object_t * object )
{
	aix_pv_data_t * pv = object->consuming_private_data;

	LOG_ENTRY();

	if ( COMPARE_TIMESTAMPS(pv->vg_head[0]->vg_timestamp, pv->vg_tail[0]->timestamp) ) {
		if ( COMPARE_TIMESTAMPS(pv->vg_head[1]->vg_timestamp, pv->vg_tail[1]->timestamp) ) {
			if ( COMPARE_TIMESTAMPS(pv->vg_head[0]->vg_timestamp, pv->vg_head[1]->vg_timestamp) ) {
				// All timestamps match. Yea!
				pv->pv_state = AIX_PV_STATE_VALID;
			}
			else {
				// Both VGDAs are good, but timestamps are
				// different. Can't tell yet which one is
				// correct.
				pv->pv_state = AIX_PV_STATE_EITHER_VGDA;
			}
		}
		else {
			// First VGDA is good, second is bad.
			pv->pv_state = AIX_PV_STATE_FIRST_VGDA;
		}
	}
	else {
		if ( COMPARE_TIMESTAMPS(pv->vg_head[1]->vg_timestamp, pv->vg_tail[1]->timestamp) ) {
			// First VGDA is bad, second is good.
			pv->pv_state = AIX_PV_STATE_SECOND_VGDA;
		}
		else {
			// This should never happen.
			LOG_CRITICAL("All four VG timestamps for %s are different. What the hell?!?\n", object->name);
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_read_pv_metadata
 *
 *	Read the AIX metadata from the specified object. Check for appropriate
 *	signatures. If found, read both VG headers and both VG trailers, and
 *	make an initial comparison of the timestamps to determine which VGDA
 *	should be used later.
 */
int aix_read_pv_metadata( storage_object_t * object )
{
	aix_ipl_rec_t   * ipl;
	aix_lvm_rec_t   * lvm;
	int             rc;

	LOG_ENTRY();

	// Read the IPL record from this object.
	rc = aix_read_ipl(object, &ipl);
	if (rc) {
		// This object is not an AIX PV.
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Read the LVM record from this object.
	rc = aix_read_lvm(object, &lvm);
	if (rc) {
		// This object is not managed by the AIX LVM.
		EngFncs->engine_free(ipl);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Allocate a PV data structure. This attaches the PV
	// private data to the specified object.
	rc = aix_allocate_pv(object, ipl, lvm);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Read the VG headers and trailers.
	rc = aix_read_vg_headers(object);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}
	rc = aix_read_vg_trailers(object);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Compare the timestamps to determine the state of the PV.
	rc = aix_compare_timestamps(object);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_check_object_for_container_inclusion
 *
 *	Check this object to make sure it is valid for adding to a container.
 */
int aix_check_object_for_container_inclusion( storage_object_t * object )
{
	int rc = 0;

	LOG_ENTRY();

	// Is this object a disk, segment, or region?
	if ( ! (object->object_type & (DISK|SEGMENT|REGION)) ) {
		LOG_WARNING("%s is not a disk, segment, or region.\n", object->name);
		rc = EINVAL;
	}

	// Is this object already in a container?
	else if ( object->consuming_container ) {
		LOG_WARNING("%s is already in a container.\n", object->name);
		rc = EINVAL;
	}

	// Does this object have any parent objects?
	else if ( ! ListEmpty(object->parent_objects) ) {
		LOG_WARNING("%s has parent objects.\n", object->name);
		rc = EINVAL;
	}

	// Is this object a compatibility volume?
	if ( object->volume ) {
		LOG_WARNING("%s is in use by volume %s. Must remove the volume first.\n",
			    object->name, object->volume->name);
		rc = EINVAL;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_check_object_for_pp_size
 *
 *	Check that the given object is large enough to use the specified PP
 *	size, and also not too large to waste disk space.
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
int aix_check_object_for_pp_size(storage_object_t       * object,
				 unsigned long           * pp_size )
{
	sector_count_t  num_sectors;
	unsigned long   num_pps;

	LOG_ENTRY();



	// Calculate the number of allocatable sectors on this object.
	num_sectors = AVAILABLE_PP_SECTORS(object);

	// Calculate the number of PPs that could be allocated on this object.
	if (*(pp_size) == 0) {
		LOG_WARNING("%s PP size is 0 %ld.\n", object->name, *pp_size);
		LOG_EXIT_INT(EINVAL);
		return EINVAL;
	}
	num_pps = num_sectors / *pp_size;

	// The object must be large enough to hold at least one PP.
	if ( num_pps < 1 ) {
		LOG_WARNING("%s is too small for PP size %ld.\n", object->name, *pp_size);
		*pp_size = num_sectors;
		aix_check_pp_size(pp_size);
		if (*pp_size > num_sectors) {
			LOG_WARNING("\t%s CANNOT be used as an AIX PV!\n", object->name);
			LOG_WARNING("\tOnly %"PRIu64" sectors available for PPs.\n", num_sectors);
			LOG_WARNING("\tSmaller than the minimum PP size: %d.\n", AIX_MIN_PP_SIZE);
			LOG_EXIT_INT(EINVAL);
			return EINVAL;
		}
		else {
			LOG_WARNING("\t%s has %"PRIu64" sectors available for PPs.\n", object->name, num_sectors);
			LOG_WARNING("\tCan handle a max PP size of %ld sectors.\n", *pp_size);
			LOG_EXIT_INT(ENOSPC);
			return ENOSPC;
		}
	}

	// The object must be small enough so it doesn't have more than the
	// max allowed number of PPs.
	else if ( num_pps > AIX_MAX_PPS_PER_PV ) {
		LOG_WARNING("%s is too large for PP size %ld\n", object->name, *pp_size);
		LOG_WARNING(" Would contain %ld PPs. Max PPs per PV is %d.\n", num_pps, AIX_MAX_PPS_PER_PV);
		*pp_size = num_sectors / AIX_MAX_PPS_PER_PV;
		aix_check_pp_size(pp_size);
		LOG_WARNING(" Needs a PP size of %ld sectors to be under max.\n", *pp_size);
		LOG_EXIT_INT(EFBIG);
		return EFBIG;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_find_free_pv_number
 *
 *	Search the existing list of PV headers in the specified container
 *	to find a free PV number.
 */
static unsigned int aix_find_free_pv_number( storage_container_t * container )
{
	aix_container_data_t    * c_data = container->private_data;
	unsigned int            i;

	LOG_ENTRY();

	for ( i = 0; i < AIX_MAX_PVS; i++ ) {
		if ( ! c_data->vgda->pv_headers[i] ) {
			LOG_EXIT_INT(i+1);
			return i+1;
		}
	}

	LOG_SERIOUS("No PV numbers available in container %s.\n", container->name);

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_initialize_ipl
 */
static int aix_initialize_ipl( aix_ipl_rec_t * ipl )
{
	int rc;

	LOG_ENTRY();

	memset(ipl, 0, sizeof(aix_ipl_rec_t));

	// Standard IPL signature.
	ipl->IPL_record_id = IPL_REC_ID;

	// Get a new Unique ID
	rc = aix_create_unique_id(&ipl->pv_id);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_initialize_lvm
 */
static int aix_initialize_lvm( aix_lvm_rec_t           * lvm,
			       storage_container_t     * container,
			       storage_object_t        * object )
{
	aix_container_data_t    * c_data = container->private_data;

	LOG_ENTRY();

	// Standard LVM signature and version.
	lvm->lvm_id             = AIX_LVM_LVMID;
	lvm->version            = LVM_VERSION_1;

	// VG information.
	lvm->vg_id              = c_data->vgda->vg_head->vg_id;
	lvm->pp_size            = c_data->vgda->vg_head->pp_size;	// Exponent form.

	// The lengths of the status area and descriptor area.
	lvm->vgsa_len           = AIX_DEFAULT_VGSA_SIZE;
	lvm->vgda_len           = AIX_DEFAULT_VGDA_SIZE;
	lvm->lvmarea_len        = AIX_MAX_VGDAS * (lvm->vgsa_len + lvm->vgda_len);

	// The sector locations of the status areas and descriptor areas.
	lvm->vgsa_psn[0]        = AIX_DEFAULT_VGSA_PSN;
	lvm->vgda_psn[0]        = lvm->vgsa_psn[0] + lvm->vgsa_len;
	lvm->vgsa_psn[1]        = lvm->vgda_psn[0] + lvm->vgda_len;
	lvm->vgda_psn[1]        = lvm->vgsa_psn[1] + lvm->vgsa_len;

	// The remap area for bad-block-relocation.
	lvm->reloc_len          = AIX_RELOC_LENGTH;
	lvm->reloc_psn          = object->size - lvm->reloc_len - 1;

	// Get a new PV number.
	lvm->pv_num = aix_find_free_pv_number(container);
	if ( ! lvm->pv_num ) {
		LOG_SERIOUS("Could not get PV number for %s\n", object->name);
		LOG_EXIT_INT(ENOSPC);
		return ENOSPC;
	}

	LOG_EXIT_INT(0);
	return 0;
}


/* aix_initialize_pv
 */
static int aix_initialize_pv( storage_object_t * object )
{
	aix_pv_data_t   * pv = object->consuming_private_data;
	sector_count_t  num_sectors = AVAILABLE_PP_SECTORS(object);

	LOG_ENTRY();

	pv->pp_map              = (pp_entries*)&(pv->pv_head[1]);
	pv->pv_head->pv_id      = pv->ipl->pv_id;
	pv->pv_head->pp_count   = num_sectors / PP_SIZE_TO_VSECTORS(pv->lvm->pp_size);
	pv->pv_head->pv_state   = LVM_PVACTIVE;
	pv->pv_head->psn_part1  = AIX_DEFAULT_FIRST_PP_PSN;
	pv->pv_head->pvnum_vgdas= AIX_MAX_VGDAS;
	pv->pv_head->pv_num     = pv->lvm->pv_num;

	LOG_EXIT_INT(0);
	return 0;
}

/* aix_create_pv_from_object
 *
 *	Create a new PV using the specified object. This will involve
 *	allocating a new PV header and PP map, and initializing them
 *	as appropriate.
 *
 *	Before calling this function, this object must first be checked to
 *	make sure it is not too big or too small for the container's PP size.
 */
int aix_create_pv_from_object(  storage_object_t        * object,
				storage_container_t     * container )
{
	aix_ipl_rec_t   * ipl;
	aix_lvm_rec_t   * lvm;
	aix_pv_data_t   * pv;
	int             rc;

	LOG_ENTRY();

	// Allocate new IPL and LVM structures.
	ipl = EngFncs->engine_alloc(sizeof(aix_ipl_rec_t));
	if ( ! ipl ) {
		MESSAGE("Memory error creating new IPL structure for %s\n", object->name);
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	lvm = EngFncs->engine_alloc(sizeof(aix_lvm_rec_t));
	if ( ! lvm ) {
		EngFncs->engine_free(ipl);
		MESSAGE("Memory error creating new LVM structure for %s\n", object->name);
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	// Initialize the IPL structure.
	rc = aix_initialize_ipl(ipl);
	if (rc) {
		EngFncs->engine_free(ipl);
		EngFncs->engine_free(lvm);
		MESSAGE("Error initializing IPL structure for %s.\n", object->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Initialize the LVM structure.
	rc = aix_initialize_lvm(lvm, container, object);
	if (rc) {
		EngFncs->engine_free(ipl);
		EngFncs->engine_free(lvm);
		MESSAGE("Error initializing LVM structure for %s.\n", object->name);
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Allocate a new PV.
	rc = aix_allocate_pv(object, ipl, lvm);
	if (rc) {
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Allocate a PV header and PP map.
	pv = object->consuming_private_data;
	pv->pv_head = EngFncs->engine_alloc(AIX_MAX_PV_SECTORS * EVMS_VSECTOR_SIZE);
	if ( ! pv->pv_head ) {
		aix_deallocate_pv(object);
		MESSAGE("Memory error creating new PV header for %s.\n", object->name);
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	aix_initialize_pv(object);

	LOG_EXIT_INT(0);
	return 0;
}


