/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libbsd.so
 *
 *   File: segments.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "bsd_plugin.h"


int bsd_can_set_volume( DISKSEG * seg, BOOLEAN flag )
{
        int rc = 0;

        LOGENTRY();

        REQUIRE( isa_bsd_segment(seg) == TRUE );

        LOGEXITRC();
        return rc;
}

int bsd_can_delete( DISKSEG * seg )
{
	int rc = EINVAL;

	LOGENTRY();

	REQUIRE( seg != NULL);
	REQUIRE( seg->object_type == SEGMENT );
	REQUIRE( seg->data_type   == DATA_TYPE );
	REQUIRE( disk_move_pending(seg) == FALSE);
	REQUIRE( isa_bsd_segment( seg ) == TRUE );

	rc = 0;

	LOGEXITRC();
	return rc;
}


int bsd_can_expand( DISKSEG         *seg,
                    sector_count_t  *expand_limit,
                    dlist_t          expansion_points )
{
        int rc=ENOSYS;

        LOGENTRY();
        LOGEXITRC();
        return rc;
}

int bsd_can_expand_by( DISKSEG * seg, sector_count_t *size)
{
        int rc=0;

        LOGENTRY();
        *size = 0;
        LOGEXITRC();
        return rc;
}


int bsd_can_shrink( DISKSEG * seg,
                    sector_count_t   * shrink_limit,
                    dlist_t            shrink_points )
{
        int rc=ENOSYS;

        LOGENTRY();
        LOGEXITRC();
        return rc;
}


int bsd_can_shrink_by( DISKSEG * seg, sector_count_t * size )
{
        int rc=0;

        LOGENTRY();
        *size=0;
        LOGEXITRC();
        return rc;
}


/*
 *  Function:  bsd_discover
 *
 *  Called to run discovery on a dlist of evms objects that
 *  the engine has found thus far.  We walk the list of objects,
 *  probing each object to see if it has a BSD disk label.  If so,
 *  we consume the object by removing it from the dlist, and place
 *  all new segment storage objects on the output object list.
 *  Any object we dont like in the input_object dlist must be
 *  copied to the output_object dlist.
 *
 */
int bsd_discover( dlist_t input_objects, dlist_t output_objects, BOOLEAN final_call)
{
        int  count = 0;
        discovery_parm_block_t  dparms = {output_objects, &count };

        LOGENTRY();

        if ( (input_objects != NULL) &&
             (output_objects != NULL) &&
             (list_item_count(input_objects) > 0)) {
 
		ForEachItem( input_objects, discover_bsd_segments, (void *) &dparms, TRUE );

        }

        LOGEXIT();
        return count;
}


int bsd_assign( storage_object_t * input_object, option_array_t * options )
{
        int rc;

        LOGENTRY();

        REQUIRE( input_object != NULL );
        REQUIRE( options != NULL);

        rc = ENOSYS;

        LOGEXITRC();
        return rc;
}


/*
 *  Function:  bsd_unassign
 *
 *  We can unassign from any logical disk storage object that
 *  has a BSD disk label on it as long as the data segments are 
 *  top-most storage objects.
 *
 */
int bsd_can_unassign( LOGICALDISK *ld )
{
        int rc=0;
        DISKSEG *seg;

        LOGENTRY();

        REQUIRE( isa_bsd_logical_disk(ld) == TRUE );

        FOR_EACH(seg, ld->parent_objects) {
                REQUIRE( list_item_count(seg->parent_objects) == 0 );
        }

        rc = 0;

        LOGEXITRC();
        return rc;
}


/*
 *  Function:  bsd_unassign
 *
 *  This routine is called to remove the BSD Segment Manager
 *  from the specified storage object (logical disk).  It will do
 *  so by:
 *
 *      - removing the BSD disk label from the storage object by writing
 *        kill sectors (zero filled sectors) to appropriate locations
 *        on the logical disk ... erasing the metadata found at
 *        those locations.
 *
 */
int bsd_unassign( LOGICALDISK *ld )
{
        int rc=0;

        LOGENTRY();

        REQUIRE( isa_bsd_logical_disk(ld) == TRUE );
  
        // toss segment storage objects
        PruneList(ld->parent_objects, prune_bsd_seg_objects_from_list, NULL);

        // toss our logical disk private data area
        delete_bsd_disk_private_data( ld );

        // blast bsd disk label 
        KILL_SECTORS(ld, 0, 1);

        LOGEXITRC();
        return rc;
}


int bsd_create( dlist_t          input_objects,
                option_array_t * options,
                dlist_t          new_objects)
{
        int rc=ENOSYS;

        LOGENTRY();
        LOGEXITRC();
        return rc;
}


int bsd_destroy( DISKSEG * seg, dlist_t child_objects )
{
	int           rc = EINVAL;
	DISKSEG      *metadata=NULL;
	DISKSEG      *md=NULL;
	LOGICALDISK  *ld=NULL;
	disk_private_data_t  *disk_pdata=NULL;

	LOGENTRY();

	LOG_DEBUG("seg: %s\n", seg->name );

	REQUIRE( seg != NULL);
	REQUIRE( seg->object_type == SEGMENT );
	REQUIRE( seg->data_type   == DATA_TYPE );
	REQUIRE( disk_move_pending(seg) == FALSE);
	REQUIRE( isa_bsd_segment( seg ) == TRUE );
 
	ld         = get_logical_disk(seg);
	REQUIRE( ld != NULL);

	disk_pdata = get_bsd_disk_private_data(ld);
	REQUIRE( disk_pdata != NULL);

	FOR_EACH(md, ld->parent_objects){
		if (md->data_type == META_DATA_TYPE) {
			metadata = md;
			break;
		}
	}

	if (metadata) {
		
		rc = remove_bsd_segment_from_list( ld->parent_objects, seg );
		if (rc == 0) {

			free_bsd_segment(seg);

			//find_freespace_on_bsd_disk( ld );

			disk_pdata->flags |= DISK_HAS_CHANGES_PENDING;
                        metadata->flags   |= SOFLAG_DIRTY;
		}
		else {
			rc = ENODEV; // so we dont return DLIST rc
		}

	}
	else {
		rc = EINVAL;
	}

	LOGEXITRC();
	return  rc;
}


int bsd_expand( DISKSEG          *seg,
                storage_object_t *expand_object,
                dlist_t           objects,
                option_array_t   *options )
{
        int rc=ENOSYS;

        LOGENTRY();
        LOGEXITRC();
        return rc;
}


int bsd_shrink( DISKSEG          * seg,
                storage_object_t * shrink_object,
                dlist_t            objects,
                option_array_t   * options )
{
        int rc=ENOSYS;

        LOGENTRY();
        LOGEXITRC();
        return rc;
}


int bsd_add_sectors_to_kill_list( DISKSEG       *seg,
                                  lsn_t          lsn,
                                  sector_count_t count)
{
        int rc = EINVAL;
        LOGICALDISK *ld=NULL;
        struct plugin_functions_s *funcs=NULL;

        LOGENTRY();

        REQUIRE( isa_bsd_segment( seg ) == TRUE );
        REQUIRE( lsn + count > seg->size );

        ld = get_logical_disk( seg );
        if (ld) {

                funcs = (struct plugin_functions_s *)ld->plugin->functions.plugin;

                rc   = funcs->add_sectors_to_kill_list( ld, seg->start+lsn, count);

        }

        LOGEXITRC();
        return rc;
}


/*
 *  Function:  bsd_commit_changes
 *
 *  The engine will call this plug-in API to save BSD metadata
 *  changes to disk.  The two arguments tell us which segment
 *  storage object is marked dirty and which commit phase is
 *  being processed.  We have no work for commit phase Zero. So,
 *  we only call the low level BSD commit routine for subsequent
 *  commit phases.
 */
int bsd_commit_changes( DISKSEG *seg, uint commit_phase )
{        
	int rc;
	LOGICALDISK *ld;
	disk_private_data_t *disk_pdata;


        LOGENTRY();
	
	REQUIRE( isa_bsd_segment(seg) == TRUE );

	ld         = get_logical_disk(seg);	
	REQUIRE( ld != NULL );

	disk_pdata = get_bsd_disk_private_data(ld);
	REQUIRE( disk_pdata != NULL);
	     
	if (commit_phase == MOVE) {
		if ( disk_pdata->flags & DISK_HAS_MOVE_PENDING ) {				
			/*
			rc = bsd_move_segment_commit( disk_pdata->copy_job );
			*/
			rc = ENOSYS;
			if (disk_pdata->copy_job) free(disk_pdata->copy_job );
			disk_pdata->copy_job = NULL;
			disk_pdata->flags   &= ~DISK_HAS_MOVE_PENDING; 
			seg->flags          &= ~SOFLAG_DIRTY;
		}
		else {
			rc = 0;
		}
	}
	else if ( commit_phase==FIRST_METADATA_WRITE || commit_phase==SECOND_METADATA_WRITE ) {
		if ( ( seg->flags & SOFLAG_DIRTY) &&
		     ( disk_pdata->flags & DISK_HAS_CHANGES_PENDING) ) {
			rc = commit_bsd_segments( seg, get_logical_disk(seg), commit_phase );
 		}
      		else {
      			rc = 0;
      		}
	}
      	else {
       		rc=0;
       	}
        

        LOGEXIT();
        return 0;   
}


int bsd_read( DISKSEG           *seg,
              lsn_t              lsn,
              sector_count_t     count,
              void              *buffer )
{
        int rc = EINVAL;
        LOGICALDISK *ld=NULL;

        LOGENTRY();

        REQUIRE( isa_bsd_segment(seg) == TRUE );
        REQUIRE( lsn + count <= seg->size );

        ld = get_logical_disk( seg );
        if (ld) {
                rc = DISK_READ( ld, lsn + seg->start, count, buffer);
        }

        LOGEXITRC();
        return rc;
}

int bsd_write( DISKSEG          *seg,
               lsn_t             lsn,
               sector_count_t    count,
               void             *buffer )
{
        int rc = EINVAL;
        LOGICALDISK *ld=NULL;

        LOGENTRY();

        REQUIRE( isa_bsd_segment(seg) == TRUE );
        REQUIRE( lsn + count <= seg->size );

        ld = get_logical_disk( seg );
        if (ld) {
                rc = DISK_WRITE( ld, lsn + seg->start, count, buffer);
        }

        LOGEXITRC();
        return rc;
}


void bsd_set_volume(storage_object_t * object, BOOLEAN flag)
{
        return;
}


int bsd_get_info( DISKSEG *seg, char * name, extended_info_array_t  ** info_array )
{
        int rc=0;

        LOGENTRY();

        REQUIRE( seg!=NULL );
        REQUIRE( seg->object_type == SEGMENT );
        REQUIRE( info_array != NULL );

        if ( name==NULL ) {
                rc = bsd_get_segment_info(seg, info_array);
        }
        else if ( ( strcmp(name, "Type")==0 ) && (seg->data_type==META_DATA_TYPE) ){
                rc = bsd_get_metadata_info(seg, info_array);
        }

        LOGEXITRC();
        return rc;
}


/*
 *  Function:  delete_all_bsd_segment_private_data
 *
 *  Only called by our plug-in cleanup API.  This routine will
 *  know how to properly free up all bsd segment private data areas.
 */
void delete_all_bsd_segment_private_data(void)
{
        dlist_t list=CreateList();
        int rc;
        storage_object_t *object;

        LOGENTRY();

        if (list != NULL) {

                rc = bsd_eng_funcs->get_object_list( SEGMENT,
                                                     0,
                                                     bsd_plugin,
                                                     NULL,
                                                     0,
                                                     &list );
                if (!rc) {

                        FOR_EACH(object, list){
                                if (object->private_data) free(object->private_data);
                        }

                }

        }

        DestroyList( &list, FALSE );

        LOGEXIT();
}

