/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libdos.so
 *
 *   File: move.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "ptables.h"
#include "move.h"
#include "dm.h"
#include "commit.h"
#include "display.h"
#include "segs.h"
#include "display.h"

// a DOS move context - used by the callback routine
typedef struct dos_move_context_s {
        DISKSEG      *src;
        DISKSEG      *trg;
        progress_t   *progress;
} dos_move_context_t;


/*
 *  Function: safe to move
 *
 *  Called to determine if it would be Ok to move
 *  the sepcified segment object.
 */
static inline BOOLEAN safe_to_move(DISKSEG *seg)
{
	LOGICALDISK        *ld=get_logical_disk(seg);	
	DISK_PRIVATE_DATA  *disk_pdata=NULL;

	if ( ld ) {

		disk_pdata = get_disk_private_data(ld);	
		if (disk_pdata) {
 
                         if ( i_can_modify_seg(seg)  ==  TRUE &&
                              seg->data_type         ==  DATA_TYPE && 
	                      seg_is_volitile(seg)   ==  TRUE  &&
			      !(seg->flags & SOFLAG_DIRTY) &&
	                      (disk_pdata->flags & DISK_HAS_MOVE_PENDING) == 0 ) { 

                                return TRUE;

                        }
                }

        }
        
        return FALSE;
}


/*
 *  Function: free a copy_job_t
 *
 *  Called to free a copy_job_t
 */
static inline void free_copy_job_t(copy_job_t *job)
{
        if (job) free(job);                         
}


/*
 *  Function: alloc copy_job_t
 *
 *  Called to allocate a copy_job_t
 */
static inline copy_job_t * alloc_copy_job_t(void)
{        
        copy_job_t *job = calloc( 1, sizeof(copy_job_t) );  
        return job;
}


/*
 *  Function: write a kill sector
 *
 *  Called to write a single sector of Zeros to the indicated
 *  LSN on the logical disk.
 */
static int write_a_kill_sector( LOGICALDISK *ld, lsn_t lsn )
{
	int rc = EINVAL;	
	char  *buffer=NULL;
	 
	LOGENTRY();

	buffer     = calloc( 1, EVMS_VSECTOR_SIZE );
	if (buffer) {
		rc = ld->plugin->functions.plugin->write(ld, lsn, 1, (void *) buffer);        			
		free(buffer);
	}
	
	LOGEXITRC();
	return rc;
}


/*
 *  Function: write ebr kill sector 
 *
 *  Called to write a kill sector to an EBR location ... but 
 *  only kill the sector if nothing else has been laid down at
 *  that location ... which would have overwritten the EBR!
 */
static int write_ebr_kill_sector( LOGICALDISK *ld, DISKSEG *ebr )
{
	int rc = 0,dlist_rc;
	DISKSEG *seg;
	 
	LOGENTRY();

	DLIST_FOR_EACH(seg,ld->parent_objects,dlist_rc){

		if ( ebr->start >= seg->start &&		     
		     ebr->start <= seg->start+seg->size-1 ) {

			// some other segment sits atop the EBR sector but
			// if it is only a freespace seg then go ahead.
			if (seg->data_type == FREE_SPACE_TYPE) {
				rc = write_a_kill_sector(ld,ebr->start);
			}

			break;
		}
		else if (seg->start > ebr->start) {		
			
			rc = write_a_kill_sector(ld,ebr->start);

			break;
		}

	}
		
	LOGEXITRC();
	return rc;
}
      

/*
 *  Function:  dos copy callback
 *
 *  Called by the engine during the copy commit phase to 
 *  report progress back to us. Every time we get called 
 *  it means that the engine copied a chunk size amount to 
 *  the target device. We are called so that we in turn can 
 *  report progress to the user interface.
 */
static void dos_copy_callback(u_int64_t chunks_copied, void *context)
{
        dos_move_context_t  *dos_context = (dos_move_context_t *)context;		

        ++dos_context->progress->count;

        SegEngFncs->progress(dos_context->progress);        
}


/*
 *  Function: create_copy_job
 *  
 *  Called by dos_move_start() to create an engine copy job that
 *  will be executed during a subsequent commit.
 */
static int create_copy_job( DISKSEG      *seg,       // data segment to be moved
                            DISKSEG      *trg,       // move target segment
                            copy_job_t  **copyjob )  // callers copy_job struct
{
        int rc=ENOMEM;
        DISK_PRIVATE_DATA   *disk_pdata=NULL;
        LOGICALDISK         *ld=NULL;
        copy_job_t          *job=NULL;
        progress_t          *progress=NULL;
        dos_move_context_t  *dos_context=NULL;
        char                *title=NULL;

	
	LOGENTRY();

	ld          = get_logical_disk(seg);
	disk_pdata  = get_disk_private_data(ld);	
        job         = alloc_copy_job_t();
        dos_context = calloc(1, sizeof(dos_move_context_t));
        progress    = calloc(1,sizeof(progress_t));
        title       = malloc(EVMS_NAME_SIZE*2+2);

        if (job && dos_context && progress && title) {
                
                sprintf(title, "Moving segment %s\n", seg->name);
                progress->total_count = seg->size / ld->geometry.sectors_per_track * disk_pdata->vsectors_per_block;
                progress->id = 0;
                progress->title = title;
                progress->description = "";
                progress->type = DISPLAY_PERCENT;
                progress->count = 0;
                progress->remaining_seconds = 0;
                progress->plugin_private_data = NULL;
                progress->ui_private_data = NULL;
 
                dos_context->src      = seg;
                dos_context->trg      = trg;
                dos_context->progress = progress;

                job->src.obj     = ld; 
                job->src.start   = seg->start;
                job->src.len     = seg->size;

                job->trg.obj     = ld; 
                job->trg.start   = trg->start;
                job->trg.len     = seg->size;

                job->chunk_size  = ld->geometry.sectors_per_track * disk_pdata->vsectors_per_block; 
                job->context     = dos_context;

                *copyjob         = job;

                rc               = 0;
        }
        
        if (rc) {
                if (job) free_copy_job_t(job);
                if (dos_context) free(dos_context);
                if (progress) free(progress);
                if (title) free(title);
        }


        LOGEXITRC();
        return rc;
}


/*
 *  Function:  create move target segment
 *
 *  Called by our private function dos_move_segment()
 *  to create a target storage_object_t that will
 *  be used to move the specified disk segment.
 */                
static int create_move_target( DISKSEG  *seg, 	     // source data segment
                               DISKSEG  *freespace,  // target freespace segment
                               DISKSEG **target,     // where to return move target storage object
			       BOOLEAN   testing )   // testing == TRUE if we just want to test
{						     // if a move object could be created and 
	LOGICALDISK       	*ld=NULL;	     // dont actually want the object returned
        DISKSEG                 *trg=NULL;
        DISKSEG                 *ebr=NULL;
	SEG_PRIVATE_DATA  	*src_pdata=NULL;
        SEG_PRIVATE_DATA  	*trg_pdata=NULL;
        SEG_PRIVATE_DATA        *ebr_pdata=NULL;
	DISK_PRIVATE_DATA 	*disk_pdata=NULL;
	lba_t			start=0;
	lba_t                   end=0;
	int 			rc = EINVAL;


        LOGENTRY();

        if (seg) {
                ld          = get_logical_disk(seg);
                disk_pdata  = get_disk_private_data(ld);
                src_pdata   = (SEG_PRIVATE_DATA *) seg->private_data;                		
        }

        if (ld && disk_pdata) {

                rc = ENOMEM;

                trg = allocate_disk_segment( ld );
                if (trg) {

                        sprintf(trg->name, "%s_move_target", seg->name );

                        trg->data_type    = META_DATA_TYPE;
                        trg->flags       |= SOFLAG_DIRTY;
			trg_pdata         = (SEG_PRIVATE_DATA *) trg->private_data;
			trg_pdata->flags |= SEG_IS_MOVE_TARGET;

                        if (src_pdata->boot_ind == ACTIVE_PARTITION) {
                                trg->flags |= SOFLAG_BIOS_READABLE;  
                        }

                        if (src_pdata->flags & SEG_IS_LOGICAL_PARTITION) {

				trg_pdata->flags |= SEG_IS_LOGICAL_PARTITION;
				
                                ebr = allocate_disk_segment( ld );
                                if (ebr) {
                                        sprintf(ebr->name, "%s_move_target", src_pdata->ebr->name);
                                        ebr->data_type          = META_DATA_TYPE;                                       
                                        ebr_pdata 		= (SEG_PRIVATE_DATA *) ebr->private_data;                                         
                                        ebr_pdata->sys_id       = DOS_EXTENDED_PARTITION ;
                                        ebr_pdata->flags        = SEG_IS_EBR;
					trg_pdata->ebr 		= ebr;
					rc = 0;
                                }

                        }
			else {
				trg_pdata->flags |= SEG_IS_PRIMARY_PARTITION;
				trg_pdata->ebr    = src_pdata->ebr;
				rc = 0;
			}
			
                }

        }

        if (!rc) {

		rc = EFBIG;
		
		// calculate the starting lba
		if (src_pdata->flags & SEG_IS_LOGICAL_PARTITION) {

			// a logical partition should always start on a cyl boundary

			if (starts_on_cylinder_boundary(ld,freespace->start)==FALSE) {
				start = roundup_to_cylinder_boundary(ld,freespace->start) + 1;
			}
			else {
				start = freespace->start;
			}

			// adjust for the ebr track
			ebr->start = start;
                        ebr->size  = ld->geometry.sectors_per_track * disk_pdata->vsectors_per_block;    

                        trg->start = ebr->start + ebr->size;
		}
		else {
			// but a primary partition is complicated !!!  Complications involve:
			//
			// (1) source segment could be track alligned NOT cylinder alligned due to following
			//     an MBR track on the disk.
			// (2) a source segment might be track alligned because it has been moved before and was
			//     track alligned because it use to follow the MBR track on the disk.
			// (3) the target freespace may be track alligned because it follows the MBR
			//     track on the disk.		
					
			// CASE 1 and CASE 2 can be solved in this manner				
			if (starts_on_cylinder_boundary(ld,seg->start)==FALSE) {
				sector_count_t sector_offset;
				lba_t          cylinder_start;

				cylinder_start = rounddown_to_cylinder_boundary(ld,seg->start);
				sector_offset  = seg->start - cylinder_start;
		
				if (rounddown_to_cylinder_boundary(ld,freespace->start)==0) {
					start = freespace->start;
				}
				else {		
					start = freespace->start + sector_offset;
				}
			}
			// CASE 3 is recognized here ... no rounding up ... keep track bdy
			else if (rounddown_to_cylinder_boundary(ld,freespace->start)==0) {
				start = freespace->start;
		      	}
			// The rest are normal cases ...
			else if (starts_on_cylinder_boundary(ld,freespace->start)==TRUE) {
 				start = freespace->start;
			}
			else {
				start = roundup_to_cylinder_boundary(ld,freespace->start);
			}

			trg->start = start;
		}

		// now calculate the ending LBA ... always must end on a cyl boundary
		end = trg->start + seg->size - 1;
		if (ends_on_cylinder_boundary(ld,end)==FALSE){
LOG_DEBUG("end lba= %"PRIu64" ... doesnt end on cyl bdy so rounding up\n", end );
			end = roundup_to_cylinder_boundary(ld,end);
LOG_DEBUG("new ... rounded up end lba = %"PRIu64"\n", end );
		}

                if (end <= freespace->start+freespace->size-1 ) {                
			trg->size = end - trg->start + 1;
			*target=trg;
			rc = 0;	 		       
                }
  
	}
 
        if (rc || testing==TRUE) {
                if (ebr) free_disk_segment(ebr);                
                if (trg) free_disk_segment(trg);                                        
        }

        LOGEXITRC();
        return rc;
}


/*  Function:  validate move target 
 *
 *  Called to validate that the specified segment
 *  can be moved to the target freespace segment.
 */
int dos_validate_move_target( DISKSEG *src, DISKSEG *trg )
{
	LOGICALDISK       	*ld=NULL;        
	DISKSEG                 *tseg;
	SEG_PRIVATE_DATA  	*src_pdata=NULL;
        SEG_PRIVATE_DATA  	*trg_pdata=NULL;        
	DISK_PRIVATE_DATA 	*disk_pdata=NULL;
	int rc=EINVAL;

	LOGENTRY();

        if (src && trg) {

		if ( src->data_type == DATA_TYPE &&
		     trg->data_type == FREE_SPACE_TYPE ) {		

			ld          = get_logical_disk(src);
			disk_pdata  = get_disk_private_data(ld);
			src_pdata   = (SEG_PRIVATE_DATA *) src->private_data;
			trg_pdata   = (SEG_PRIVATE_DATA *) trg->private_data;

			if (ld && disk_pdata){			
				rc = 0;
			}
		}

		if (!rc) {		
			if (src_pdata->flags & SEG_IS_LOGICAL_PARTITION) {
				if (seg_is_within_or_adjacant_to_extended_partition(ld,trg)==FALSE) {
					rc = EINVAL;
				}
			}
			else {
				if (seg_is_within_the_extended_partition(ld,trg)==TRUE) {
					rc = EINVAL;
				}
			}
		}

		if (!rc) {
			rc = create_move_target( src, trg, &tseg, TRUE);					
		}

	}

	LOGEXITRC();
	return rc;
}


/*
 *  Function:  can move segment
 *
 *  Check if the specified segment object can be moved
 *  elsewhere on the disk. Do so by ... checking each 
 *  freespace segment we find ... stopping at the 1st
 *  freespace segment that could be used for the move.
 */
int dos_can_move_segment( DISKSEG *seg )
{        
	DISKSEG       *freespace=NULL;	
	LOGICALDISK   *ld=NULL;	
	int rc=EINVAL,dlistrc;
	dlist_t        seglist;

        LOGENTRY();
	
	if ( safe_to_move(seg) == TRUE ) {		

		ld = get_logical_disk(seg);
 		if (ld) {
		
			seglist = CreateList();
			if (seglist) {

				dlistrc = CopyList(seglist,ld->parent_objects,AppendToList);				
				if (!dlistrc) {
				
					DLIST_FOR_EACH(freespace, seglist, dlistrc) {
 			      
						if ( (freespace->data_type == FREE_SPACE_TYPE) &&
					             (freespace->size      >= seg->size) ) {

							rc = dos_validate_move_target(seg,freespace);
							if (rc == 0) {			          					
								break;
							}

						}

					}
 
				}

				DestroyList(&seglist, FALSE);
			}

		}

	} 	
  
        LOGEXITRC();
        return rc;
}


/*
 *  Function:  dos swap segments
 *
 *  Called when finishing a move to update the partition
 *  table metadata info. This means we need to swap out 
 *  the source data segment, replacing it with the target
 *  segment that we just finished copying sectors to.
 */        
static int dos_swap_segments( DISKSEG *src, DISKSEG *trg)
{
        int rc=EINVAL;             
        LOGICALDISK        *ld=NULL;
        DISK_PRIVATE_DATA  *disk_pdata=NULL;
        SEG_PRIVATE_DATA   *src_pdata=NULL;
        SEG_PRIVATE_DATA   *src_ebr_pdata=NULL;
	SEG_PRIVATE_DATA   *trg_pdata=NULL;
        SEG_PRIVATE_DATA   *trg_ebr_pdata=NULL;
	SEG_PRIVATE_DATA   *ebr_pdata=NULL;
        BOOLEAN             logical_partition=FALSE;

	LOGENTRY(); 
						
	if (src && trg){         

     	        ld         = get_logical_disk(src);
	        disk_pdata = get_disk_private_data(ld);	
		src_pdata  = (SEG_PRIVATE_DATA *) src->private_data;		
		trg_pdata  = (SEG_PRIVATE_DATA *) trg->private_data;
 	       
                // check if we moved a logical partition
                if (trg_pdata->flags & SEG_IS_LOGICAL_PARTITION){			
                        logical_partition=TRUE;
                }

                // remove segments from the disks segment list cuz ...
                // doing so will unregister the names and create freespace
                // areas on the disk.
                remove_diskseg_from_list( ld->parent_objects, src );
                remove_diskseg_from_list( ld->parent_objects, trg );
                if (logical_partition == TRUE){			
			remove_diskseg_from_list(ld->parent_objects, src_pdata->ebr );
                        remove_diskseg_from_list(ld->parent_objects, trg_pdata->ebr );
                }
                
                // copy source segment information to target segment
		trg->data_type           = src->data_type;
		trg_pdata->flags        &= ~SEG_IS_MOVE_TARGET;
                trg_pdata->sys_id        = src_pdata->sys_id;
                trg_pdata->boot_ind      = src_pdata->boot_ind;
                trg_pdata->ptable_index  = src_pdata->ptable_index;
                trg_pdata->part_number   = src_pdata->part_number;		
                trg_pdata->ebr_number    = src_pdata->ebr_number; 
                trg_pdata->prev_ebr      = src_pdata->prev_ebr;
                trg_pdata->next_ebr      = src_pdata->next_ebr;
		
/*
LOG_DEBUG("SEG Info ...\n");
LOG_DEBUG("target ...\n");
LOG_DEBUG("         name = %s\n", trg->name);
LOG_DEBUG("        start = %"PRIu64"\n", trg->start);
LOG_DEBUG("     next ebr = %p\n", trg_pdata->next_ebr);
LOG_DEBUG("     prev ebr = %p\n", trg_pdata->prev_ebr);
LOG_DEBUG("          ebr = %p\n", trg_pdata->ebr);
LOG_DEBUG("     part number= %d\n", trg_pdata->part_number);
LOG_DEBUG("source ...\n");
LOG_DEBUG("         name = %s\n", src->name);
LOG_DEBUG("        start = %"PRIu64"\n", src->start);
LOG_DEBUG("     next ebr = %p\n", src_pdata->next_ebr);
LOG_DEBUG("     prev ebr = %p\n", src_pdata->prev_ebr);
LOG_DEBUG("          ebr = %p\n", src_pdata->ebr);
LOG_DEBUG("     part number= %d\n", src_pdata->part_number);
*/

		// fixup ebr chain in the ebr private data areas
		if (logical_partition == TRUE) {		

			src_ebr_pdata = (SEG_PRIVATE_DATA *) src_pdata->ebr->private_data;
			trg_ebr_pdata = (SEG_PRIVATE_DATA *) trg_pdata->ebr->private_data;

			trg_ebr_pdata->ptable_index  = src_ebr_pdata->ptable_index;
			trg_ebr_pdata->part_number   = src_ebr_pdata->part_number;
			trg_ebr_pdata->next_ebr      = src_ebr_pdata->next_ebr;
			trg_ebr_pdata->prev_ebr      = src_ebr_pdata->prev_ebr;
			trg_ebr_pdata->ebr           = src_ebr_pdata->ebr;
			trg_ebr_pdata->ebr_number    = src_ebr_pdata->ebr_number;

			if (trg_ebr_pdata->prev_ebr) {
				ebr_pdata = (SEG_PRIVATE_DATA *) trg_ebr_pdata->prev_ebr->private_data;
				ebr_pdata->next_ebr = trg_pdata->ebr;
			}
			if (trg_ebr_pdata->next_ebr) {
				ebr_pdata = (SEG_PRIVATE_DATA *) trg_ebr_pdata->next_ebr->private_data;
				ebr_pdata->prev_ebr = trg_pdata->ebr; 
				ebr_pdata->ebr      = trg_pdata->ebr; 
			}

/*
LOG_DEBUG("EBR Info ...\n");
LOG_DEBUG("target ...\n");
LOG_DEBUG("         name = %s\n", trg_pdata->ebr->name);
LOG_DEBUG("     next ebr = %p\n", trg_ebr_pdata->next_ebr);
LOG_DEBUG("     prev ebr = %p\n", trg_ebr_pdata->prev_ebr);
LOG_DEBUG("          ebr = %p\n", trg_ebr_pdata->ebr);
LOG_DEBUG("     part number= %d\n", trg_ebr_pdata->part_number);
LOG_DEBUG("      ebr number= %d\n", trg_ebr_pdata->ebr_number);
LOG_DEBUG("source ...\n");
LOG_DEBUG("         name = %s\n", src_pdata->ebr->name);
LOG_DEBUG("     next ebr = %p\n", src_ebr_pdata->next_ebr);
LOG_DEBUG("     prev ebr = %p\n", src_ebr_pdata->prev_ebr);
LOG_DEBUG("          ebr = %p\n", src_ebr_pdata->ebr);
LOG_DEBUG("     part number= %d\n", src_ebr_pdata->part_number);
LOG_DEBUG("      ebr number= %d\n", src_ebr_pdata->ebr_number);
*/

		}

		// take care of OS/2 dlat information
                trg_pdata->dlat          = src_pdata->dlat;
                trg_pdata->dla_entry     = src_pdata->dla_entry;

                if (trg_pdata->dla_entry) {
                        trg_pdata->dla_entry->Partition_Start = trg->start;
                        trg_pdata->dla_entry->Partition_Size  = trg->size;
                }

                // reinsert the target segment cuz ...
                // doing so will now register correct names.                
                if (logical_partition == TRUE){			
                        insert_diskseg_into_list(ld->parent_objects, trg_pdata->ebr );
                }               
                insert_diskseg_into_list( ld->parent_objects, trg );

		// fixup the extended partition info if needed
		if (logical_partition == TRUE){
       			
			// recalc the size of the extended partition
			fixup_disk_extd_partition_dimensions( ld );

			// if logical drives are offset in the extd partition we need to
			// keep an anchor EBR in the 1st sector of the extd partition.
			fixup_disk_extd_partition_anchor(ld);

			// calc sizes of each std ebr partition
			fixup_EBR_Sizes(ld);			
		}

                rc = 0;
        }

        LOGEXITRC();
        return rc;
}


/*  Function:  DOS move segment 
 *
 *  Called as a seg mgr private function to move a data segment
 *  to a freespace area on the same disk.
 */
int dos_move_segment( DISKSEG *src, DISKSEG *freespace )
{
	int rc=EINVAL;
        LOGICALDISK        *ld=NULL;
	DISKSEG            *trg=NULL;
        SEG_PRIVATE_DATA   *trg_pdata=NULL;
	SEG_PRIVATE_DATA   *src_pdata=NULL;
	DISK_PRIVATE_DATA  *disk_pdata=NULL;
        copy_job_t         *job=NULL;

	LOGENTRY();

	if ( safe_to_move(src) == TRUE ) {
 
		ld         = get_logical_disk(src);
		disk_pdata = get_disk_private_data(ld);
		src_pdata  = (SEG_PRIVATE_DATA *) src->private_data;				                
                
                // create a move target segment
                rc = create_move_target(src, freespace, &trg, FALSE);
		if (!rc)  trg_pdata=(SEG_PRIVATE_DATA *)trg->private_data;


      		// create the engine copy job
		if (!rc) {

                        rc = create_copy_job( src, trg, &job );                              		
                        if (rc){                	
				if (trg_pdata->flags & SEG_IS_LOGICAL_PARTITION) {					
					free_disk_segment(trg_pdata->ebr);
				}
        
				free_disk_segment( trg );
                                find_freespace_on_disk(ld);
                        }

		}

                // remove the freespace segment from the disk segment
                // list and insert the new move-target data segment.
                if (!rc) {    						

                        remove_diskseg_from_list( ld->parent_objects, freespace );
                        free_disk_segment( freespace );
                        
                        if (trg_pdata->flags & SEG_IS_LOGICAL_PARTITION){			
                                insert_diskseg_into_ordered_list(ld->parent_objects, trg_pdata->ebr );
                        }               
                        insert_diskseg_into_ordered_list( ld->parent_objects, trg );
		
			disk_pdata->flags     |= DISK_HAS_MOVE_PENDING;		
			disk_pdata->copy_job   = job;
			src_pdata->move_target = trg;

			find_freespace_on_disk(ld);
		}

	}

	LOGEXITRC();
	return rc;
}


/*  Function:  DOS move segment finish
 *
 *  Called as a seg mgr private function to move a data segment
 *  to a freespace on the same disk.
 */
int dos_move_segment_finish( DISKSEG *src, DISKSEG *trg, int copy_rc )
{
	int rc=EINVAL;	
        LOGICALDISK        *ld=NULL;
	SEG_PRIVATE_DATA   *src_pdata=NULL;
	SEG_PRIVATE_DATA   *trg_pdata=NULL;
        DISK_PRIVATE_DATA  *disk_pdata=NULL;

	LOGENTRY(); 
						
	if (src && trg){         

     	        ld         = get_logical_disk(src);
	        disk_pdata = get_disk_private_data(ld);	
		src_pdata  = (SEG_PRIVATE_DATA *)src->private_data;
		trg_pdata  = (SEG_PRIVATE_DATA *)trg->private_data;		

		SEG_deactivate(src);		

                if ( !copy_rc ) {

                        rc = dos_swap_segments(src,trg);                       
                        if (!rc) {
                                 rc = Commit_Disk_Partition_Tables(ld);                                              
			}

                }

		if (!rc) {			
			if (src_pdata->flags & SEG_IS_LOGICAL_PARTITION){							
				write_ebr_kill_sector(ld, src_pdata->ebr);
				free_disk_segment(src_pdata->ebr);
			}		
			free_disk_segment( src );
						
			trg->flags |= SOFLAG_NEEDS_ACTIVATE; 
			SEG_activate( trg ); 
		}
                else {	 
			if (trg_pdata->flags & SEG_IS_LOGICAL_PARTITION) {
				remove_diskseg_from_list( ld->parent_objects, trg_pdata->ebr);                
				write_ebr_kill_sector(ld, trg_pdata->ebr);
				free_disk_segment(trg_pdata->ebr);
			}
			remove_diskseg_from_list( ld->parent_objects, trg);
			free_disk_segment( trg );                        

                        src_pdata->move_target = NULL;
                        src->flags |= SOFLAG_NEEDS_ACTIVATE; 
                        SEG_activate(src);    			
                }

                find_freespace_on_disk(ld); 
        }


	LOGEXITRC();
	return rc;
}
  
/*
 *  Function: dos move commit
 *
 *  Called during MOVE commit phase to commit a segment
 *  move to disk. This is done by calling the engine copy
 *  service and afterwards calling our own move finish
 *  routine that takes care of committing metadata changes
 *  or backing out changes.
 */
int dos_move_segment_commit( copy_job_t *job )
{
	int rc;	
        dos_move_context_t  *dos_context;

	LOGENTRY();

	dos_context = (dos_move_context_t *) job->context;

	// call the engine copy service 
	rc = SegEngFncs->copy( job, dos_copy_callback);
	
	// done ... call finish routine with results
        dos_move_segment_finish( dos_context->src, dos_context->trg, rc);		
	
	// only free context here ... seg mgr frees the copy job
	free(dos_context->progress->title);
	free(dos_context->progress);
	free(dos_context);

       	LOGEXITRC();
	return rc;
}
