/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mdregmgr
 * File: md_main.c
 *
 * Description: Global and core for MD plugins
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include <fcntl.h>
#define MY_PLUGIN my_plugin
#include <sys/ioctl.h>
#include "md.h"


plugin_record_t	    		* linear_plugin = &linear_plugin_record;
plugin_record_t	    		* raid1_plugin = &raid1_plugin_record;
plugin_record_t	    		* raid0_plugin = &raid0_plugin_record;
plugin_record_t                 * raid5_plugin = &raid5_plugin_record;
md_volume_t			* volume_list_head = NULL;	// List of real volume groups, indexed by vg_number.
engine_functions_t		* EngFncs;		// The Engine's internal API set.
plugin_record_t			* my_plugin = NULL;
char message_buffer[1024];
plugin_record_t * evms_plugin_records[] = {&linear_plugin_record,
	                                   &raid1_plugin_record,
	                                   &raid0_plugin_record,
	                                   &raid5_plugin_record,
					   &multipath_plugin_record,
	                                   NULL};


/*
 * Function: calc_log2
 *
 *	Calculate the log-base2 of the argument
 */
long calc_log2( long arg )
{
	long	result = -1;
	long	tmp;

	if ( arg ) {
		tmp = arg;
		result++;
		while ( ! (tmp & 1) ) {
			result++;
			tmp >>= 1;
		}
		if ( tmp != 1 ) {
			// arg isn't a power of 2!
			result = -2;
		}
	}
	RETURN((int)result);
}

static storage_object_t * md_get_region_for_object(storage_object_t *object)
{
	md_volume_t * volume = volume_list_head;
	storage_object_t *region;
	int i;

	LOG_ENTRY;
	
	while (volume) {
		region = volume->region;
		if (region  && (volume->flags & MD_DISCOVERED)) {
			for (i = 0; i< MAX_MD_DEVICES; i++) {
				if (volume->child_object[i] == object) {
					LOG("[%s] belongs to %s\n", object->name, region->name);
					return region;
				}
			}
		}
		volume = volume->next;
	}
	LOG("MD does not own this object [%s]\n", object->name);
	return NULL;
}

/* Function: md_can_replace_child
 *
 *  Can we replace a child?
 */
int md_can_replace_child(storage_object_t *region,
			 storage_object_t *child,
			 storage_object_t *new_child)
{
	u_int32_t child_size, new_child_size;

	LOG_ENTRY;

	/*
	 * We can not replace a child if the region belongs to an EVMS volume.
	 * The reason is that we can not deactivate the region.
	 * The kernel MD refuses to stop the array because Device Mapper 
	 * owns a usage count on the MD region for the EVMS volume device node.
	 */
	if ( md_is_region_active(region)) {
		LOG_DETAILS("Region %s is currently active, checking for EVMS volume...\n", region->name);
		if (region->volume && !(region->volume->flags & VOLFLAG_COMPATIBILITY)) {
			LOG_ERROR("Region %s belongs to an EVMS volume, abort!\n", region->name);
			RETURN(EBUSY);
		} else {
			LOG_DETAILS("Region %s belongs to a compatibility volume\n", region->name);
		}
	}
	

	LOG_DEBUG("region: %s, child:%s, new child:%s\n",
		  region->name, child->name, new_child ? new_child->name : "<unknown>");

	if (region != md_get_region_for_object(child)) {
		LOG_ERROR("[%s] does not belong to %s\n", child->name, region->name);
		RETURN(EINVAL);
	}
	if (new_child) {
		/*
		 * The Engine passed in new child object,
		 * make sure it is a data object and
		 * make sure that the sizes are the same
		 */
		if (new_child->data_type != DATA_TYPE)
			RETURN(EINVAL);

		child_size = MD_NEW_SIZE_SECTORS(child->size);
		new_child_size = MD_NEW_SIZE_SECTORS(new_child->size);
		if (child_size != new_child_size)
			RETURN(EINVAL);
	}

	RETURN(0);
}


/* Function: md_replace_child
 *
 * 
 */
int md_replace_child(storage_object_t *region,
		     storage_object_t *child,
		     storage_object_t *new_child)
{
	int rc=0;
	md_volume_t *volume;
	int i;
	u_int32_t child_size, new_child_size;

	LOG_ENTRY;

	if (region == md_get_region_for_object(child)) {
		
		child_size = MD_NEW_SIZE_SECTORS(child->size);
		new_child_size = MD_NEW_SIZE_SECTORS(new_child->size);

		LOG_DEBUG("region: %s, CHILD: %s, NEW CHILD: %s\n", region->name, child->name, new_child->name);
		LOG_DEBUG(" CHILD object size=%llu, MD size=%d and NEW CHILD: object size=%llu, MD size=%d\n",
			  child->size, child_size, new_child->size, new_child_size);

		if (child_size != new_child_size) {
			LOG_ERROR("Child size=%llu (MD size=%d) is not equal to new child size=%llu (MD size=%d)\n",
				  child->size, child_size, new_child->size, new_child_size);
			RETURN(EINVAL);
		}

		volume = (md_volume_t *)region->private_data;

		for (i = 0; i< MAX_MD_DEVICES; i++) {
			if (volume->child_object[i] == child) {
				
				KILL_SECTORS(child, MD_NEW_SIZE_SECTORS(child->size), MD_RESERVED_SECTORS);
				md_remove_region_from_object(region, child);
				md_append_region_to_object(region, new_child);
				volume->child_object[i] = new_child;
				volume->super_array[i]->this_disk.major = new_child->dev_major;
				volume->super_array[i]->this_disk.minor = new_child->dev_minor;
				volume->super_block->disks[i].major = new_child->dev_major;
				volume->super_block->disks[i].minor = new_child->dev_minor;
				if (md_is_region_active(region)) {
					region->flags |= SOFLAG_NEEDS_DEACTIVATE;
					/* 
					 * Turn off dirty bit for "healthy" MD region, 
					 * so that the kernel MD will not resync.
					 */
					if (!(volume->flags & (MD_CORRUPT | MD_DEGRADED)))
						volume->super_block->state |= (1 << MD_SB_CLEAN);
				}
				region->flags |= SOFLAG_DIRTY;
				/*
				 * The Engines calls us twice to replace a child.
				 * We don't want to activate the MD region the first time, when
				 * the new child is the temp "replaced" object.
				 */
				if (new_child->dev_major == 0)
					volume->flags |= MD_DELAY_ACTIVATE;
				else {
					volume->flags &= ~MD_DELAY_ACTIVATE;
					region->flags |= SOFLAG_NEEDS_ACTIVATE;
				}
				break;
			}
		}
	} else {
		LOG_ERROR("%s is not an MD region\n", region->name);
		rc = EINVAL;
	}

	RETURN(rc);
}

int md_ioctl_start_array(storage_object_t *region, int dev_major, int dev_minor)
{
	int rc = 0, fd, ioctl_cmd;
	ulong arg = 0;
	LOG_ENTRY;

	fd = EngFncs->open_object(region, O_RDWR);
	if (fd <= 0) {
		LOG_ERROR("Unable to open md object %s to send ioctl\n", region->name);
		rc = -fd;
	} else {
		ioctl_cmd = START_ARRAY;
		arg = MKDEV(dev_major, dev_minor);
		rc = EngFncs->ioctl_object(region, fd, ioctl_cmd, arg);
		if (rc) {
			LOG_ERROR("Error starting MD array %s (major=%d, minor=%d), rc=%d\n", region->name, dev_major, dev_minor, rc);
		}
		EngFncs->close_object(region, fd);
	}

	RETURN(rc);

}

int md_ioctl_get_array_info(storage_object_t *region, mdu_array_info_t *md_info)
{
	int rc = 0, fd, ioctl_cmd;
	ulong arg = 0;
	LOG_ENTRY;

	fd = EngFncs->open_object(region, O_RDWR);
	if (fd <= 0) {
		LOG_ERROR("Unable to open md object %s to send ioctl\n", region->name);
		rc = -fd;
	} else {
		ioctl_cmd = GET_ARRAY_INFO;
		arg = (ulong)md_info;
		rc = EngFncs->ioctl_object(region, fd, ioctl_cmd, arg);
		if (rc) {
			LOG_ERROR("Error getting info from MD array %s, rc=%d\n", region->name, rc);
		}
		EngFncs->close_object(region, fd);
	}

	RETURN(rc);

}

#define BLKGETSIZE64 _IOR(0x12,114,sizeof(u_int64_t)) 
int md_ioctl_get_blk_size(storage_object_t *region, u_int64_t *size)
{
	int rc = 0, fd, ioctl_cmd;
	ulong arg = 0;
	LOG_ENTRY;

	fd = EngFncs->open_object(region, O_RDWR);
	if (fd <= 0) {
		LOG_ERROR("Unable to open md object %s to send ioctl\n", region->name);
		rc = -fd;
	} else {
		ioctl_cmd = BLKGETSIZE64;
		arg = (ulong)size;
		rc = EngFncs->ioctl_object(region, fd, ioctl_cmd, arg);
		if (rc) {
			LOG_ERROR("Error getting size of md region %s, rc=%d size=%"PRIu64".\n", region->name, rc, *size);
		}
		EngFncs->close_object(region, fd);
	}

	RETURN(rc);

}

int md_ioctl_hot_add_spare(storage_object_t *region, int dev_major, int dev_minor)
{
	int rc = 0, fd, ioctl_cmd;
	ulong arg = 0;
	LOG_ENTRY;

	fd = EngFncs->open_object(region, O_RDWR);
	if (fd <= 0) {
		LOG_ERROR("Unable to open md object %s to send ioctl\n", region->name);
		rc = -fd;
	} else {
		ioctl_cmd = HOT_ADD_DISK;
		arg = MKDEV(dev_major, dev_minor);
		rc = EngFncs->ioctl_object(region, fd, ioctl_cmd, arg);
		if (rc) {
			LOG_ERROR("Unable to hot add the spare (major=%d, minor=%d) to [%s]\n",
					  dev_major, dev_minor, region->name);
		}
		EngFncs->close_object(region, fd);
	}

	RETURN(rc);
}


/*
 * Function: md_find_first_active_disk
 *
 * Return the first active disk
 */
u_int32_t md_find_first_active_disk(md_volume_t *volume)
{
	int i;
	u_int32_t state;

	LOG_ENTRY;

	for (i=0; i< MAX_MD_DEVICES; i++) {
		if (!volume->child_object[i])
			continue;
		state = volume->super_block->disks[i].state;
		if (
			(state == (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC)) ||
			(state == (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_PENDING_ACTIVE )) ||
			(state == (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_PENDING_ACTIVE | 1<<MD_DISK_NEW)) ||
			(state == (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_NEW)) )
			break;
	}
	if (i == MAX_MD_DEVICES)
		LOG("could not find any active disk in region %s\n", volume->name);
	RETURN(i);
}
