/*
 *
 *   Copyright (c) International Business Machines  Corp., 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libs390.so
 *
 *   File: move.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "390segmgr.h"
#include "helpers.h"
#include "move.h"
#include "dm.h"
#include "commit.h"


// move context - used by the callback routine
typedef struct move_context_s {
	DISKSEG      *src;
	DISKSEG      *trg;
	progress_t   *progress;
} move_context_t;


/*
 *  Function: safe to move
 *
 *  Called to determine if it would be Ok to move
 *  the sepcified segment object.
 */
static inline BOOLEAN safe_to_move(DISKSEG *seg)
{
	LOGICALDISK        *ld=get_logical_disk(seg);   
	DISK_PRIVATE_DATA  *disk_pdata=NULL;

	if ( ld ) {

		disk_pdata = get_s390_disk_private_data(ld);    
		if (disk_pdata) {

			if (  i_can_modify(seg)  ==  TRUE &&
			      seg->data_type     ==  DATA_TYPE &&
			      (disk_pdata->flags & DISK_HAS_MOVE_PENDING) == 0 ) {

				return TRUE;

			}
		}

	}

	return FALSE;
}


/*
 *  Function: free a copy_job_t
 *
 *  Called to free a copy_job_t
 */
static inline void free_copy_job_t(copy_job_t *job)
{
	if (job) free(job);
}


/*
 *  Function: alloc copy_job_t
 *
 *  Called to allocate a copy_job_t
 */
static inline copy_job_t * alloc_copy_job_t(void)
{        
	copy_job_t *job = calloc( 1, sizeof(copy_job_t) );  
	return job;
}


/*
 *  Function: write kill sectors 
 *
 *  Called to write a Zero filled sectors to the indicated
 *  LSN on the logical disk.
 */
static int write_kill_sectors( LOGICALDISK *ld, lsn_t lsn, sector_count_t count )
{
	int rc = EINVAL;        
	char  *buffer=NULL;

	LOGENTRY();

	buffer     = calloc( count, EVMS_VSECTOR_SIZE );
	if (buffer) {
		rc = ld->plugin->functions.plugin->write(ld, lsn, count, (void *) buffer);                              
		free(buffer);
	}

	LOGEXITRC();
	return rc;
}



/*
 *  Function:  copy callback
 *
 *  Called by the engine during the copy commit phase to 
 *  report progress back to us. Every time we get called 
 *  it means that the engine copied a chunk size amount to 
 *  the target device. We are called so that we in turn can 
 *  report progress to the user interface.
 */
static void copy_callback(u_int64_t chunks_copied, void *context)
{
	move_context_t  *copy_context = (move_context_t *)context;              

	++copy_context->progress->count;

	S390EngFuncs->progress(copy_context->progress);        
}


/*
 *  Function: create_copy_job
 *  
 *  Called by s390_move_start() to create an engine copy job that
 *  will be executed during a subsequent commit.
 */
static int create_copy_job( DISKSEG      *seg,	     // data segment to be moved
			    DISKSEG      *trg,	     // move target segment
			    copy_job_t  **copyjob )  // callers copy_job struct
{
	int rc=ENOMEM;
	DISK_PRIVATE_DATA   *disk_pdata=NULL;
	LOGICALDISK         *ld=NULL;
	copy_job_t          *job=NULL;
	progress_t          *progress=NULL;
	move_context_t      *copy_context=NULL;
	char                *title=NULL;


	LOGENTRY();

	ld           = get_logical_disk(seg);
	disk_pdata   = get_s390_disk_private_data(ld);  
	job          = alloc_copy_job_t();
	copy_context = calloc(1, sizeof(move_context_t));
	progress     = calloc(1,sizeof(progress_t));
	title        = malloc(EVMS_NAME_SIZE*2+2);

	if (job && copy_context && progress && title) {

		sprintf(title, "Moving segment %s\n", seg->name);
		progress->total_count = seg->size / (ld->geometry.sectors_per_track * disk_pdata->vsectors_per_block);
		progress->id = 0;
		progress->title = title;
		progress->description = "";
		progress->type = DISPLAY_PERCENT;
		progress->count = 0;
		progress->remaining_seconds = 0;
		progress->plugin_private_data = NULL;
		progress->ui_private_data = NULL;

		copy_context->src      = seg;
		copy_context->trg      = trg;
		copy_context->progress = progress;

		job->src.obj     = ld; 
		job->src.start   = seg->start;
		job->src.len     = seg->size;

		job->trg.obj     = ld; 
		job->trg.start   = trg->start;
		job->trg.len     = seg->size;

		job->chunk_size  = ld->geometry.sectors_per_track * disk_pdata->vsectors_per_block; 
		job->context     = copy_context;

		*copyjob         = job;

		rc               = 0;
	}

	if (rc) {
		if (job) free_copy_job_t(job);
		if (copy_context) free(copy_context);
		if (progress) free(progress);
		if (title) free(title);
	}


	LOGEXITRC();
	return rc;
}


/*
 *  Function:  create move target segment
 *
 *  Called by our private function s390_move_segment()
 *  to create a target storage_object_t that will
 *  be used to move the specified disk segment.
 */                
static int create_move_target( DISKSEG  *seg,	     // source data segment
			       DISKSEG  *freespace,  // target freespace segment
			       DISKSEG **target,     // where to return move target storage object
			       BOOLEAN   testing )   // testing == TRUE if we just want to test
{						     // if a move object could be created and 
	LOGICALDISK             *ld=NULL;	     // dont actually want the object returned
	DISKSEG                 *trg=NULL;
	SEG_PRIVATE_DATA        *src_pdata=NULL;
	SEG_PRIVATE_DATA        *trg_pdata=NULL;
	DISK_PRIVATE_DATA       *disk_pdata=NULL;
	lba_t                   start=0;
	lba_t                   end=0;
	int                     rc = EINVAL;


	LOGENTRY();

	if (seg) {
		ld          = get_logical_disk(seg);
		disk_pdata  = get_s390_disk_private_data(ld);
		src_pdata   = (SEG_PRIVATE_DATA *) seg->private_data;                           
	}

	if (ld && disk_pdata) {

		rc = ENOMEM;

		trg = allocate_s390_disk_segment( ld );
		if (trg) {

			sprintf(trg->name, "%s_move_target", seg->name );

			trg->data_type = DATA_TYPE;
			trg->flags    |= SOFLAG_DIRTY;
			trg_pdata      = (SEG_PRIVATE_DATA *) trg->private_data;

			rc = 0;
		}

	}

	if (!rc) {

		rc = EFBIG;

		// calculate the starting lba     							
		if (starts_on_cylinder_boundary(ld,seg->start)==TRUE) {
			if (starts_on_cylinder_boundary(ld,freespace->start)==TRUE) {
				start = freespace->start;
			}
			else {
				start = roundup_to_cylinder_boundary(ld,freespace->start);
			}
		}
		else {
			sector_count_t sector_offset;
			lba_t          cylinder_start;

			if (starts_on_cylinder_boundary(ld,freespace->start)==TRUE) {
				cylinder_start = rounddown_to_cylinder_boundary(ld,seg->start);
				sector_offset  = seg->start - cylinder_start;
				start = freespace->start + sector_offset;
			}
			else {
				start = freespace->start;
			}                               
		}

		trg->start = start;

		// now calculate the ending LBA ... always must end on a cyl boundary
		end = trg->start + seg->size - 1;
		if (ends_on_cylinder_boundary(ld,end)==FALSE) {
			end = roundup_to_cylinder_boundary(ld,end);
		}

		if (end <= freespace->start+freespace->size-1 ) {
			trg->size = end - trg->start + 1;
			*target=trg;
			rc = 0;                        
		}

	}

	if (rc || testing==TRUE) {
		if (trg) free_s390_disk_segment(trg);
	}

	LOGEXITRC();
	return rc;
}


/*  Function:  validate move target 
 *
 *  Called to validate that the specified segment
 *  can be moved to the target freespace segment.
 */
int s390_validate_move_target( DISKSEG *src, DISKSEG *trg )
{
	LOGICALDISK             *ld=NULL;        
	DISKSEG                 *tseg;
	SEG_PRIVATE_DATA        *src_pdata=NULL;
	SEG_PRIVATE_DATA        *trg_pdata=NULL;        
	DISK_PRIVATE_DATA       *disk_pdata=NULL;
	int rc=EINVAL;

	LOGENTRY();

	if (src && trg) {

		if ( src->data_type == DATA_TYPE &&
		     trg->data_type == FREE_SPACE_TYPE ) {

			ld          = get_logical_disk(src);
			disk_pdata  = get_s390_disk_private_data(ld);
			src_pdata   = (SEG_PRIVATE_DATA *) src->private_data;
			trg_pdata   = (SEG_PRIVATE_DATA *) trg->private_data;

			if (ld && disk_pdata) {
				rc = 0;
			}
		}

		if (!rc) {
			rc = create_move_target( src, trg, &tseg, TRUE);                                        
		}

	}

	LOGEXITRC();
	return rc;
}


/*
 *  Function:  can move segment
 *
 *  Check if the specified segment object can be moved
 *  elsewhere on the disk. Do so by ... checking each 
 *  freespace segment we find ... stopping at the 1st
 *  freespace segment that could be used for the move.
 */
int s390_can_move_segment( DISKSEG *seg )
{        
	DISKSEG       *freespace=NULL;  
	LOGICALDISK   *ld=NULL; 
	int rc=EINVAL,dlistrc;
	dlist_t        seglist;

	LOGENTRY();

	if ( safe_to_move(seg) == TRUE ) {

		ld = get_logical_disk(seg);
		if (ld) {

			seglist = CreateList();
			if (seglist) {

				dlistrc = CopyList(seglist,ld->parent_objects,AppendToList);                            
				if (!dlistrc) {

					DLIST_FOR_EACH(freespace, seglist, dlistrc) {

						if ( (freespace->data_type == FREE_SPACE_TYPE) &&
						     (freespace->size      >= seg->size) ) {

							rc = s390_validate_move_target(seg,freespace);
							if (rc == 0) {
								break;
							}

						}

					}

				}

				DestroyList(&seglist, FALSE);
			}

		}

	}

	LOGEXITRC();
	return rc;
}


/*
 *  Function:  swap segments
 *
 *  Called when finishing a move to update the partition
 *  table metadata info. This means we need to swap out 
 *  the source data segment, replacing it with the target
 *  segment that we just finished copying sectors to.
 */        
static int swap_segments( DISKSEG *src, DISKSEG *trg)
{
	int rc=EINVAL;             
	LOGICALDISK        *ld=NULL;
	DISK_PRIVATE_DATA  *disk_pdata=NULL;

	LOGENTRY(); 

	if (src && trg) {

		ld         = get_logical_disk(src);
		disk_pdata = get_s390_disk_private_data(ld);    

		// remove segments from the disks segment list cuz ...
		// doing so will unregister the names and create freespace
		// areas on the disk.
		remove_s390_segment_from_list( ld->parent_objects, src );
		remove_s390_segment_from_list( ld->parent_objects, trg );

		// copy source segment information to target segment
		memcpy( trg->private_data, src->private_data, sizeof(SEG_PRIVATE_DATA));

		// copy name
		strcpy( trg->name, src->name);

		// reinsert the target segment 
		insert_s390_segment_into_list( ld->parent_objects, trg );

		rc = 0;
	}

	LOGEXITRC();
	return rc;
}


/*  Function:  390 move segment 
 *
 *  Called as a seg mgr private function to move a data segment
 *  to a freespace area on the same disk.
 */
int s390_move_segment( DISKSEG *src, DISKSEG *freespace )
{
	int rc=EINVAL;
	LOGICALDISK        *ld=NULL;
	DISKSEG            *trg=NULL;
	SEG_PRIVATE_DATA   *trg_pdata=NULL;
	SEG_PRIVATE_DATA   *src_pdata=NULL;
	DISK_PRIVATE_DATA  *disk_pdata=NULL;
	copy_job_t         *job=NULL;

	LOGENTRY();

	if ( safe_to_move(src) == TRUE ) {

		ld         = get_logical_disk(src);
		disk_pdata = get_s390_disk_private_data(ld);
		src_pdata  = (SEG_PRIVATE_DATA *) src->private_data;                                            

		// create a move target segment
		rc = create_move_target(src, freespace, &trg, FALSE);
		if (!rc)  trg_pdata=(SEG_PRIVATE_DATA *)trg->private_data;


		// create the engine copy job
		if (!rc) {

			rc = create_copy_job( src, trg, &job );                                         
			if (rc) {
				free_s390_disk_segment( trg );
				find_freespace_on_s390_disk(ld);
			}

		}

		// remove the freespace segment from the disk segment
		// list and insert the new move-target data segment.
		if (!rc) {

			remove_s390_segment_from_list( ld->parent_objects, freespace );
			free_s390_disk_segment( freespace );

			insert_s390_segment_into_ordered_list( ld->parent_objects, trg );

			disk_pdata->flags     |= DISK_HAS_MOVE_PENDING;         
			disk_pdata->copy_job   = job;
			src_pdata->move_target = trg;

			find_freespace_on_s390_disk(ld);
		}

	}

	LOGEXITRC();
	return rc;
}


/*  Function:  390 move segment finish
 *
 *  Called as a seg mgr private function to move a data segment
 *  to a freespace on the same disk.
 */
int s390_move_segment_finish( DISKSEG *src, DISKSEG *trg, int copy_rc )
{
	int rc=EINVAL, rc2;
	LOGICALDISK        *ld=NULL;
	DISKSEG            *seg=NULL;
	SEG_PRIVATE_DATA   *src_pdata=NULL;
	SEG_PRIVATE_DATA   *trg_pdata=NULL;
	SEG_PRIVATE_DATA   *pdata=NULL;
	DISK_PRIVATE_DATA  *disk_pdata=NULL;

	LOGENTRY(); 

	if (src && trg) {

		ld         = get_logical_disk(src);
		disk_pdata = get_s390_disk_private_data(ld);    
		src_pdata  = (SEG_PRIVATE_DATA *)src->private_data;
		trg_pdata  = (SEG_PRIVATE_DATA *)trg->private_data;             

		s390_deactivate(src);           

		if ( !copy_rc ) {
			rc = swap_segments(src,trg);                       
			if (!rc) {
				rc = commit_390_partition_table(ld,FIRST_METADATA_WRITE);
			}
		}

		if (!rc) {
			write_kill_sectors( ld,
					    src->start,
					    ld->geometry.sectors_per_track*disk_pdata->vsectors_per_block);
			free_s390_disk_segment( src );
			fixup_390_partition_names( ld );    // new
			//s390_activate( trg );
			DLIST_FOR_EACH(seg, ld->parent_objects, rc2){
				if (seg->data_type == DATA_TYPE) {						
					pdata = (SEG_PRIVATE_DATA *)seg->private_data;
					if (pdata) {
						if (pdata->cflags & SEG_NEEDS_DM_RENAME) {
							s390_activate(seg);
						}
					}
				}
			}
			trg->flags |= SOFLAG_NEEDS_ACTIVATE; 
			s390_activate(trg);
			rc = 0;

  		}
		else {
			remove_s390_segment_from_list( ld->parent_objects, trg);
			remove_s390_segment_from_list( ld->parent_objects, src);
			insert_s390_segment_into_ordered_list( ld->parent_objects, trg );
			src_pdata->move_target = NULL;
			free_s390_disk_segment( trg );
			src->flags |= SOFLAG_NEEDS_ACTIVATE; 
			s390_activate(src);                     
		}

		find_freespace_on_s390_disk(ld); 
	}


	LOGEXITRC();
	return rc;
}

/*
 *  Function: 390 move commit
 *
 *  Called during MOVE commit phase to commit a segment
 *  move to disk. This is done by calling the engine copy
 *  service and afterwards calling our own move finish
 *  routine that takes care of committing metadata changes
 *  or backing out changes.
 */
int s390_move_segment_commit( copy_job_t *job )
{
	int rc; 
	move_context_t  *copy_context;

	LOGENTRY();

	copy_context = (move_context_t *) job->context;

	// call the engine copy service 
	rc = S390EngFuncs->copy( job, copy_callback);

	// done ... call finish routine with results
	s390_move_segment_finish( copy_context->src, copy_context->trg, rc);            

	// only free context here ... seg mgr frees the copy job
	free(copy_context->progress->title);
	free(copy_context->progress);
	free(copy_context);

	LOGEXITRC();
	return rc;
}
