/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: fs_xfs.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <plugin.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "fsimxfs.h"

plugin_record_t * my_plugin_record = &xfs_plugin_record;
engine_functions_t * EngFncs = NULL;
char xfsutils_version[10];
u_int32_t xfsutils_support;

/*-------------------------------------------------------------------------------------+
+                                                                                      +
+                            Start Of EVMS Plugin Functions                            +
+                        (exported to engine via function table)                       +
+                                                                                      +
+-------------------------------------------------------------------------------------*/


static int fs_setup( engine_functions_t *engine_function_table)
{
	int rc = 0;
	EngFncs = engine_function_table;

	LOG_ENTRY();

	xfsutils_support = 0;

	/*
	 * xfsutils must be at version 2.0.0 or greater to work properly with this FSIM.
	 * Also, initialize xfsutils_version and xfsutils_support.
	 */
	rc = fsim_test_version();

	if ( rc ) {
		LOG_WARNING( "xfsutils must be version 2.0.0 or later to function properly with this FSIM.\n" );
		LOG_WARNING( "xfsutils must also reside in the search path specified by PATH.\n" );
		LOG_WARNING( "Please get/install the current version of xfsutils from http://oss.sgi.com/projects/xfs\n" );
		rc = ENOSYS;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Free all of the private data item we have left on volumes.
 */
static void fs_cleanup()
{
	int rc = 0;
	dlist_t global_volumes;
	logical_volume_t * volume;
	LOG_ENTRY();

	rc = EngFncs->get_volume_list(my_plugin_record, NULL, 0, &global_volumes);
	if (!rc) {
		while (ExtractObject(global_volumes, VOLUME_TAG, NULL, (void**)&volume)==0) {
			if (volume->private_data) {
				EngFncs->engine_free(volume->private_data);
			}
		}
	}
	LOG_EXIT_VOID();
}

/*
 * Get the size limits for this volume.
 */
static int fs_get_fs_limits( logical_volume_t * volume,
			     sector_count_t   * min_fs_size,
			     sector_count_t   * max_fs_size,
			     sector_count_t   * max_object_size)
{
	int rc = EINVAL;
	xfs_volume_t * xfs_vol = (xfs_volume_t *) volume->private_data;

	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}

	if (xfs_vol->sb) {
		/* get and validate current XFS superblock */
		rc = fsim_get_xfs_superblock( volume, xfs_vol->sb );

		if ( !rc ) {
			//rc = fsim_get_volume_limits( xfs_vol->sb, min_size, max_volume_size, max_object_size);

			*max_fs_size = (u_int64_t)1 << 63;
			*max_object_size = (u_int64_t)1 << 63;
			*min_fs_size     = xfs_vol->sb->sb_dblocks * (xfs_vol->sb->sb_blocksize/EVMS_VSECTOR_SIZE);
			LOG_EXTRA("volume:%s, min:%"PRIu64", max:%"PRIu64"\n",volume->dev_node, *min_fs_size, *max_fs_size);
			LOG_EXTRA("fssize:%"PRIu64", vol_size:%"PRIu64"\n",volume->fs_size,volume->vol_size );

			if (*min_fs_size > volume->vol_size) {
				LOG_ERROR("XFS FSIM returned min size > volume size, setting min size to volume size\n");
				*min_fs_size = volume->vol_size;
			}
		}
	} else {
		/* XFS external log */
		if (xfs_vol->fs_vol) {	  // if log associated with a fs
		        *min_fs_size = *max_fs_size = ((xfs_volume_t *)(xfs_vol->fs_vol->private_data))->sb->sb_logblocks *
			     	   (((xfs_volume_t *)(xfs_vol->fs_vol->private_data))->sb->sb_blocksize /
			     	    EVMS_VSECTOR_SIZE);
			*max_object_size = (u_int64_t)1 << 63;
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Does this FSIM manage the file system on this volume?
 * Return 0 for "yes", else a reason code.
 */
static int fs_probe(logical_volume_t * volume)
{
	int  rc = 0;
	xfs_volume_t * xfs_vol;
	dlist_t XFS_vols = NULL;
	xfs_volume_t * tmp_vol;
	logical_volume_t * tmp;
	int tag;

	LOG_ENTRY();

	//BUGBUG make sure not memory leak here!!!
	volume->private_data = NULL;

	/* allocate space for copy of superblock in private data */
	xfs_vol = EngFncs->engine_alloc(sizeof(xfs_volume_t));
	if ( !xfs_vol ) {
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	xfs_vol->sb = EngFncs->engine_alloc( SIZE_OF_SUPER );
	if ( !xfs_vol->sb ) {
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	/* get and validate XFS superblock */
	rc = fsim_get_xfs_superblock( volume, xfs_vol->sb );

	if ( !rc ) {
		/* store copy of valid XFS superblock in private data */
		volume->private_data = (void *)xfs_vol;
	} else {
		/* didn't find valid XFS superblock,  */
		/* look for XFS log superblock        */
		EngFncs->engine_free(xfs_vol->sb);
		xfs_vol->sb = NULL;
		xfs_vol->log_sb = EngFncs->engine_alloc(sizeof (xlog_rec_header_t));
		if ( !xfs_vol->log_sb ) {
			LOG_EXIT_INT(ENOMEM);
			return ENOMEM;
		}
		rc = fsim_get_log_superblock( volume, (xlog_rec_header_t *)xfs_vol->log_sb );
		if ( !rc ) {
			/* store copy of valid XFS log superblock in private data */
			volume->private_data = (void *)xfs_vol;
		} else {
			EngFncs->engine_free(xfs_vol->log_sb);
			EngFncs->engine_free(xfs_vol);
		}
	}

	if (!rc) {
		/* get list of XFS claimed volumes */
		rc = EngFncs->get_volume_list(my_plugin_record, NULL, 0, &XFS_vols);
		GoToStartOfList(XFS_vols);

		/*
		 * Look through all of the volumes to one that is
		 * an external log for this XFS volume or the FS vol fof this external log.  We do this by
		 * checking for a match of the actual uuid of the external log, and the
		 * uuid of the external log that is stored in the XFS volume's superblock.
		 */
		while (BlindExtractObject(XFS_vols, (TAG *) &tag, NULL, (void**)&tmp)==0) {
			tmp_vol = (xfs_volume_t *)tmp->private_data;
			/* just found a new XFS vol, look for pre-discovered log volumes that match */
			if (xfs_vol->sb && tmp_vol->log_sb) {
				/* compare the log uuid stored in the XFS superblock with the       */
				/* actual uuid of the external log volume to see if we have a match */
				if (!uuid_compare(xfs_vol->sb->sb_uuid, tmp_vol->log_sb->h_fs_uuid)) {
					/* If we have a match, add the XFS vol/external log vol pair to the list */
					xfs_vol->log_vol = tmp;
					tmp_vol->fs_vol = volume;
					// since we discovered the external log before the fs volume,
					// the size information for the log volume is incorrect
					// due the the log size only being stored if the FS supoerblock
					// so go update the size infor for the external log volume.
					fs_get_fs_limits(tmp, &tmp->min_fs_size,
							 &tmp->max_fs_size,
							 &tmp->max_vol_size);
				}
			} else if (xfs_vol->log_sb && tmp_vol->sb) { /* log volume, look for fs volume */
				if (!uuid_compare(xfs_vol->log_sb->h_fs_uuid, tmp_vol->sb->sb_uuid)) {
					xfs_vol->fs_vol = tmp;
					tmp_vol->log_vol = volume;
				}
			}
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Can mkfs this volume?
 */
static int fs_can_mkfs(logical_volume_t * volume)
{
	int  rc=0;

	LOG_ENTRY();

	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't format. */
		rc = EBUSY;
	} else if ((volume->vol_size * EVMS_VSECTOR_SIZE) < MINXFS) {
		/* volume size must be >= MINXFS */
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Can unmkfs this volume?
 */
static int fs_can_unmkfs(logical_volume_t * volume)
{
	int  rc=0;
	xfs_volume_t * xfs_vol = (xfs_volume_t *) volume->private_data;

	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}

	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't unmkfs. */
		rc = EBUSY;
	} else if (xfs_vol->log_sb && xfs_vol->fs_vol) {
		/* If XFS external log volume,and has an associated file system
		can't unmkfs */
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Can fsck this volume?
 */
static int fs_can_fsck(logical_volume_t * volume)
{
	int  rc=0;
	xfs_volume_t * xfs_vol = (xfs_volume_t *) volume->private_data;

	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}

	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't unmkfs. */
		rc = EBUSY;
	} else if (xfs_vol->log_sb) {
		/* If XFS external log volume, can't fsck */
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Get the current size of this volume
 */
static int fs_get_fs_size( logical_volume_t * volume,
			   sector_count_t   * size    )
{
	int  rc = EINVAL;
	xfs_volume_t * xfs_vol = (xfs_volume_t *) volume->private_data;

	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}

	if (xfs_vol->sb) {
		/* get and validate current XFS superblock */
		rc = fsim_get_xfs_superblock(volume, xfs_vol->sb);

		if (!rc) {
			*size = (xfs_vol->sb->sb_dblocks)
				* (xfs_vol->sb->sb_blocksize/EVMS_VSECTOR_SIZE);
		}
	} else {
		if (xfs_vol->fs_vol) {	  // if log associated with a fs
			*size = ((xfs_volume_t *)(xfs_vol->fs_vol->private_data))->sb->sb_logblocks *
			     	   (((xfs_volume_t *)(xfs_vol->fs_vol->private_data))->sb->sb_blocksize /
			     	    EVMS_VSECTOR_SIZE);
		} else {
			rc = EINVAL;
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}



/*
 * Expand the volume to new_size.  If the volume is not expanded exactly to
 * new_size, set new_size to the new_size of the volume.
 */
static int fs_expand( logical_volume_t * volume,
		      sector_count_t   * new_size )
{
	int     rc = FSIM_ERROR;
	char   *argv[4];
	pid_t   pidf;
	int     status, bytes_read;
	int     fds2[2];
	char    *buffer = NULL;

	LOG_ENTRY();
	/* open pipe, alloc buffer for collecting fsck.xfs output */
	if ( (rc = pipe(fds2)) ) {
		return rc;
	}
	if (!(buffer = EngFncs->engine_alloc(MAX_USER_MESSAGE_LEN))) {
		close(fds2[0]);
		close(fds2[1]);
		return ENOMEM;
	}

	argv[0] = "xfs_growfs";
	SET_STRING(argv[1],volume->mount_point);
	argv[2] = NULL;

	pidf = EngFncs->fork_and_execvp(volume, argv, NULL, fds2, fds2);
	if (pidf != -1) {
		fcntl(fds2[0], F_SETFL, fcntl(fds2[0], F_GETFL,0) | O_NONBLOCK);
		while (!(waitpid( pidf, &status, WNOHANG ))) {
			bytes_read = read(fds2[0],buffer,MAX_USER_MESSAGE_LEN);
			if (bytes_read > 0) {
				MESSAGE("XFS growfs output: \n%s",buffer); // on prompt if there is one
				memset(buffer,0,bytes_read); //clear out message
			}
			usleep(10000); // don't hog all the cpu
		}
		if ( WIFEXITED(status) && (WEXITSTATUS(status) != ENOENT)) {
			do {
				bytes_read = read(fds2[0],buffer,MAX_USER_MESSAGE_LEN);
				if (bytes_read > 0) {
					MESSAGE("XFS growfs output: \n%s",buffer);
				}
			} while (bytes_read == MAX_USER_MESSAGE_LEN);
			rc = WEXITSTATUS(status);
			if (rc == 0) {
				LOG_DETAILS("XFS growfs completed with rc = %d \n",status);
			} else {
				LOG_ERROR("XFS growfs completed with rc = %d \n",status);
			}
		} else {
			rc = EINTR;
		}

	} else {
		rc = EIO;
	}

	EngFncs->engine_free(buffer);
	close(fds2[0]);
	close(fds2[1]);

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * unmkfs has been scheduled.  Do any setup work such as
 * releasing another volume that was used for an external log.
 */
static int fs_unmkfs_setup(logical_volume_t * volume)
{
	int rc = 0;
	xfs_volume_t  *xfs_vol = (xfs_volume_t *)volume->private_data;

	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}

	/* see if the volume to be unmkfs'ed has an external log */
	/* if so, free the private data on the external log and call */
	/* call unassign to dissacociate XFS from the volume */
	/* private data on real fs vol still has reference to external */
	/* log to nuke the header rec */
	if (xfs_vol->sb && xfs_vol->log_vol) {
		EngFncs->engine_free(((xfs_volume_t *)xfs_vol->log_vol->private_data)->log_sb);
		EngFncs->engine_free(xfs_vol->log_vol->private_data);
		EngFncs->unassign_fsim_from_volume(xfs_vol->log_vol);
		xfs_vol->log_vol->private_data = NULL;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * "unmkfs" the volume
 *
 */
static int fs_unmkfs_xfs(logical_volume_t * volume)
{
	int rc = EINVAL;
	xfs_volume_t  * xfs_vol = (xfs_volume_t *)volume->private_data;

	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}
	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If mounted, can't unmkfs. */
		LOG_EXIT_INT(EBUSY);
		return EBUSY;
	}
	/* unmkfs the volume */
	rc = fsim_unmkfs_xfs(volume);
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Shrink the volume to new_size. Not supported.
 */
static int fs_shrink( logical_volume_t * volume,
		      sector_count_t     requested_size,
		      sector_count_t   * new_size )
{
	/* unsupported at this time */
	int  rc = ENOSYS;
	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * mkfs has been scheduled.  Do any setup work such as claiming another
 * volume for an external log.
 */
static int fs_mkfs_setup(logical_volume_t * volume,
			 option_array_t   * options)
{
	int  i, tag, rc = 0;
	char  * ext_log_vol = NULL;
	dlist_t  avail_ext_logs;
	logical_volume_t * vol;
	xfs_volume_t * xfs_vol;

	LOG_ENTRY();

	for ( i=0; i<options->count; i++ ) {
		if ( options->option[i].is_number_based ) {
			if (options->option[i].number == MKFS_JOURNAL_VOL_INDEX) {
				/* 'set journal volume' option */
				if ( options->option[i].value.s && strcmp(options->option[i].value.s, NO_SELECTION) ) {
					ext_log_vol = options->option[i].value.s;
				}
			}
		} else {
			if ( !strcmp(options->option[i].name, "journalvol") ) {
				/* 'set journal volume' option */
				if ( options->option[i].value.s && strcmp(options->option[i].value.s, NO_SELECTION) ) {
					ext_log_vol = options->option[i].value.s;
				}
			}
		}
	}
	xfs_vol = EngFncs->engine_alloc(sizeof(xfs_volume_t));
	if ( !xfs_vol ) {
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}
	xfs_vol->sb = EngFncs->engine_alloc(SIZE_OF_SUPER);
	if ( !xfs_vol->sb ) {
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}
	volume->private_data = xfs_vol;


	if (ext_log_vol) {
		/*
		 * Search through the volumes to find and validate the one
		 * selected for an external log, and if valid, claim it.
		 */
		rc = EngFncs->get_volume_list(NULL, NULL, 0, &avail_ext_logs);
		GoToStartOfList(avail_ext_logs);
		while (BlindExtractObject(avail_ext_logs, (TAG *) &tag, NULL, (void**)&vol)==0) {
			/*************************************************************
			 *  TODO: When a min log volume size is put into mkfs.xfs,   *
			 *  add that requirement here for validation of a volume to  *
			 *  be considered for a XFS external log.                    *
			 *************************************************************/
			if ( (vol->file_system_manager == NULL ) &&
			     (!EngFncs->is_mounted(vol->dev_node, NULL)) &&
			     (!strcmp(vol->dev_node, ext_log_vol))) {
				EngFncs->assign_fsim_to_volume(my_plugin_record, vol);
				xfs_vol = EngFncs->engine_alloc(sizeof(xfs_volume_t));
				if ( !xfs_vol ) {
					LOG_EXIT_INT(ENOMEM);
					return ENOMEM;
				}
				xfs_vol->log_sb = EngFncs->engine_alloc(sizeof (xlog_rec_header_t));
				if ( !xfs_vol->log_sb ) {
					LOG_EXIT_INT(ENOMEM);
					return ENOMEM;
				}
				xfs_vol->fs_vol = volume;
				vol->private_data = xfs_vol;
				((xfs_volume_t*)volume->private_data)->log_vol = vol;
			}
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Format the volume.
 */
static int fs_mkfs_xfs(logical_volume_t * volume, option_array_t * options )
{
	int rc = 0;

	LOG_ENTRY();

	/* don't format if mounted */
	if (EngFncs->is_mounted(volume->dev_node, NULL)) {
		return EBUSY;
	}

	rc = fsim_mkfs(volume, options);

	/* re-probe to set up private data */
	if ( !rc ) {
		rc = fs_probe(volume);
	}
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Run fsck on the volume.
 */
static int fs_fsck(logical_volume_t * volume, option_array_t * options )
{
	int rc = EINVAL;

	LOG_ENTRY();

	rc = fsim_fsck( volume, options );

	/*
	 * If fsck.xfs returns FSCK_CORRECTED, the
	 * file system is clean, so return FSCK_OK.
	 */
	if ( rc == FSCK_CORRECTED ) {
		rc = FSCK_OK;
		/*
		 * The value of FSCK_CORRECTED is the same as
		 * EPERM, so fsim_fsck will return -1 for EPERM.
		 */
	} else if (rc == -1) {
		rc = EPERM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Return the total number of supported options for the specified task.
 */
static int fs_get_option_count(task_context_t * context)
{
	int count = 0;

	LOG_ENTRY();

	switch (context->action) {
	case EVMS_Task_mkfs:
		count = MKFS_XFS_OPTIONS_COUNT;
		break;
	case EVMS_Task_fsck:
		count = FSCK_XFS_OPTIONS_COUNT;
		break;
	default:
		count = -1;
		break;
	}

	LOG_EXIT_INT(count);
	return count;
}


/* Get the list of volumes on the system that we can use. */
static int get_volume_list( value_list_t ** value_list,
			    dlist_t         selected_volumes )
{
	int rc = 0;
	int count, i, tag;
	logical_volume_t * vol;
	dlist_t tmp_list, selected_tmp_list;

	LOG_ENTRY();

	rc = EngFncs->get_volume_list(NULL, NULL, 0, &tmp_list);

	/* allocate temp list for 'selected' volumes */
	selected_tmp_list = CreateList();
	if (!selected_tmp_list) {
		LOG_ERROR("Error DLIST Create failed, give up \n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	/* populate temp list with actual 'selected' volumes */
	rc = CopyList(selected_tmp_list, selected_volumes, InsertAtStart);
	if (rc) {
		LOG_ERROR("Error copying list\n");
		LOG_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	/*
	 * Loop through 'selected' volumes, remove them from temp list.
	 */
	GoToStartOfList(selected_tmp_list);
	while (!BlindExtractObject(selected_tmp_list, (TAG *)&tag, NULL, (void **)&vol)) {
		LOG_DETAILS("Volume %s selected, removing from 'available' list\n",vol->name);
		rc = DeleteObject(tmp_list, vol);
		if (rc) {
			LOG_ERROR("Error removing volume %s from list, rc = %d\n",vol->name,rc);
			rc = 0;	  /* Needed to fudge list on modify */
		}
	}
	DestroyList(&selected_tmp_list, FALSE);

	/* if 'available volumes' list already exists, free it and start over */
	if (*value_list) {
		for (i = 0; i < (*value_list)->count; i++) {
			if ((*value_list)->value[i].s) {
				EngFncs->engine_free((*value_list)->value[i].s);
			}
		}
		EngFncs->engine_free(*value_list);
	}
	*value_list = NULL;

	GetListSize(tmp_list, &count);
	/* Increment count to hold the 'None' selection. */
	count++;
	/* allocate and populate 'available volumes' list */
	*value_list = EngFncs->engine_alloc(count * sizeof(value_t) + sizeof(value_list_t));  /* yeah it's too big, but so what */
	if (!rc) {
		i = 0;
		SET_STRING((*value_list)->value[i].s, NO_SELECTION);
		i++;
		while (BlindExtractObject(tmp_list, (TAG *) &tag, NULL, (void**)&vol)==0) {
			/* add proper volumes to 'available volumes' list */
			/*************************************************************
			 *  TODO: When a min log volume size is put into mkfs.xfs,   *
			 *  add that requirement here for validation of a volume to  *
			 *  be considered for a XFS external log.                    *
			 *************************************************************/
			if ( (vol->file_system_manager == NULL ) &&
			     (!EngFncs->is_mounted(vol->dev_node, NULL)) ) {
				(*value_list)->value[i].s = EngFncs->engine_alloc(strlen(vol->name) + 1);
				strcpy((*value_list)->value[i].s, vol->name);
				i++;
			}
		}
		(*value_list)->count = i;
	}
	rc = DestroyList(&tmp_list, FALSE);

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Fill in the initial list of acceptable objects.  Fill in the minimum and
 * maximum nuber of objects that must/can be selected.  Set up all initial
 * values in the option_descriptors in the context record for the given
 * task.  Some fields in the option_descriptor may be dependent on a
 * selected object.  Leave such fields blank for now, and fill in during the
 * set_objects call.
 */
static int fs_init_task( task_context_t * context )
{
	dlist_t global_volumes;
	logical_volume_t * volume;
	void* waste;
	int tag;
	int  rc = 0;
	option_descriptor_t * option = context->option_descriptors->option;  // tmp for option array

	LOG_ENTRY();

	context->min_selected_objects = 0;
	context->max_selected_objects = 1;
	context->option_descriptors->count = 0;

	/* Parameter check */
	if (!context) {
		LOG_EXIT_INT(EFAULT);
		return EFAULT;
	}

	rc = EngFncs->get_volume_list(NULL, NULL, 0, &global_volumes);

	while (!(rc = BlindExtractObject(global_volumes, (TAG *)&tag, NULL, (void **)&volume))) {

		switch (context->action) {
		case EVMS_Task_mkfs:
			/* only mkfs unformatted, unmounted, 'large enough' volumes */
			if ( (volume->file_system_manager == NULL) &&
			     ((volume->vol_size * EVMS_VSECTOR_SIZE) >= MINXFS) &&
			     (!EngFncs->is_mounted(volume->dev_node, NULL))  ) {
				rc = InsertObject(context->acceptable_objects, volume, VOLUME_TAG, NULL, InsertAtStart, TRUE, (void **)&waste);
			}
			break;

		case EVMS_Task_fsck:
			rc = 0;
			break;

		default:
			rc = ENOSYS;
			break;
		}
	}

	if (rc == DLIST_EMPTY || rc == DLIST_END_OF_LIST) {
		rc = 0;
	}

	switch (context->action) {
	
	case EVMS_Task_mkfs:
		if (!xfsutils_support) {
			int answer = 0;
			char * choice_text[2] = {"Ok", NULL};
			EngFncs->user_message(my_plugin_record, &answer,choice_text,
				"xfsutils were either not found or not at the proper version. "
				"xfsutils must be version 2.0.0 or later to function properly with this FSIM. "
				"xfsutils must also reside in the search path specified by PATH. "
				"Please get/install the current version of xfsutils from http://oss.sgi.com/projects/xfs\n" );
			rc = E_CANCELED;
			break;
		}

		context->option_descriptors->count = MKFS_XFS_OPTIONS_COUNT;

		/* Set Volume Label option */
		SET_STRING(option[MKFS_SETVOL_INDEX].name, "vollabel" );
		SET_STRING(option[MKFS_SETVOL_INDEX].title, "Volume Label" );
		SET_STRING(option[MKFS_SETVOL_INDEX].tip, "Set the volume label for the file system." );
		option[MKFS_SETVOL_INDEX].help = NULL;
		option[MKFS_SETVOL_INDEX].type = EVMS_Type_String;
		option[MKFS_SETVOL_INDEX].unit = EVMS_Unit_None;
		option[MKFS_SETVOL_INDEX].min_len = 1;
		option[MKFS_SETVOL_INDEX].max_len = LV_NAME_SIZE;
		option[MKFS_SETVOL_INDEX].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED | EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
		option[MKFS_SETVOL_INDEX].constraint_type = EVMS_Collection_None;
		option[MKFS_SETVOL_INDEX].value.s = EngFncs->engine_alloc(LV_NAME_SIZE + 1);
		if (option[MKFS_SETVOL_INDEX].value.s == NULL) {
			LOG_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		/* Set Journal Volume option */
		SET_STRING(option[MKFS_JOURNAL_VOL_INDEX].name, "journalvol" );
		SET_STRING(option[MKFS_JOURNAL_VOL_INDEX].title, "External Journal Volume" );
		SET_STRING(option[MKFS_JOURNAL_VOL_INDEX].tip, "Set the volume for the journal if other than the XFS file system volume." );
		option[MKFS_JOURNAL_VOL_INDEX].help = NULL;
		option[MKFS_JOURNAL_VOL_INDEX].type = EVMS_Type_String;
		option[MKFS_JOURNAL_VOL_INDEX].unit = EVMS_Unit_None;
		option[MKFS_JOURNAL_VOL_INDEX].min_len = 0;
		option[MKFS_JOURNAL_VOL_INDEX].max_len = EVMS_VOLUME_NAME_SIZE + 1;
			option[MKFS_JOURNAL_VOL_INDEX].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED | EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
		option[MKFS_JOURNAL_VOL_INDEX].constraint_type = EVMS_Collection_List;
		get_volume_list((value_list_t **)&option[MKFS_JOURNAL_VOL_INDEX].constraint.list,
				context->selected_objects);
		option[MKFS_JOURNAL_VOL_INDEX].value.s = EngFncs->engine_alloc(EVMS_VOLUME_NAME_SIZE + 1);
		if (option[MKFS_JOURNAL_VOL_INDEX].value.s == NULL) {
			LOG_EXIT_INT(ENOMEM);
			return ENOMEM;
		}

		/* Log size option */
		SET_STRING(option[MKFS_SETLOGSIZE_INDEX].name, "logsize" );
		SET_STRING(option[MKFS_SETLOGSIZE_INDEX].title, "Log Size" );
		SET_STRING(option[MKFS_SETLOGSIZE_INDEX].tip, "Set log size (in megabytes).  Default log size is 0.4% of volume size.  Used only if no external journal volume is specified." );
		option[MKFS_SETLOGSIZE_INDEX].help = NULL;
		option[MKFS_SETLOGSIZE_INDEX].type = EVMS_Type_Real32;
		option[MKFS_SETLOGSIZE_INDEX].unit = EVMS_Unit_Megabytes;
		option[MKFS_SETLOGSIZE_INDEX].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED;
		option[MKFS_SETLOGSIZE_INDEX].constraint_type = EVMS_Collection_Range;
		option[MKFS_SETLOGSIZE_INDEX].constraint.range = EngFncs->engine_alloc( sizeof(value_range_t) );
		if (option[MKFS_SETLOGSIZE_INDEX].constraint.range == NULL) {
			LOG_EXIT_INT(ENOMEM);
			return ENOMEM;
		}
		option[MKFS_SETLOGSIZE_INDEX].constraint.range->min.r32 = 2.0;
		option[MKFS_SETLOGSIZE_INDEX].constraint.range->max.r32 = 256.0;
		option[MKFS_SETLOGSIZE_INDEX].constraint.range->increment.r32 = 0.1;
		option[MKFS_SETLOGSIZE_INDEX].value.r32 = 4.6;

		context->min_selected_objects = 1;
		context->max_selected_objects = 1;
		break;

	case EVMS_Task_fsck:

		if (!xfsutils_support) {
			int answer = 0;
			char * choice_text[2] = {"Ok", NULL};
			EngFncs->user_message(my_plugin_record, &answer,choice_text,
				"xfsutils were either not found or not at the proper version. "
				"xfsutils must be version 2.0.0 or later to function properly with this FSIM. "
				"xfsutils must also reside in the search path specified by PATH. "
				"Please get/install the current version of xfsutils from http://oss.sgi.com/projects/xfs\n" );
			rc = E_CANCELED;
			break;
		}

		context->option_descriptors->count = FSCK_XFS_OPTIONS_COUNT;

		/* read-only check option */
		SET_STRING(option[FSCK_READONLY_INDEX].name, "readonly" );
		SET_STRING(option[FSCK_READONLY_INDEX].title, "Check Read-Only" );
		SET_STRING(option[FSCK_READONLY_INDEX].tip, "Check the file system READ ONLY.  Report but do not correct errors." );
		option[FSCK_READONLY_INDEX].help = NULL;
		option[FSCK_READONLY_INDEX].type = EVMS_Type_Boolean;
		option[FSCK_READONLY_INDEX].unit = EVMS_Unit_None;
		option[FSCK_READONLY_INDEX].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED;
		option[FSCK_READONLY_INDEX].constraint_type = EVMS_Collection_None;
		option[FSCK_READONLY_INDEX].value.bool = TRUE;

		/* verbose output option */
		SET_STRING(option[FSCK_VERBOSE_INDEX].name, "verbose" );
		SET_STRING(option[FSCK_VERBOSE_INDEX].title, "Verbose Output" );
		SET_STRING(option[FSCK_VERBOSE_INDEX].tip, "Log fsck.xfs details and debug information." );
		option[FSCK_VERBOSE_INDEX].help = NULL;
		option[FSCK_VERBOSE_INDEX].type = EVMS_Type_Boolean;
		option[FSCK_VERBOSE_INDEX].unit = EVMS_Unit_None;
		option[FSCK_VERBOSE_INDEX].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED;
		option[FSCK_VERBOSE_INDEX].constraint_type = EVMS_Collection_None;
		option[FSCK_VERBOSE_INDEX].value.bool = FALSE;

		context->min_selected_objects = 0;
		context->max_selected_objects = 0;

		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);

	return rc;

}


/*
 * Examine the specified value, and determine if it is valid for the task
 * and option_descriptor index. If it is acceptable, set that value in the
 * appropriate entry in the option_descriptor. The value may be adjusted
 * if necessary/allowed. If so, set the effect return value accordingly.
 */
static int fs_set_option( task_context_t * context,
			  u_int32_t        index,
			  value_t        * value,
			  task_effect_t  * effect )
{
	int  rc= 0;

	LOG_ENTRY();

	/* Parameter check */
	if (!context || !value || !effect) {
		return EFAULT;
	}

	*effect = 0;

	switch (context->action) {
	
	case EVMS_Task_mkfs:
		switch (index) {
		
		case MKFS_SETVOL_INDEX:
			/* 'set volume label' option set? */
			strncpy(context->option_descriptors->option[index].value.s, value->s, LV_NAME_SIZE);
			break;

		case MKFS_JOURNAL_VOL_INDEX:
			/* 'set journal volume' option set? */
			strncpy(context->option_descriptors->option[index].value.s, value->s, EVMS_VOLUME_NAME_SIZE);
			break;

		case MKFS_SETLOGSIZE_INDEX:
			/* 'set log size' option set? */
			context->option_descriptors->option[index].value.r32 = value->r32;
			break;

		default:
			break;
		}
		break;

	case EVMS_Task_fsck:
		switch (index) {
		
		case FSCK_READONLY_INDEX:
			/* 'check read only' option set? */
			context->option_descriptors->option[index].value.bool = value->bool;
			break;

		case FSCK_VERBOSE_INDEX:
			/* 'verbose' option set? */
			context->option_descriptors->option[index].value.bool = value->bool;
			break;

		default:
			break;
		}
		break;

	default:
		rc = EINVAL;
		break;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Validate the volumes in the selected_objects dlist in the task context.
 * Remove from the selected objects lists any volumes which are not
 * acceptable.  For unacceptable volumes, create a declined_handle_t
 * structure with the reason why it is not acceptable, and add it to the
 * declined_volumes dlist.  Modify the accepatble_objects dlist in the task
 * context as necessary based on the selected objects and the current
 * settings of the options.  Modify any option settings as necessary based
 * on the selected objects.  Return the appropriate task_effect_t settings
 * if the object list(s), minimum or maximum objects selected, or option
 * settings have changed.
 */
static int fs_set_volumes( task_context_t * context,
			   dlist_t          declined_volumes,
			   task_effect_t  * effect )
{
	int  rc = 0;
	int64_t  log_size;
	double  max_log_size;
	logical_volume_t * vol;

	LOG_ENTRY();

	*effect = 0;

	if (context->action == EVMS_Task_mkfs) {

		/* get the selected volume */
		rc = GetObject(context->selected_objects,VOLUME_TAG,NULL,FALSE,(ADDRESS *) &vol);

		if (!rc) {
			if (EngFncs->is_mounted(vol->dev_node, NULL)) {
				/* If mounted, can't mkfs.xfs. */
				rc = EBUSY;
				LOG_EXIT_INT(rc);
				return rc;
			}
			if ( (vol->vol_size * EVMS_VSECTOR_SIZE) < MINXFS) {

				/***************************************************
				 *  TODO - move this volume to unacceptable list   *
				 ***************************************************/

				MESSAGE( "The size of volume %s is %"PRIu64" bytes.", vol->dev_node, vol->vol_size * EVMS_VSECTOR_SIZE);
				MESSAGE( "mkfs.xfs requires a minimum of %u bytes to build the XFS file system.", MINXFS );
				rc = EPERM;
			} else {
				log_size = vol->vol_size/8196;
				log_size = log_size / 2; // = K
				log_size = log_size / 1024; // = M
				log_size = max(log_size, XFS_MIN_LOG_SIZE);
				log_size = max(log_size, XFS_DEFAULT_LOG_SIZE);

				/* calculate and set upper log size bound */
				max_log_size = min (XFS_MAX_LOG_SIZE, (vol->vol_size/8)/2048);// 1/8th max /2048 = convert to MB
				max_log_size = min (max_log_size, (vol->vol_size - XFS_MIN_DATA_SECTORS)/2048) ;//
				max_log_size = max (max_log_size, XFS_MIN_LOG_SIZE) ;// max can't be less than min
				context->option_descriptors->option[MKFS_SETLOGSIZE_INDEX].constraint.range->max.r32 = max_log_size;

				/* make sure log size is not greater than max log size */
				if ( log_size > max_log_size ) {
					log_size = max_log_size;
				}
				/* initialize log size option */
				context->option_descriptors->option[MKFS_SETLOGSIZE_INDEX].value.r32 = (unsigned int)log_size;
				*effect = EVMS_Effect_Reload_Options;

				/* Remove selected volume from external log selection list. */
				get_volume_list((value_list_t **)&context->option_descriptors->option[MKFS_JOURNAL_VOL_INDEX].constraint.list,
						context->selected_objects);
			}
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * Return any additional information that you wish to provide about the
 * volume.  The Engine privides an external API to get the information
 * stored in the logical_volume_t.  This call is to get any other
 * information about the volume that is not specified in the
 * logical_volume_t.  Any piece of information you wish to provide must be
 * in an extended_info_t structure.  Use the Engine's engine_alloc() to
 * allocate the memory for the extended_info_t.  Also use engine_alloc() to
 * allocate any strings that may go into the extended_info_t.  Then use
 * engine_alloc() to allocate an extended_info_array_t with enough entries
 * for the number of exteneded_info_t structures you are returning.  Fill
 * in the array and return it in *info.
 * If you have extended_info_t descriptors that themselves may have more
 * extended information, set the EVMS_EINFO_FLAGS_MORE_INFO_AVAILABLE flag
 * in the extended_info_t flags field.  If the caller wants more information
 * about a particular extended_info_t item, this API will be called with a
 * pointer to the sotrage_object_t and with a pointer to the name of the
 * extended_info_t item.  In that case, return an extended_info_array_t with
 * further information about the item.  Each of those items may have the
 * EVMS_EINFO_FLAGS_MORE_INFO_AVAILABLE flag set if you desire.  It is your
 * resposibility to give the items unique names so that you know which item
 * the caller is asking additional information for.  If info_name is NULL,
 * the caller just wants top level information about the object.
 */
static int fs_get_volume_info( logical_volume_t        * volume,
			       char                    * info_name,
			       extended_info_array_t * * info )
{
	int rc = EINVAL;
	xfs_volume_t * xfs_vol = (xfs_volume_t *) volume->private_data;
	extended_info_array_t  *Info;
	char magic[12];


	LOG_ENTRY();

	if (!xfs_vol) {
		rc = EINVAL;
		LOG_EXIT_INT(rc);
		return rc;
	}

	if (xfs_vol->sb) {
		/* get and validate current XFS superblock */
		rc = fsim_get_xfs_superblock( volume, xfs_vol->sb );

		if ( (info_name == NULL) && !rc ) {
			fs_get_fs_limits( volume,	       // reset limits.
					  &volume->min_fs_size,
					  &volume->max_vol_size,
					  &volume->max_fs_size);

			Info = EngFncs->engine_alloc( sizeof(extended_info_array_t) + ( 10 * sizeof(extended_info_t) ) );
			if (Info) {

				Info->count = 10;

				SET_STRING( Info->info[0].name, "MagicNumber" );
				SET_STRING( Info->info[0].title, "XFS Magic Number" );
				SET_STRING( Info->info[0].desc, "XFS Magic Number.");
				Info->info[0].type               = EVMS_Type_String;
				Info->info[0].unit               = EVMS_Unit_None;
				sprintf(magic, "0x%x", xfs_vol->sb->sb_magicnum);
				SET_STRING(Info->info[0].value.s, magic);
				//SET_STRING(Info->info[0].value.s, sb_ptr->sb_magicnum);
				Info->info[0].collection_type    = EVMS_Collection_None;
				memset( &Info->info[0].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[1].name, "Version" );
				SET_STRING( Info->info[1].title, "XFS SB Version Number" );
				SET_STRING( Info->info[1].desc, "XFS Super Block Version Number.");
				Info->info[1].type               = EVMS_Type_Unsigned_Int32;
				Info->info[1].unit               = EVMS_Unit_None;
				Info->info[1].value.ui32         = xfs_vol->sb->sb_versionnum & XFS_SB_VERSION_NUMBITS;
				Info->info[1].collection_type    = EVMS_Collection_None;
				memset( &Info->info[1].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[2].name, "State" );
				SET_STRING( Info->info[2].title, "XFS State" );
				SET_STRING( Info->info[2].desc, "The state of XFS.");
				Info->info[2].type               = EVMS_Type_String;
				Info->info[2].unit               = EVMS_Unit_None;
				switch (xfs_vol->sb->sb_flags) {
				case XFS_SBF_READONLY:
					SET_STRING(Info->info[2].value.s, "ReadOnly");
					break;
				default:
					SET_STRING(Info->info[2].value.s, "None");
					break;
				}
				Info->info[2].collection_type    = EVMS_Collection_None;
				memset( &Info->info[2].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[3].name, "VolLabel" );
				SET_STRING( Info->info[3].title, "Volume Label" );
				SET_STRING( Info->info[3].desc, "File system volume label chosen via mkfs.xfs.");
				Info->info[3].type               = EVMS_Type_String;
				Info->info[3].unit               = EVMS_Unit_None;
				SET_STRING(Info->info[3].value.s, xfs_vol->sb->sb_fname);
				Info->info[3].collection_type    = EVMS_Collection_None;
				memset( &Info->info[3].group, 0, sizeof(group_info_t));

        			if (xfs_vol->log_vol) {
        				SET_STRING( Info->info[4].name, "ExtLog" );
        				SET_STRING( Info->info[4].title, "External Log Volume" );
        				SET_STRING( Info->info[4].desc, "Location of exteral journal ");
        				Info->info[4].type               = EVMS_Type_String;
        				Info->info[4].unit               = EVMS_Unit_None;
        				SET_STRING(Info->info[4].value.s, xfs_vol->log_vol->name);
        				Info->info[4].collection_type    = EVMS_Collection_None;
        				memset( &Info->info[4].group, 0, sizeof(group_info_t));
        			} else {
        				SET_STRING( Info->info[4].name, "LogSize" );
        				SET_STRING( Info->info[4].title, "Inline Log Size" );
        				SET_STRING( Info->info[4].desc, "Size of the inline file system log (included in the file system size).");
        				Info->info[4].type               = EVMS_Type_Unsigned_Int64;
        				Info->info[4].unit               = EVMS_Unit_Sectors;
        				Info->info[4].value.ui64         = xfs_vol->sb->sb_logblocks * xfs_vol->sb->sb_blocksize/EVMS_VSECTOR_SIZE;
        				Info->info[4].collection_type    = EVMS_Collection_None;
					memset( &Info->info[4].group, 0, sizeof(group_info_t));
				}

				SET_STRING( Info->info[5].name, "xfsutils" );
				SET_STRING( Info->info[5].title, "XFS Utilities Version" );
				SET_STRING( Info->info[5].desc, "XFS utilities version number.");
				Info->info[5].type               = EVMS_Type_String;
				Info->info[5].unit               = EVMS_Unit_None;
				SET_STRING(Info->info[5].value.s, xfsutils_version);
				Info->info[5].collection_type    = EVMS_Collection_None;
				memset( &Info->info[5].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[6].name, "agcount" );
				SET_STRING( Info->info[6].title, "AG Count" );
				SET_STRING( Info->info[6].desc, "Number of allocation groups");
				Info->info[6].type               = EVMS_Type_Unsigned_Int32;
				Info->info[6].unit               = EVMS_Unit_None;
				Info->info[6].value.ui32         = xfs_vol->sb->sb_agcount;
				Info->info[6].collection_type    = EVMS_Collection_None;
				memset( &Info->info[6].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[7].name, "agsize" );
				SET_STRING( Info->info[7].title, "AG Size" );
				SET_STRING( Info->info[7].desc, "Size of allocation groups");
				Info->info[7].type               = EVMS_Type_Unsigned_Int64;
				Info->info[7].unit               = EVMS_Unit_Kilobytes;
				Info->info[7].value.ui64         = xfs_vol->sb->sb_agblocks * 4;
				Info->info[7].collection_type    = EVMS_Collection_None;
				memset( &Info->info[7].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[8].name, "sunit" );
				SET_STRING( Info->info[8].title, "Stripe Unit" );
				SET_STRING( Info->info[8].desc, "Size of stripe unit");
				Info->info[8].type               = EVMS_Type_Unsigned_Int32;
				Info->info[8].unit               = EVMS_Unit_Sectors;
				Info->info[8].value.ui32         = xfs_vol->sb->sb_unit;
				Info->info[8].collection_type    = EVMS_Collection_None;
				memset( &Info->info[8].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[9].name, "swidth" );
				SET_STRING( Info->info[9].title, "Stripe Width" );
				SET_STRING( Info->info[9].desc, "Width of stripe");
				Info->info[9].type               = EVMS_Type_Unsigned_Int32;
				Info->info[9].unit               = EVMS_Unit_Sectors;
				Info->info[9].value.ui32         = xfs_vol->sb->sb_width;
				Info->info[9].collection_type    = EVMS_Collection_None;
				memset( &Info->info[9].group, 0, sizeof(group_info_t));

				*info = Info;

				rc = 0;
			} else {
				rc = ENOMEM;
			}

		} else {
			/* There is no more information about any of the descriptors. */
			rc = EINVAL;
		}
	} else {
		/* get and validate current XFS external log superblock */
		rc = fsim_get_log_superblock( volume, xfs_vol->log_sb);

		if ( (info_name == NULL) && !rc ) {
			Info = EngFncs->engine_alloc( sizeof(extended_info_array_t) + ( 6 * sizeof(extended_info_t) ) );
			if (Info) {

				Info->count = 5;

				SET_STRING( Info->info[0].name, "LogMagicNumber" );
				SET_STRING( Info->info[0].title, "XFS External Log Magic Number" );
				SET_STRING( Info->info[0].desc, "XFS External Log Magic Number.");
				Info->info[0].type               = EVMS_Type_String;
				Info->info[0].unit               = EVMS_Unit_None;
				sprintf(magic, "0x%x", xfs_vol->log_sb->h_magicno);
				SET_STRING(Info->info[0].value.s, magic);
				Info->info[0].collection_type    = EVMS_Collection_None;
				memset( &Info->info[0].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[1].name, "LogVersion" );
				SET_STRING( Info->info[1].title, "XFS External Log Version Number" );
				SET_STRING( Info->info[1].desc, "XFS External Log Version Number.");
				Info->info[1].type               = EVMS_Type_Unsigned_Int32;
				Info->info[1].unit               = EVMS_Unit_None;
				Info->info[1].value.i32          = xfs_vol->log_sb->h_version;
				Info->info[1].collection_type    = EVMS_Collection_None;
				memset( &Info->info[1].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[2].name, "fsvol" );
				SET_STRING( Info->info[2].title, "Journal for:" );
				SET_STRING( Info->info[2].desc, "Volume for which this is an external log volume");
				Info->info[2].type               = EVMS_Type_String;
				Info->info[2].unit               = EVMS_Unit_None;
				if (xfs_vol->fs_vol) {
					SET_STRING(Info->info[2].value.s, xfs_vol->fs_vol->name);
				} else {
					SET_STRING(Info->info[2].value.s, "UNKNOWN");
				}
				Info->info[2].collection_type    = EVMS_Collection_None;
				memset( &Info->info[2].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[3].name, "LogSize" );
				SET_STRING( Info->info[3].title, "Log Size" );
				SET_STRING( Info->info[3].desc, "Size of the file system log (included in the file system size).");
				Info->info[3].type               = EVMS_Type_Unsigned_Int64;
				Info->info[3].unit               = EVMS_Unit_Sectors;
				if (xfs_vol->fs_vol) {
					Info->info[3].value.ui64         = ((xfs_volume_t *)(xfs_vol->fs_vol->private_data))->sb->sb_logblocks *
									   (((xfs_volume_t *)(xfs_vol->fs_vol->private_data))->sb->sb_blocksize /
									    EVMS_VSECTOR_SIZE);
				} else {
					Info->info[3].value.ui64         = 0;
				}
				Info->info[3].collection_type    = EVMS_Collection_None;
				memset( &Info->info[3].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[4].name, "xfsutils" );
				SET_STRING( Info->info[4].title, "XFS Utilities Version" );
				SET_STRING( Info->info[4].desc, "XFS utilities version number.");
				Info->info[4].type               = EVMS_Type_String;
				Info->info[4].unit               = EVMS_Unit_None;
				SET_STRING(Info->info[4].value.s, xfsutils_version);
				Info->info[4].collection_type    = EVMS_Collection_None;
				memset( &Info->info[4].group, 0, sizeof(group_info_t));

				*info = Info;

				rc = 0;
			} else {
				rc = ENOMEM;
			}

		} else {
			/* There is no more information about any of the descriptors. */
			rc = EINVAL;
		}
	}
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 *  Returns Plugin specific information ...
 */
static int fs_get_plugin_info( char * descriptor_name, extended_info_array_t * * info )
{
	int                      rc = EINVAL;
	extended_info_array_t   *Info;
	char                     version_string[64];
	char                     required_engine_api_version_string[64];
	char                     required_fsim_api_version_string[64];

	LOG_ENTRY();

	if (info) {

		if (descriptor_name == NULL) {
			*info = NULL;	  // init to no info returned

			Info = EngFncs->engine_alloc( sizeof(extended_info_array_t) + (6*sizeof(extended_info_t))  );
			if (Info) {

				Info->count = 6;

				sprintf(version_string, "%d.%d.%d",
					MAJOR_VERSION,
					MINOR_VERSION,
					PATCH_LEVEL );

				sprintf(required_engine_api_version_string, "%d.%d.%d",
					my_plugin_record->required_engine_api_version.major,
					my_plugin_record->required_engine_api_version.minor,
					my_plugin_record->required_engine_api_version.patchlevel );

				sprintf(required_fsim_api_version_string, "%d.%d.%d",
					my_plugin_record->required_plugin_api_version.fsim.major,
					my_plugin_record->required_plugin_api_version.fsim.minor,
					my_plugin_record->required_plugin_api_version.fsim.patchlevel );

				SET_STRING( Info->info[0].name, "Short Name" );
				SET_STRING( Info->info[0].title, "Short Name" );
				SET_STRING( Info->info[0].desc, "A short name given to this plugin.");
				Info->info[0].type               = EVMS_Type_String;
				Info->info[0].unit               = EVMS_Unit_None;
				SET_STRING( Info->info[0].value.s, my_plugin_record->short_name );
				Info->info[0].collection_type    = EVMS_Collection_None;
				memset( &Info->info[0].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[1].name, "Long Name" );
				SET_STRING( Info->info[1].title, "Long Name" );
				SET_STRING( Info->info[1].desc, "A long name given to this plugin.");
				Info->info[1].type               = EVMS_Type_String;
				Info->info[1].unit               = EVMS_Unit_None;
				SET_STRING( Info->info[1].value.s, my_plugin_record->long_name );
				Info->info[1].collection_type    = EVMS_Collection_None;
				memset( &Info->info[1].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[2].name, "Type" );
				SET_STRING( Info->info[2].title, "Plugin Type" );
				SET_STRING( Info->info[2].desc, "There are various types of plugins; each responsible for some kind of storage object.");
				Info->info[2].type               = EVMS_Type_String;
				Info->info[2].unit               = EVMS_Unit_None;
				SET_STRING( Info->info[2].value.s, "File System Interface Module" );
				Info->info[2].collection_type    = EVMS_Collection_None;
				memset( &Info->info[2].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[3].name, "Version" );
				SET_STRING( Info->info[3].title, "Plugin Version" );
				SET_STRING( Info->info[3].desc, "This is the version number of the plugin.");
				Info->info[3].type               = EVMS_Type_String;
				Info->info[3].unit               = EVMS_Unit_None;
				SET_STRING( Info->info[3].value.s, version_string );
				Info->info[3].collection_type    = EVMS_Collection_None;
				memset( &Info->info[3].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[4].name, "Required Engine Services API Version" );
				SET_STRING( Info->info[4].title, "Required Engine Services API Version" );
				SET_STRING( Info->info[4].desc, "This is the version of the Engine services that this plug-in requires. It will not run on older versions of the Engine services.");
				Info->info[4].type               = EVMS_Type_String;
				Info->info[4].unit               = EVMS_Unit_None;
				SET_STRING( Info->info[4].value.s, required_engine_api_version_string );
				Info->info[4].collection_type    = EVMS_Collection_None;
				memset( &Info->info[4].group, 0, sizeof(group_info_t));

				SET_STRING( Info->info[5].name, "Required Engine FSIM API Version" );
				SET_STRING( Info->info[5].title, "Required Engine FSIM API Version" );
				SET_STRING( Info->info[5].desc, "This is the version of the Engine FSIM API that this plug-in requires. It will not run on older versions of the Engine FSIM API.");
				Info->info[5].type               = EVMS_Type_String;
				Info->info[5].unit               = EVMS_Unit_None;
				SET_STRING( Info->info[5].value.s, required_fsim_api_version_string );
				Info->info[5].collection_type    = EVMS_Collection_None;
				memset( &Info->info[5].group, 0, sizeof(group_info_t));

				*info = Info;

				rc = 0;
			} else {
				rc = ENOMEM;
			}

		} else {
			/* There is no more information on any of the descriptors. */
			rc = EINVAL;
		}
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * How much can file system expand?
 */
static int fs_can_expand_by(logical_volume_t * volume,
			    sector_count_t   * delta)
{
	int  rc = 0;

	LOG_ENTRY();
	if (!xfsutils_support) {
		rc = EINVAL;
	} else if (!EngFncs->is_mounted(volume->dev_node, NULL)) {
		/* If not mounted, can't expand. online only. */
		rc = EBUSY;
	} else if (volume->fs_size + *delta > volume->max_fs_size) {
		*delta = volume->max_fs_size - volume->fs_size;
	}
	LOG_EXIT_INT(rc);
	return rc;

}


/*
 * How much can file system shrink?
 */
static int fs_can_shrink_by(logical_volume_t * volume,
			    sector_count_t * delta)
{
	/* unsupported at this time */
	int  rc = ENOSYS;
	LOG_ENTRY();
	LOG_EXIT_INT(rc);
	return rc;
}

static fsim_functions_t fsim_ops = {
	.setup_evms_plugin	= fs_setup,
	.cleanup_evms_plugin	= fs_cleanup,
	.is_this_yours		= fs_probe,
	.can_mkfs		= fs_can_mkfs,
	.can_unmkfs		= fs_can_unmkfs,
	.can_fsck		= fs_can_fsck,
	.get_fs_size		= fs_get_fs_size,
	.get_fs_limits		= fs_get_fs_limits,
	.can_expand_by		= fs_can_expand_by,
	.can_shrink_by		= fs_can_shrink_by,
	.expand			= fs_expand,
	.shrink			= fs_shrink,
	.mkfs			= fs_mkfs_xfs,
	.mkfs_setup		= fs_mkfs_setup,
	.fsck			= fs_fsck,
	.unmkfs			= fs_unmkfs_xfs,
	.unmkfs_setup		= fs_unmkfs_setup,
	.get_option_count	= fs_get_option_count,
	.init_task		= fs_init_task,
	.set_option		= fs_set_option,
	.set_volumes		= fs_set_volumes,
	.get_volume_info	= fs_get_volume_info,
	.get_plugin_info	= fs_get_plugin_info
};

plugin_record_t xfs_plugin_record = {
	.id = SetPluginID(EVMS_OEM_IBM,
			  EVMS_FILESYSTEM_INTERFACE_MODULE,
			  FS_TYPE_XFS),
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 11,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.fsim = {
			.major		= 8,
			.minor		= 0,
			.patchlevel	= 0
		}
	},
	.short_name = "XFS",
	.long_name = "XFS File System Interface Module",
	.oem_name = "IBM",
	.functions = {
		.fsim = &fsim_ops
	},
	.container_functions = NULL
};

plugin_record_t *evms_plugin_records[] = {
	&xfs_plugin_record,
	NULL
};

