/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: fsimxfs.c
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <plugin.h>
#include "fsimxfs.h"

void fsim_swap_xfs_superblock( xfs_sb_t * );
void fsim_swap_log_superblock( xlog_rec_header_t * );

int fsim_rw_diskblocks( logical_volume_t *,int, int64_t, int32_t, void *, int );

extern char xfsutils_version[10];
extern u_int32_t xfsutils_support;

/*-------------------------------------------------------------------------------------+
+                                                                                      +
+                                   Common Routines                                    +
+                                                                                      +
+--------------------------------------------------------------------------------------*/

/*
 * Un-Format the volume.
 */
int fsim_unmkfs_xfs( logical_volume_t * volume )
{
	int   fd, fd2, rc = 0;
	xfs_volume_t * xfs_vol = (xfs_volume_t*)volume->private_data;

	fd = EngFncs->open_volume(volume, O_RDWR|O_EXCL, 0);
	if (fd < 0) return -1;

	if (xfs_vol->sb) {
		/* clear private data */
		memset( (void *)xfs_vol->sb, 0, SIZE_OF_SUPER );
		/* zero primary superblock */
		rc =  fsim_rw_diskblocks( volume, fd, XFS_SUPER1_OFF, SIZE_OF_SUPER, xfs_vol->sb, PUT );

		/* if volume has external log, delete it as well */
		if (xfs_vol->log_vol) {
			fd2 = EngFncs->open_volume(xfs_vol->log_vol, O_RDWR|O_EXCL, 0);
			if (fd < 0) return -1;
			rc =  fsim_rw_diskblocks( volume, fd2, XFS_SUPER1_OFF, SIZE_OF_SUPER, xfs_vol->sb, PUT );
			fd = EngFncs->close_volume(volume, fd2);
		}
		EngFncs->engine_free(xfs_vol->sb);
		EngFncs->engine_free(volume->private_data);
		volume->private_data = NULL;
	} else if (xfs_vol->log_sb && !xfs_vol->fs_vol) {     // if orphaned log
		fd2 = EngFncs->open_volume(volume, O_RDWR|O_EXCL, 0);
		if (fd < 0) return -1;
		memset(xfs_vol->log_sb, 0x00, sizeof (xlog_rec_header_t));
		rc =  fsim_rw_diskblocks( volume, fd2, XFS_SUPER1_OFF, SIZE_OF_SUPER, xfs_vol->log_sb, PUT );
		fd = EngFncs->close_volume(volume, fd2);
		EngFncs->engine_free(xfs_vol->log_sb);
		EngFncs->engine_free(volume->private_data);
		volume->private_data = NULL;
	} else {
		rc = ERROR;
	}

	fd = EngFncs->close_volume(volume, fd);

	return rc;
}


/*
 * NAME: set_mkfs_options
 *
 * FUNCTION: Build options array (argv) for mkfs.xfs
 *
 * PARAMETERS:
 *      options   - options array passed from EVMS engine
 *      argv      - mkfs options array
 *      volume    - volume on which program will be executed
 *
 * return number of args used
 */
static int set_mkfs_options( option_array_t * options,
			     char ** argv,
			     logical_volume_t * volume)
{
	int i, opt_count = 0;

	strncpy(argv[opt_count++], "mkfs.xfs", 64);

	for ( i=0; i<options->count; i++ ) {

		if ( options->option[i].is_number_based ) {

			switch (options->option[i].number) {
			
			case MKFS_SETVOL_INDEX:
				/* 'set volume name' option */
				if ( options->option[i].value.s ) {
					sprintf(argv[opt_count++],"-L%s",options->option[i].value.s);
				}
				break;

			case MKFS_JOURNAL_VOL_INDEX:
				/* 'set journal volume' option */
				if ( options->option[i].value.s && strcmp(options->option[i].value.s, NO_SELECTION) ) {
					sprintf(argv[opt_count++],"-llogdev=%s",options->option[i].value.s);
				}
				break;

			case MKFS_SETLOGSIZE_INDEX:
				/* 'set log size' option */
				if ( options->option[i].value.r32 ) {
					u_int32_t tmp = (u_int32_t)(options->option[i].value.r32*1024*1024);
					tmp = tmp & ~(4095); //4k multiple
					sprintf(argv[opt_count++], "-lsize=%d", tmp);
				}
				break;

			default:
				break;
			}

		} else {

			if ( !strcmp(options->option[i].name, "vollabel") ) {
				/* 'set volume label' option */
				if ( options->option[i].value.s ) {
					sprintf(argv[opt_count++],"-L%s",options->option[i].value.s);
				}
			}

			if ( !strcmp(options->option[i].name, "journalvol") ) {
				/* 'set journal vol' option */
				if ( options->option[i].value.s && strcmp(options->option[i].value.s, NO_SELECTION) ) {
					sprintf(argv[opt_count++],"-llogdev=%s",options->option[i].value.s);
				}
			}

			if ( !strcmp(options->option[i].name, "logsize") ) {
				/* 'set log size' option */
				if ( options->option[i].value.r32 ) {
					sprintf(argv[opt_count++], "-lsize=%d", (int)options->option[i].value.r32*1024*1024);
				}
			}
		}
	}

	sprintf(argv[opt_count++],"%s", volume->dev_node);

	return (opt_count);
}



/*
 * Format the volume.
 */
int fsim_mkfs(logical_volume_t * volume, option_array_t * options )
{
	int     rc = FSIM_ERROR;
	char   *argv[MKFS_XFS_OPTIONS_COUNT_TOTAL];
	pid_t   pidm;
	int     status;
	int     fds2[2];
	char    *buffer = NULL;
	int i, last;
	int bytes_read = 0;

	LOG_ENTRY();

	/* open pipe, alloc buffer for collecting fsck.xfs output */
	if ( (rc = pipe(fds2)) ) {
		return rc;
	}

	for (i=0; i < (MKFS_XFS_OPTIONS_COUNT_TOTAL); i++) {
		argv[i] = EngFncs->engine_alloc(64);
	}

	if (!(buffer = EngFncs->engine_alloc(MAX_USER_MESSAGE_LEN))) {
		close(fds2[0]);
		close(fds2[1]);
		return ENOMEM;
	}

	last = set_mkfs_options( options, argv, volume);

	for (i=last; i < MKFS_XFS_OPTIONS_COUNT_TOTAL; i++) {
		EngFncs->engine_free(argv[i]);
		argv[i] = NULL;
	}
/* debug
	for (i=0; i < (MKFS_XFS_OPTIONS_COUNT +7); i++) {
		printf("arg %d = %s \n",i,argv[i]);
	} */

	pidm = EngFncs->fork_and_execvp(volume, argv, NULL, fds2, fds2);
	if (pidm != -1) {
		fcntl(fds2[0], F_SETFL, fcntl(fds2[0], F_GETFL,0) | O_NONBLOCK);
		waitpid( pidm, &status, 0 );
		if ( WIFEXITED(status) ) {
			bytes_read = read(fds2[0],buffer,MAX_USER_MESSAGE_LEN);
			if (bytes_read > 0) {
				LOG_DETAILS("mkfs output: \n%s",buffer); // on prompt if there is one
				memset(buffer,0,bytes_read); //clear out message
			}
			/* get mkfs.xfs exit code */
			rc = WEXITSTATUS(status);
			if (rc == 0) {
				LOG_DETAILS("mkfs.xfs completed with exit code %d \n", status);
			} else {
				LOG_ERROR("mkfs.xfs completed with exit code %d \n", status);
			}
		} else {
			rc = EINTR;
		}
	} else {
		rc = EIO;
	}
	EngFncs->engine_free(buffer);

	for (i=0; i < last; i++) {
		EngFncs->engine_free(argv[i]);
	}
	close(fds2[0]);
	close(fds2[1]);

	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * NAME: set_fsck_options
 *
 * FUNCTION: Build options array (argv) for fsck.xfs
 *
 * PARAMETERS:
 *      options   - options array passed from EVMS engine
 *      argv      - fsck options array
 *      volume    - volume on which program will be executed
 *
 */
static int set_fsck_options( option_array_t * options, char ** argv, logical_volume_t * volume )
{
	int i, opt_count = 0;

	sprintf(argv[opt_count++], "xfs_repair");

	for ( i=0; i<options->count; i++) {

		if ( options->option[i].is_number_based ) {


			/* 'check read only' option or mounted */
			if ( ( (options->option[i].number == FSCK_READONLY_INDEX) &&
			       (options->option[i].value.bool == TRUE) ) ||
			     EngFncs->is_mounted(volume->dev_node, NULL) ) {
				/* set fsck.xfs -n option */
				sprintf(argv[opt_count++],"-n");
			}

			/* 'verbose output' option */
			if ( (options->option[i].number == FSCK_VERBOSE_INDEX) &&
			     (options->option[i].value.bool == TRUE) ) {
				/* set fsck.xfs -v option */
				sprintf(argv[opt_count++], "-v");
			}


		} else {


			/* 'check read only' option selected or mounted */
			if ( ( (!strcmp(options->option[i].name, "readonly")) &&
			       (options->option[i].value.bool == TRUE) ) ||
			     EngFncs->is_mounted(volume->dev_node, NULL) ) {
				/* set fsck.xfs -n option */
				sprintf(argv[opt_count++],"-n");
			}

			/* 'verbose output' option selected */
			if ( !strcmp(options->option[i].name, "verbose") &&
			     (options->option[i].value.bool == TRUE) ) {
				/* set fsck.xfs -v option */
				sprintf(argv[opt_count++], "-v");
			}

		}
	}

	sprintf(argv[opt_count++],"%s", volume->dev_node);

	return(opt_count);
}



/*
 * Run fsck on the volume.
 */
int fsim_fsck(logical_volume_t * volume, option_array_t * options )
{
	int     rc = FSIM_ERROR;
	char   *argv[FSCK_XFS_OPTIONS_COUNT_TOTAL];
	pid_t   pidf;
	int     status, bytes_read;
	int     fds2[2];
	char    *buffer = NULL;
	int i, last;


	for (i=0; i < FSCK_XFS_OPTIONS_COUNT_TOTAL; i++) {
		argv[i] = EngFncs->engine_alloc(64);
	}

	/* open pipe, alloc buffer for collecting fsck.xfs output */
	if ( (rc = pipe(fds2)) ) {
		return rc;
	}
	if (!(buffer = EngFncs->engine_alloc(MAX_USER_MESSAGE_LEN))) {
		close(fds2[0]);
		close(fds2[1]);
		return ENOMEM;
	}

	last = set_fsck_options( options, argv, volume );

	for (i=last; i < FSCK_XFS_OPTIONS_COUNT_TOTAL; i++) {
		EngFncs->engine_free(argv[i]);
		argv[i] = NULL;
	}
/* debug
	for (i=0; i < (MKFS_XFS_OPTIONS_COUNT +7); i++) {
		printf("arg %d = %s \n",i,argv[i]);
	} */

	pidf = EngFncs->fork_and_execvp(volume, argv, NULL, fds2, fds2);
	if (pidf != -1) {
		fcntl(fds2[0], F_SETFL, fcntl(fds2[0], F_GETFL,0) | O_NONBLOCK);
		while (!(waitpid( pidf, &status, WNOHANG ))) {
			bytes_read = read(fds2[0],buffer,MAX_USER_MESSAGE_LEN);
			if (bytes_read > 0) {
				MESSAGE("check output: \n%s",buffer); // on prompt if there is one
				memset(buffer,0,bytes_read); //clear out message
			}
			usleep(10000); // don't hog all the cpu
		}
		if ( WIFEXITED(status) && (WEXITSTATUS(status) != ENOENT)) {
			do {
				bytes_read = read(fds2[0],buffer,MAX_USER_MESSAGE_LEN);
				if (bytes_read > 0) {
					MESSAGE("check output1: \n%s",buffer);
				}
			} while (bytes_read == MAX_USER_MESSAGE_LEN);
			rc = WEXITSTATUS(status);
			if (rc == 0) {
				LOG_DETAILS("check completed with rc = %d \n",status);
			} else {
				LOG_ERROR("check completed with rc = %d \n",status);
			}
		} else {
			rc = EINTR;
		}

	} else {
		rc = EIO;
	}

	EngFncs->engine_free(buffer);
	for (i=0; i < (FSCK_XFS_OPTIONS_COUNT_TOTAL); i++) {
		EngFncs->engine_free(argv[i]);
	}
	close(fds2[0]);
	close(fds2[1]);

	return rc;
}


/*
 * NAME: fsim_swap_xfs_superblock
 *
 * FUNCTION: Endian swap a XFS superblock
 *
 * PARAMETERS:
 *      sblk   - pointer to superblock
 */
void fsim_swap_xfs_superblock( xfs_sb_t *sblk )
{
	sblk->sb_versionnum   = XDISK_TO_CPU16(sblk->sb_versionnum);
	sblk->sb_blocksize    = XDISK_TO_CPU32(sblk->sb_blocksize);
	sblk->sb_magicnum     = XDISK_TO_CPU32(sblk->sb_magicnum);
	sblk->sb_dblocks      = XDISK_TO_CPU64(sblk->sb_dblocks);
	sblk->sb_rblocks      = XDISK_TO_CPU64(sblk->sb_rblocks);
	sblk->sb_logblocks    = XDISK_TO_CPU32(sblk->sb_logblocks);
	sblk->sb_agcount      = XDISK_TO_CPU32(sblk->sb_agcount);
	sblk->sb_agblocks     = XDISK_TO_CPU32(sblk->sb_agblocks);
	sblk->sb_unit         = XDISK_TO_CPU32(sblk->sb_unit);
	sblk->sb_width        = XDISK_TO_CPU32(sblk->sb_width);
//	BUGBUG need more ?????
	return;
}


/*
 * NAME: fsim_get_xfs_superblock
 *
 * FUNCTION: Get and validate a XFS superblock
 *
 * PARAMETERS:
 *      volume   - pointer to volume from which to get the superblock
 *      sb_ptr   - pointer to superblock
 *
 * RETURNS:
 *      (0) for success
 *      != 0 otherwise
 *
 */
int fsim_get_xfs_superblock( logical_volume_t *volume, xfs_sb_t *sb_ptr )
{
	int  fd, rc = 0;
	LOG_ENTRY();

	fd = EngFncs->open_volume(volume, O_RDONLY, 0);
	if (fd < 0) return EIO;

	/* get primary superblock */
	rc = fsim_rw_diskblocks( volume, fd, XFS_SUPER1_OFF, SIZE_OF_SUPER, sb_ptr, GET );
	fsim_swap_xfs_superblock( sb_ptr );

	if ( rc == FSIM_SUCCESS ) {
		/* see if primary superblock is XFS */
		if ( ( (sb_ptr->sb_versionnum & XFS_SB_VERSION_NUMBITS) < XFS_SB_VERSION_1 ) ||
		     ( sb_ptr->sb_magicnum != XFS_SB_MAGIC)) {
			rc = FSIM_ERROR;
		}
	}

	EngFncs->close_volume(volume, fd);
	return rc;
}

/*
 * NAME: fsim_swap_log_superblock
 *
 * FUNCTION: Endian swaps a XFS external log superblock
 *
 * PARAMETERS:
 *      sblk   - pointer to log header
 */
void fsim_swap_log_superblock( xlog_rec_header_t * sblk )
{

	LOG_ENTRY();
	sblk->h_magicno = XDISK_TO_CPU32(sblk->h_magicno);
	sblk->h_version = XDISK_TO_CPU32(sblk->h_version);
	LOG_EXIT_VOID();
}


/*
 * NAME: fsim_get_log_superblock
 *
 * FUNCTION: Get and validate a XFS log superblock
 *
 * PARAMETERS:
 *      dev_node     - pointer to volume from which to get the superblock
 *      log_sb_ptr   - pointer to superblock
 *
 * RETURNS:
 *      (0) for success
 *      != 0 otherwise
 *
 */
int fsim_get_log_superblock( logical_volume_t * volume, xlog_rec_header_t *log_sb_ptr )
{
	int  fd, rc = 0;

	LOG_ENTRY();
	fd = EngFncs->open_volume(volume, O_RDONLY, 0);
	if (fd < 0) return EIO;

	/* get log superblock */
	rc = fsim_rw_diskblocks( volume, fd, 0, sizeof(xlog_rec_header_t), log_sb_ptr, GET );
	fsim_swap_log_superblock( log_sb_ptr );

	if ( rc == FSIM_SUCCESS ) {
		/* see if superblock is XFS log superblock */
		if ( log_sb_ptr->h_magicno != XLOG_HEADER_MAGIC_NUM ) {
			rc = FSIM_ERROR;
		}
	}

	EngFncs->close_volume(volume, fd);
	LOG_EXIT_INT(rc);
	return rc;
}


/*
 * NAME: fsim_rw_diskblocks
 *
 * FUNCTION: Read or write specific number of bytes for an opened device.
 *
 * PARAMETERS:
 *      dev_ptr         - file handle of an opened device to read/write
 *      disk_offset     - byte offset from beginning of device for start of disk
 *                        block read/write
 *      disk_count      - number of bytes to read/write
 *      data_buffer     - On read this will be filled in with data read from
 *                        disk; on write this contains data to be written
 *      mode            - GET (read) or PUT (write)
 *
 * RETURNS:
 *      FSIM_SUCCESS (0) for success
 *      ERROR       (-1) can't lseek
 *      EINVAL
 *      EIO
 *
 */
int fsim_rw_diskblocks( logical_volume_t * volume,
			int      dev_ptr,
			int64_t  disk_offset,
			int32_t  disk_count,
			void     *data_buffer,
			int      mode )
{
	size_t  Bytes_Transferred;

	switch ( mode ) {
	case GET:
		Bytes_Transferred = EngFncs->read_volume(volume,dev_ptr,data_buffer,disk_count,disk_offset);
		break;
	case PUT:
		Bytes_Transferred = EngFncs->write_volume(volume,dev_ptr,data_buffer,disk_count,disk_offset);
		break;
	default:
		return EINVAL;
		break;
	}

	if ( Bytes_Transferred != disk_count ) {
		return EIO;
	}

	return FSIM_SUCCESS;
}


/*
 * Test mkfs.xfs version.
 *    xfsutils must be at version 1.0.9 or greater to function properly
 *    with this FSIM.  That also happens to be the first version in which
 *    mkfs.xfs supported the -V (version) option.  So, run mkfs.xfs with
 *    the -V option,and if it succeeds, the installed version is OK.  If
 *    it fails, pass back the exit code (should be EINVAL) of the failure.
 */
int fsim_test_version( )
{
	int     rc = FSIM_ERROR;
//	int     dev_rc = 0;
	char   *argv[3];
	pid_t   pidm;
	int     status, bytes_read;
	int     fds2[2];
	char    *buffer = NULL;
	char    *buf_ptr = NULL;
	int     ver_string_size;

	/* open pipe, alloc buffer for collecting fsck.xfs output */
	rc = pipe(fds2);
	if (rc) {
		return(rc);
	}
	if (!(buffer = EngFncs->engine_alloc(PAGE_SIZE))) {
		close(fds2[0]);
		close(fds2[1]);
		return(ENOMEM);
	}

	argv[0] = "mkfs.xfs";
	argv[1] = "-V";
	argv[2] = NULL;

	pidm = EngFncs->fork_and_execvp(NULL, argv, NULL, fds2, fds2);
	if (pidm != -1) {
		fcntl(fds2[0], F_SETFL, fcntl(fds2[0], F_GETFL,0) | O_NONBLOCK);
		waitpid( pidm, &status, 0 );
		if ( WIFEXITED(status) ) {
			bytes_read = read(fds2[0],buffer,PAGE_SIZE);
			if (bytes_read > 0) {
				/* find start and end of version number, store in xfsutils_version */
				buf_ptr = strstr(buffer, "version");
				if (buf_ptr) {
					buf_ptr = strstr(buf_ptr, " ") + 1;
					ver_string_size = strstr(buf_ptr, "\n") - buf_ptr;
					if (ver_string_size > sizeof(xfsutils_version)) {
						ver_string_size = sizeof(xfsutils_version);
					}
					xfsutils_support = XFS_VALID_UTILS;
					strncpy(xfsutils_version, buf_ptr, ver_string_size);
					if ( strcmp(xfsutils_version, "2.1.0") >= 0 ) {
						xfsutils_support |= XFS_V2_LOG;
					}
				} else {
					memset(xfsutils_version, 0, 9);
				}
			}
			/* get mkfs.xfs exit code */
			rc = WEXITSTATUS(status);
			if (rc == 1) {
				rc = 0; //old xfs exits with rc=1 on version check
			}
			LOG_DETAILS("mkfs.xfs test version completed with exit code %d \n", rc);
		} else {
//			rc = EINTR;
		}
	} else {
		rc = EIO;
	}

	EngFncs->engine_free(buffer);

	return(0);
}


